/*
 * $Id: CORPUSER.C,v 1.2 2005-10-04 09:07:18 tonydodd Exp $
 *
 * $Log: not supported by cvs2svn $
 * Revision 1.3  1996/02/22 11:37:52  scotty
 * Now includes limits.h; OSF/1 was picking it up anyway but SunOS didn't.
 *
 * Revision 1.2  1996/02/15 02:13:03  scotty
 * Now creates files with permissions set to ACC_MODE (defined
 * in Makefile).
 *
 * Revision 1.1  1996/02/15 01:29:39  scotty
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include <sys/stat.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#ifdef HAVE_CRYPT_H
#include <crypt.h>
#endif
#include "Corpinc.h"
#include "CORPUSER.H"

static char rcsid[] = "$Id: CORPUSER.C,v 1.2 2005-10-04 09:07:18 tonydodd Exp $";

#define BUFLEN    4096
#define MAXTRIES  3
#define TRYWAIT   2

//char *crypt(const char *, const char *);

#if defined(__ultrix__)
char *strdup(char *src)
{
   char *dst;

   if ((dst = malloc(strlen(src) + 1)) == 0)
      return 0;
   else
      strcpy(dst, src);

   return dst;
}

#endif /* __ultrix__ */

enum {
   U_PWD, U_REAL, U_ORG, U_CTCT, U_SUCC, U_UNSUCC, U_EXP
};

static char *rgszNames[] = {
   "pwd", "realname", "organization", "contact",
   "lastsucc", "lastunsucc", "expires"
};

static USR *pusrParse(char *szName, FILE *fpIn)
{
   char    *pch;
   char    *pchWord;
   char    *pchValue;
   USR     *pusrNew;
   char     rgchBuf[BUFLEN];
   int      i;
   int      f_pwd = 0;
   int      f_real = 0;
   int      f_org = 0;
   int      f_ctct = 0;
   int      f_succ = 0;
   int      f_unsucc = 0;
   int      f_exp = 0;
   int      line = 0;

   if ((pusrNew = (USR*) malloc(sizeof(USR))) == 0) {
      syslog(LOG_ERR, "OUT OF MEMORY");
      return 0;
   }

   pusrNew->u_name = 0;

   while (pch = fgets(rgchBuf, BUFLEN, fpIn)) {
      line++;
      if (pch == 0) {
         syslog(LOG_ERR, "Premature end of stanza file for %s", szName);
         free(pusrNew);
         return 0;
      }
      if (rgchBuf[0] != '#')
         break;
   }

   for (pch = rgchBuf; *pch && !isspace(*pch) && *pch != ':'; pch++)
      ;

   if (*pch != ':') {
      *pch = '\0';
      for (++pch; *pch && *pch != ':'; pch++)
         ;
      if (*pch != ':') {
         syslog(LOG_ERR, "Badly formed stanza file for %s, line %d",
                szName, line);
         free(pusrNew);
         return 0;
      }
   } else
      *pch = '\0';

   if (strcasecmp(rgchBuf, szName)) {
      syslog(LOG_ERR, "Corrupted stanza file for %s", szName);
      free(pusrNew);
      return 0;
   } else {
      if ((pusrNew->u_name = strdup(rgchBuf)) == 0) {
         syslog(LOG_ERR, "OUT OF MEMORY");
         return 0;
      }
   }

   while (fgets(rgchBuf, BUFLEN, fpIn) != 0) {
      line++;
      for (pch = rgchBuf; *pch && isspace(*pch); pch++)
         ;
      if (*pch == '#')
         continue;

      for (pchWord = pch; *pch && !isspace(*pch) && *pch != '='; pch++)
         ;

      if (*pch != '=') {
         *pch = '\0';
         for (++pch; *pch && *pch != '='; pch++)
            ;
         if (*pch != '=') {
            syslog(LOG_ERR, "Badly formed stanza file for %s, line %d",
                   szName, line);
            free(pusrNew);
            return 0;
         }
      } else
         *pch = '\0';

      for (i = 0; rgszNames[i] && strcasecmp(rgszNames[i], pchWord); i++)
         ;

      if (rgszNames[i] == 0) {
         syslog(LOG_WARNING, "Discarding unknown stanza entry '%s' for %s",
                pchWord, szName);
         continue;
      }

      for (++pch; *pch && isspace(*pch); pch++)
         ;

      pchValue = pch;

      for (pch = pchValue + strlen(pchValue) - 1; isspace(*pch); pch--)
         ;
      if (pch < pchValue)
         pchValue = 0;
      else
         *++pch = '\0';

      switch (i) {
       case U_PWD:
         if (pchValue == 0)
            pusrNew->u_pwd = 0;
         else if ((pusrNew->u_pwd = strdup(pchValue)) == 0)
            syslog(LOG_ERR, "OUT OF MEMORY");
         else
            f_pwd++;
         break;

       case U_REAL:
         if (pchValue == 0)
            pusrNew->u_realname = 0;
         else if ((pusrNew->u_realname = strdup(pchValue)) == 0)
            syslog(LOG_ERR, "OUT OF MEMORY");
         else
            f_real++;
         break;

       case U_ORG:
         if (pchValue == 0)
            pusrNew->u_organization = 0;
         else if ((pusrNew->u_organization = strdup(pchValue)) == 0)
            syslog(LOG_ERR, "OUT OF MEMORY");
         else
            f_org++;
         break;

       case U_CTCT:
         if (pchValue == 0)
            pusrNew->u_contact = 0;
         else if ((pusrNew->u_contact = strdup(pchValue)) == 0)
            syslog(LOG_ERR, "OUT OF MEMORY");
         else
            f_ctct++;
         break;

       case U_SUCC:
         if (sscanf(pchValue, "%i", (int *)&pusrNew->u_lastsucc) != 1)
            syslog(LOG_ERR, "Bad lastsucc in stanza for %s", szName);
         else
            f_succ++;
         break;

       case U_UNSUCC:
         if (sscanf(pchValue, "%i", (int *)&pusrNew->u_lastunsucc) != 1)
            syslog(LOG_ERR, "Bad lastunsucc in stanza for %s", szName);
         else
            f_unsucc++;
         break;

       case U_EXP:
         if (sscanf(pchValue, "%i", (int *)&pusrNew->u_expires) != 1)
            syslog(LOG_ERR, "Bad expires in stanza for %s", szName);
         else
            f_exp++;
         break;

       default:
         syslog(LOG_ERR, "INTERNAL ERROR -- %s:%d", __FILE__, __LINE__);
         break;
      }
   }

   if (!f_pwd) {
      syslog(LOG_ERR, "Badly formed stanza file for %s", szName);
      return 0;
   }

   if (!f_real)
      pusrNew->u_realname = 0;

   if (!f_org)
      pusrNew->u_organization = 0;

   if (!f_ctct)
      pusrNew->u_contact = 0;

   if (!f_succ)
      pusrNew->u_lastsucc = 0;

   if (!f_unsucc)
      pusrNew->u_lastunsucc = 0;

   if (!f_exp)
      pusrNew->u_expires = 0;

   return pusrNew;
}


static char *lowerdup(char *src)
{
   char *dst;
   char *p;
   
   if ((dst = (char*) malloc(strlen(src) + 1)) != 0) {
      for (p = dst; *src; *p++ = tolower(*src++))
         ;
      *p = '\0';
   }

   return dst;
}


USR *pusrRead(char *szName)
{
   char           rgchPath[_MAX_PATH];
   int            fdUser;
   struct flock   flLock;
   int            currTimeout;
   sigset_t       stSave;
   sigset_t       stBlock;
   USR           *pusrNew;
   FILE          *fpIn;
   int            tries;
   char          *szFile;

   errno = 0;
   if (!szName || *szName == '\0') {
      syslog(LOG_ERR, "No name");
      errno = EINVAL;
      return 0;
   }

   if ((szFile = lowerdup(szName)) == 0) {
      syslog(LOG_ERR, "Out of memory");
      errno = ENOMEM;
      return 0;
   }
       
   sprintf(rgchPath, "%s%c/%s", accpath, *szFile, szFile);

   if ((fdUser = open(rgchPath, O_RDONLY)) < 0) {
      syslog(LOG_ERR, "open() failed, errno = %d", errno);
      return 0;
   }

   flLock.l_type = F_RDLCK;
   flLock.l_whence = SEEK_SET;
   flLock.l_start = (off_t) 0;
   flLock.l_len = (off_t) 0;


   /*
    * Block all blockable signals
    */
   sigfillset(&stBlock);
   sigdelset(&stBlock, SIGSEGV);
   sigdelset(&stBlock, SIGQUIT);
   sigprocmask(SIG_BLOCK, &stBlock, &stSave);

   for (tries = 0; tries < MAXTRIES; tries++) {
      if (fcntl(fdUser, F_SETLK, &flLock) < 0) {
         sleep(TRYWAIT);
      } else
         break;
   }

   if (tries < MAXTRIES) {
      if (fpIn = fdopen(fdUser, "r")) {
         pusrNew = pusrParse(szName, fpIn);
      } else {
         syslog(LOG_ERR, "fdopen() failed, errno = %d", errno);
         pusrNew = 0;
      }
   } else
      syslog(LOG_ERR, "Couldn't read lock file for %s", szName);

   sigprocmask(SIG_SETMASK,  &stSave, 0);

   flLock.l_type = F_UNLCK;
   fcntl(fdUser, F_SETLK, &flLock);

   if (fpIn)
      fclose(fpIn);
   else if (fdUser >= 0)
      close(fdUser);

   return pusrNew;
}

int WritePUsr(USR *pusrNew)
{
   char           rgchPath[_MAX_PATH];
   int            fdUser;
   struct flock   flLock;
   int            currTimeout;
   sigset_t       stSave;
   sigset_t       stBlock;
   FILE          *fpOut;
   int            tries;
   char          *szFile;

   errno = 0;

   if (!pusrNew || !pusrNew->u_name || *pusrNew->u_name == '\0') {
      syslog(LOG_ERR, "No name");
      return EINVAL;
   }

   if ((szFile = lowerdup(pusrNew->u_name)) == 0) {
      syslog(LOG_ERR, "Out of memory");
      errno = ENOMEM;
      return 0;
   }
       
   sprintf(rgchPath, "%s%c/%s", accpath, *szFile, szFile);

   if ((fdUser = open(rgchPath, O_WRONLY|O_TRUNC|O_CREAT, 0)) < 0) {
      syslog(LOG_ERR, "open() failed, errno = %d", errno);
      return EINVAL;
   }

   flLock.l_type = F_WRLCK;
   flLock.l_whence = SEEK_SET;
   flLock.l_start = (off_t) 0;
   flLock.l_len = (off_t) 0;


   /*
    * Block all blockable signals
    */
   sigfillset(&stBlock);
   sigdelset(&stBlock, SIGSEGV);
   sigdelset(&stBlock, SIGQUIT);
   sigprocmask(SIG_BLOCK, &stBlock, &stSave);

   for (tries = 0; tries < MAXTRIES; tries++) {
      if (fcntl(fdUser, F_SETLK, &flLock) < 0) {
         sleep(TRYWAIT);
      } else
         break;
   }

   if (tries < MAXTRIES) {
      if (fpOut = fdopen(fdUser, "w")) {
         fprintf(fpOut, "%s:\n", pusrNew->u_name);
         fprintf(fpOut, "\tpwd          = %s\n", pusrNew->u_pwd);
         if (pusrNew->u_realname)
            fprintf(fpOut, "\trealname     = %s\n", pusrNew->u_realname);
         if (pusrNew->u_organization)
            fprintf(fpOut, "\torganization = %s\n", pusrNew->u_organization);
         if (pusrNew->u_contact)
            fprintf(fpOut, "\tcontact      = %s\n", pusrNew->u_contact);
         if (pusrNew->u_lastsucc != 0)
            fprintf(fpOut, "\tlastsucc     = %i\n", pusrNew->u_lastsucc);
         if (pusrNew->u_lastunsucc != 0)
            fprintf(fpOut, "\tlastunsucc   = %i\n", pusrNew->u_lastunsucc);
         if (pusrNew->u_expires != 0)
            fprintf(fpOut, "\texpires      = %i\n", pusrNew->u_expires);
      } else {
         syslog(LOG_ERR, "fdopen() failed, errno = %d", errno);
      }
   } else {
      syslog(LOG_WARNING, "Couldn't write lock file for %s",
             pusrNew->u_name);
   }

   sigprocmask(SIG_SETMASK,  &stSave, 0);

   flLock.l_type = F_UNLCK;
   fcntl(fdUser, F_SETLK, &flLock);

   if (fpOut)
       fclose(fpOut);
   else if (fdUser >= 0)
      close(fdUser);

   return errno;
}

BOOL LogUser(char *name, char *password)
{
   USR     *pusrLog;
   time_t   ttNow;

   time(&ttNow);
   if ((pusrLog = pusrRead(name)) == 0) {
      syslog(LOG_WARNING, "Failed login for unknown user");
      return 0;
   }

   if (strcmp(crypt(password, pusrLog->u_pwd), pusrLog->u_pwd)) {
      /* failed */
      pusrLog->u_lastunsucc = ttNow;
      syslog(LOG_WARNING, "Failed login for %s", name);
      if (WritePUsr(pusrLog))
         syslog(LOG_ERR, "Couldn't update stanza file for %s", name);
      return 0;
   } else {
      if (pusrLog->u_expires && pusrLog->u_expires < ttNow) {
         syslog(LOG_WARNING, "Attempted login by expired user %s", name);
         return 0;
      }
      pusrLog->u_lastsucc = ttNow;
      syslog(LOG_NOTICE, "Successful login for %s", name);
      if (WritePUsr(pusrLog)) {
         syslog(LOG_ERR, "Couldn't update stanzafile for %s", name);
         return 0;
      }
      return 1;
  }

   return 0;                    /* just in case */
}

BOOL SetPassword(char *name, char *old, char *newn)
{
   USR  *pusrChange;
   char  salt[3];
   char  rgchSalt[] =
            "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789/.";


   if ((pusrChange = pusrRead(name)) == 0) {
      syslog(LOG_WARNING, "Attempt to change password for unknown user");
      return 0;
   }

   if (old != (char *)-1) {
      if (strcmp(crypt(old, pusrChange->u_pwd), pusrChange->u_pwd)) {
         syslog(LOG_WARNING, "Failed password change for %s", name);
         return 0;
      }
   }

   salt[0] = rgchSalt[rand() % 64];
   salt[1] = rgchSalt[rand() % 64];
   salt[3] = '\0';

   if ((pusrChange->u_pwd = strdup(crypt(newn, salt))) == 0) {
      syslog(LOG_ERR, "OUT OF MEMORY");
      return 0;
   }

   if (WritePUsr(pusrChange)) {
      syslog(LOG_ERR, "Couldn't update stanza file for %s", name);
      return 0;
   } else
      return 1;

   return 0;                    /* just in case */
}
