#include "Corpinc.h"
#ifdef _WIN32
#include "../XairaVersion.h"
#endif

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include    <xercesc/sax2/DefaultHandler.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/util/TranscodingException.hpp>
#include <unicode/brkiter.h>
#include <unicode/rbbi.h>
#include <unicode/uchar.h>
#include <unicode/unorm.h>
#include <unicode/coll.h>
#include <unicode/regex.h>
#include <unicode/ucnv.h>
#include <zlib.h>

using namespace XERCES_CPP_NAMESPACE;

enum param {PARAM_ETCPATH,PARAM_NAME,PARAM_TEXTPATH,PARAM_ROOTPATH,PARAM_INDEXPATH,PARAM_HEADER,
	PARAM_TEMPPATH,PARAM_LOG,PARAM_SOURCEPATH,PARAM_BIB,PARAM_VALIDATE,PARAM_GRAMMARCACHE,PARAM_MAX};

struct _tfl 
{
	const char* name;
	_tfl* next;
};


void SetParam(char* name,const xairo_uchar* value);
const char* exppath(param pa,const char* s);
const xairo_uchar* Param(param p); 
const char* ParamA(param p); 
void AddFile(const char* s);
void BuildFileArray();
struct stream* MakeMultipleStream(xara_form);
struct stream *MakeStream(int k);
void MakeEntityTable();
int LookupType(char*,char*);
int LookupTypeW(xairo_uchar*,xairo_uchar*);
xara_streambase* GetSB(int);
bool OpenSB();
bool AKTest(xara_streambase*);
void FakeDSC();
void GetLocale();
extern void MakeDTD();

int nfiles;
_tfl* tfl;
const char** files;
BreakIterator *brit;
char wblocale[1000];
char* wbrules=NULL;

extern "C" struct pihead idxhead;
extern "C" BOOL SetPos(struct stream *s);
extern "C" int CountDocuments();
extern "C" BOOL WAdStream(struct stream *, sara_ulong, WORDNO, int, BOOL);
extern "C" const char* WindowsTempPath();
extern "C" void MakeLocToDocTable();
extern "C" BOOL GetNElt(sara_ulong, const xairo_uchar *, xairo_uchar *,int,int);
extern "C" void Loc(int,struct xara_loc*);


char* Ascii(const XMLCh* w)
{
	size_t k=wcslen(w);
	char* s=new char [k+1];
	wcstombs(s,w,k+1);
	s[k]=0;
	return s;
}

class DHandle : public DOMErrorHandler {
public:
	DHandle():bErr(false) {}
	bool handleError(const DOMError&);
	bool bErr;
	const xairo_uchar* msg;
};

//DHandle eh;

int places, shift, hashsize;
int ncont, nlemma;
int wlsize;

xairo_uchar* param_vals[PARAM_MAX];

BOOL ReadParameters(const char* pfile)
{
	const char* ss=strrchr(pfile,SEPCHAR);
	if (ss==NULL) {
		bootlog("Cannot set root\n",pfile);
		return FALSE;
	}
	param_vals[PARAM_ROOTPATH]=new xairo_uchar [ss-pfile+2];
	mbstowcs(param_vals[PARAM_ROOTPATH],pfile,ss-pfile+1);
	param_vals[PARAM_ROOTPATH][ss-pfile+1]=0;

	FILE* f=fopen(pfile,"r");
	if (f==NULL) {
		bootlog("Parameter file %s does not exist\n",pfile);
		return FALSE;
	}
	fclose(f);
	DOMImplementationLS* di=(DOMImplementationLS*) DOMImplementation::getImplementation();
	DOMLSParser* d=di->createLSParser(DOMImplementationLS::MODE_SYNCHRONOUS,NULL);
	DOMConfiguration* cfg=d->getDomConfig();
	cfg->setParameter(XMLUni::fgDOMNamespaces,true);
    cfg->setParameter(XMLUni::fgXercesSchema,true);
	cfg->setParameter(XMLUni::fgDOMValidate, false);
	DHandle eh;
	cfg->setParameter(L("error-handler"),&eh);
	DOMDocument* doc=NULL;
	doc=d->parseURI(pfile);
	if (eh.bErr) {
		bootlog("Parse error in parameter file: %S",eh.msg);
		return FALSE;
	}
	DOMElement* r=doc->getDocumentElement();
	if (wcscmp(r->getTagName(),L("corpus_parameters")) != 0) {
		bootlog("Parameter file has wrong document element: %S\n",r->getTagName());		
		return FALSE;
	}
	DOMNodeList* n=doc->getElementsByTagName(L("parameter"));
	for (unsigned i=0;i<n->getLength();i++) {
		DOMNode* m=n->item(i);
		const XMLCh *x;
		const XMLCh *y;
		DOMNamedNodeMap* a=m->getAttributes();
		DOMNode* na=a->getNamedItem(L("name"));
		if (na==NULL) continue;
		x=na->getNodeValue();
		DOMNode* mm=m->getFirstChild();
		if (mm==NULL) continue;
		if (mm->getNodeType()!=DOMNode::TEXT_NODE) continue;
		y=mm->getNodeValue();
		SetParam(Ascii(x),y);
	}
	wcstombs(accpath,Param(PARAM_ROOTPATH),_MAX_PATH);
	strcat(accpath,SEPCHARS);
	wcstombs(etcpath,Param(PARAM_ETCPATH),_MAX_PATH);
	strcat(etcpath,SEPCHARS);
	wcstombs(idxpath,Param(PARAM_INDEXPATH),_MAX_PATH);
	strcat(idxpath,SEPCHARS);
	wcstombs(txtpath,Param(PARAM_TEXTPATH),_MAX_PATH);
	strcat(txtpath,SEPCHARS);
	wcstombs(corpname,Param(PARAM_NAME),_MAX_PATH);
	bSysLog=wcscmp(Param(PARAM_LOG),L("on"))==0;
	return TRUE;
}




char* param_names[PARAM_MAX]=
{"etcpath","name","textpath","rootpath","indexpath","headerpath","temppath","log","sourcepath","bibpath",
"validate","grammarcache"};

void SetParam(char* name,const xairo_uchar* value)
{
	int i;
	for (i=0;i<PARAM_MAX;i++)
	{
		if (strcmp(name,param_names[i])==0) break;
	}
	if (i>=PARAM_MAX) return;
	if (i==PARAM_ROOTPATH) return;
	param_vals[i]=_wcsdup(value);
	delete name;
}

bool DHandle::handleError(const DOMError& err)
{
	bErr=true;
	msg=err.getMessage();
	return false;
}

/*

void DHandle::fatalError(const SAXParseException& e)
{
	throw e;
}

void DHandle::error(const SAXParseException& e)
{
	throw e;
}

*/

BOOL ReadFileList()
{
	tfl=NULL; nfiles=0;
	DOMImplementationLS* di=(DOMImplementationLS*) DOMImplementation::getImplementation();
	DOMLSParser* d=di->createLSParser(DOMImplementationLS::MODE_SYNCHRONOUS,NULL);
	DOMConfiguration* cfg=d->getDomConfig();
	cfg->setParameter(XMLUni::fgDOMNamespaces,true);
    cfg->setParameter(XMLUni::fgXercesSchema,true);
	DHandle eh;
	cfg->setParameter(L("error-handler"),&eh);
	DOMDocument* doc=NULL;
	try
	{
		doc=d->parseURI(exppath(PARAM_ETCPATH,"file_index.xml"));
	}
	catch (SAXParseException& e)
	{
		bootlog("Parser error in file_index.xml: %S\n",e.getMessage());
		return FALSE;
	}
	DOMNodeList* n=doc->getElementsByTagName(L("file"));
	for (unsigned i=0;i<n->getLength();i++) {
		DOMNode* m=n->item(i);
		if (m==NULL) continue;
		DOMNode* m1=m->getFirstChild();
		if (m1==NULL) continue;
		if (m1->getNodeType() != DOMNode::TEXT_NODE) continue;
		const XMLCh* s=m1->getNodeValue();
		AddFile(Ascii(s));
	}
	BuildFileArray();
	ntexts=idxhead.n=nfiles+1;
	if (nfiles==0) {
		bootlog("No texts in corpus\n");
		return FALSE;
	}
	return TRUE;
}

void BuildFileArray()
{
	files=new const char* [nfiles];
	_tfl* t=tfl;
	for (int i=0;i<nfiles;i++) {
		files[i]=t->name;
		t=t->next;
	}
}

extern "C" const char* GetTextPath(int i)
{
	if (i==0) return ParamA(PARAM_HEADER);
	else if (i==1) return ParamA(PARAM_BIB);
	struct stream *s0=MakeAtomStream("path",0,it_global_attribute_value);
	s0->cursor=i;
	SetPos(s0);
	const char* path=GetWordList(s0->wpos);
	FreeStream(s0);
	return exppath(PARAM_TEXTPATH,path);

}

void GetLocale()
{
	strcpy(wblocale,"#def");
	struct stream* s=MakeAtomStream("locale",0,it_global_attribute_value);
	if (s) {
		s->cursor=0;
		if (SetPos(s)) {
			strcpy(wblocale,GetWordList(s->wpos));
		}
		FreeStream(s);
	}
	wbrules=NULL;
	s=MakeAtomStream("wbrules",0,it_global_attribute_value);
	if (s) {
		s->cursor=0;
		if (SetPos(s)) {
			wbrules=_strdup(GetWordList(s->wpos));
		}
		FreeStream(s);
	}
}





void AddFile(const char* s)
{
	_tfl** t=&tfl;	
	while (*t) t=&(*t)->next;
	*t=new _tfl;
	(*t)->name=_strdup(s);
	(*t)->next=NULL;
	nfiles++;
}

#define ISSEPCHAR(c) (c=='/'||c=='\\')

const char* exppath(param pa,const char* s)
{
	static char buffer[_MAX_PATH];
	const xairo_uchar* p=Param(pa);
	size_t k=wcslen(p);
	wcstombs(buffer,p,k);
	buffer[k]=0;
	if (k > 0 && !ISSEPCHAR(buffer[k-1])) 
		strcat(buffer,"/");
	strcat(buffer,s);
	return buffer;
}



const xairo_uchar* Param0(param p) 
{
	return param_vals[p]?param_vals[p]:L("");
}

const xairo_uchar* Param(param p) 
{
	const xairo_uchar* s=Param0(p);
	if ((wcslen(s) > 1 && s[1]==':') || p==PARAM_NAME || p==PARAM_TEMPPATH)
		return s;
	xairo_uchar* ss=new xairo_uchar [_MAX_PATH];
	wcscpy(ss,Param0(PARAM_ROOTPATH));
	wcscat(ss,s);
	return ss;
}

const char* ParamA(param p) 
{
	static char pp[_MAX_PATH];
	wcstombs(pp,Param(p),_MAX_PATH);
	return pp;
}

#define MAXNF 100
#define MAXTAX 100

struct _catab {
	char name[MAXLTNAME];
	int idx;
	int nf;
	int fil[MAXNF];
	xairo_uchar def[MAXWORD];
};

_catab conttab[MAXNAK];

struct _tax {
	xairo_uchar name[MAXNAME];
	int nc;
};

_tax taxtab[MAXTAX];

static int ntax=0;

static int nctab0=0;

xara_wl *wl;
int *hi;

BOOL LoadWordList()
{
	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbwl"),"rb");
	if (f==NULL) {
		bootlog("The file xdbwl is missing\n");
		return FALSE;
	}
	xara_wl_header h; 
	fread(&h,sizeof h,1,f);
	if (h.sig != DBSIG || h.version != DBVER) {
		fclose(f);
		bootlog("Bad index version %d (%d required)\n",h.version,DBVER);
		return FALSE;
	}
	hashsize=h.hashsize;
	wlsize=h.size;
	places=h.places;
	shift=h.shift;
	ncont=h.cont;
	ntax=h.ntax;
	nlemma=-1; // for now
	int* t=new int[h.size];
	xara_wl_header_addkey* cont = new xara_wl_header_addkey [ncont];
	fread(cont,sizeof(xara_wl_header_addkey),ncont,f);
	xara_wl_header_tax* tax=new xara_wl_header_tax [ntax];
	fread(tax,sizeof(xara_wl_header_tax),ntax,f);
	wl=new xara_wl [h.size];
	hi=new int[h.hashsize];
	fread(hi,sizeof(int),h.hashsize,f);
	int i;
	for (i=0;i<h.size;i++) {
		fread(&t[i],sizeof(int),1,f);
		fread(&wl[i].next,sizeof(int),1,f);
		fread(&wl[i].dicptr,sizeof(int),1,f);
	}
	for (i=0;i<h.size;i++) {
		xairo_uchar* x=new xairo_uchar [t[i]+1];
		fread(x,2,t[i],f);
		x[t[i]]=0;
		wl[i].word=x;
	}
	delete [] t;
	fclose(f);
	int nl=1;
	strcpy(lemmatab[0].name,"null");
	lemmatab[0].idx=-1;
	char info[MAXLTNAME];
	sprintf(info,"%d = ",ncont);
	for (i=0;i<ncont;i++) {
		strcat(info,GetWordList(cont[i].name));
		strcat(info,"=");
	}
	strcpy(lemmatab[0].info,info);
   	for (i=0;i<ncont;i++) {
		if (!cont[i].bLemma) continue;
		strcpy(lemmatab[nl].name,GetWordList(cont[i].name));
		lemmatab[nl].idx=i;
		strcpy(lemmatab[nl].info,info);
		nl++;
	}
	int nc=0;
   	for (i=0;i<ncont;i++) {
		if (cont[i].bLemma) continue;
		strcpy(conttab[nc].name,GetWordList(cont[i].name));
		conttab[nc].idx=i;
		conttab[nc].nf=0;
		wcscpy(conttab[nc].def,GetWordListW(cont[i].def));
		nc++;
	}
   	for (i=0;i<ntax;i++) {
		wcscpy(taxtab[i].name,GetWordListW(tax[i].name));
		taxtab[i].nc=tax[i].nc;
	}

	nlemmatab=nl;
	nctab0=nc;
	return TRUE;
}

extern "C" const xairo_uchar* GetContName(int i)
{
	static xairo_uchar b[100]; int j;
	for (j=0;j<nctab0;j++) 
		if (conttab[j].idx==i) break;
	mbstowcs(b,conttab[j].name,100);
	return b;
}

static char buffer[10000];

// until protocol conversion
xairo_uchar* ISOToUnicode(const char* c)
{
	xairo_uchar* b=(xairo_uchar*) buffer; 
	while (*c) {
		if (*c != '&') {
			*b=(unsigned char) *c;
			b++; 
			c++;
		}
		else 
		{
			int n=0; int nn=0;
			int x=sscanf(c,"&#%d;%n",&n,&nn);
			if (x) {
				*b=n;
				b++;
				c+=nn;
			} else {
				*b='&';
				b++;
				c++;
			}
		}
	}
	*b=0;
	return (xairo_uchar*) buffer;
}

char* UnicodeToISO(xairo_uchar* w)
{
	char* s=buffer;
	memset(buffer,0,sizeof buffer);
	while (*w) {
		if ((*w) < 256) {
			*s=(char) *w;
			s++;
		}
		else {
			char t[10];
			sprintf(t,"&#%d;",*w);
			strcat(s,t);
			s+=strlen(t);
		}
		w++;
	}
	*s=0;
	return buffer;
}

char* GetWordList(int n)
{
	return UnicodeToISO(wl[n].word);
}

xairo_uchar* GetWordListW(int n)
{
	return wl[n].word;
}

struct xara_lemma *ll=NULL;
int *llr=NULL;

struct xara_type *tl=NULL;

int ktl=0;

extern "C" BOOL LoadTypeList()
{
	const char* path=exppath(PARAM_INDEXPATH,"xdbtype");
	FILE* f=fopen(path,"rb");
	if (f==NULL){
		bootlog("The file xdbtype is missing\n");
		return FALSE;
	}
	fseek(f,0,SEEK_END);
	int k=ftell(f);
	fseek(f,0,SEEK_SET);
	ktl=k/sizeof (struct xara_type);
	tl=new xara_type [k];
	fread(tl,sizeof (struct xara_type),ktl,f);
	fclose(f);
	return TRUE;
}


extern "C" BOOL LoadLemmaList()
{
	char nm[20];
	sprintf(nm,"xdblemma%d",nlemma+1);
	const char* path=exppath(PARAM_INDEXPATH,nm);
	FILE* f=fopen(path,"rb");
	if (f==NULL) {
		bootlog("The file %s is missing\n",nm);
		return FALSE;
	}
	fread(&dicthead,sizeof dicthead,1,f);
	if (ll) delete [] ll;
	if (llr) delete [] llr;
	ll=new xara_lemma [dicthead.ndix];
	llr=new int [wlsize];
	for (int i=0;i<wlsize;i++) llr[i]=-1;
	for (int i=0;i<dicthead.ndix;i++) {
		fread(&ll[i],sizeof (int),3,f);
		fread(&ll[i].word[0],sizeof (int),dicthead.nfacets,f);
		llr[ll[i].word[0]]=i;
		int j=ll[i].forms;
		ll[i].pt=new xara_form[j];
		fread(ll[i].pt,sizeof(xara_form),j,f);
	}
	fclose(f);
	return TRUE;
}


int hash(const xairo_uchar *w)
{
	unsigned int k=0;
//	int n=w.length();
//	if (n > m_places) n=m_places;
//	for (int i=0;i<n;i++) {
	for (int i=0;i<places;i++) {
		if (w[i]==0) break; // new
//		k <<= m_shift;
//		k+=w[i];
		k+=(w[i] << shift*i);
	}
	return k % hashsize;
}

xara_loc_hdr loc_hdr, loc_hdr_t;
xara_loc_page* loctab, *threadtab;
int *thpa=NULL;

extern "C" void ReadPage(int n,char* b)
{
	int off=loctab[n].off;
	int len=loctab[n].len;
	int nf=loctab[n].fi;
	unsigned char* c=new unsigned char[len];
	fseek(fLocs[nf],off,SEEK_SET);
	fread(c,1,len,fLocs[nf]);
	unsigned long d=loc_hdr.ps*sizeof(xara_loc);
	uncompress((unsigned char*) b,&d,c,len);
	delete [] c;
}

Collator* theCollator;

extern "C" BOOL NewStartup(const char* f,BOOL bLog)
{
    XMLPlatformUtils::Initialize();
	if (!ReadParameters(f)) return FALSE;
	opensyslog("sara", 0, 0,bLog);
//	if(!ReadFileList()) return FALSE;
//	syslog(LOG_NOTICE,"Read file list");
	ntexts=idxhead.n;
#ifdef _WIN32
	syslog(LOG_NOTICE,"Xaira-server version %d.%d.%d.%d built %s",XV1,XV2,XV3,XV4,__DATE__);
#else
	syslog(LOG_NOTICE,"Xaira-server version " VERSION " built " __DATE__);
#endif
	if (!LoadWordList()) return FALSE;
	syslog(LOG_NOTICE,"Loaded word list");
	if (!LoadLemmaList()) return FALSE;
	if (!LoadTypeList()) return FALSE;
	syslog(LOG_NOTICE,"Loaded lemma list");
//	fStreamBase=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	if (fStreamBase==NULL) return FALSE;
	if (!OpenSB()) return FALSE;
	FILE *fL=fopen(exppath(PARAM_INDEXPATH,"xdblocs"),"rb");
	if (fL==NULL) 
	{
		bootlog("Locs file 0 missing");
		return FALSE;
	}
	fread(&loc_hdr,sizeof(loc_hdr),1,fL);
	fLocs=new FILE* [loc_hdr.nf];
	fLocs[0]=fL;
	for (int i=1;i<loc_hdr.nf;i++) {
		char p[_MAX_PATH];
		sprintf(p,"xdblocs%d",i);
		fLocs[i]=fopen(exppath(PARAM_INDEXPATH,p),"rb");
		if (fLocs[i]==NULL) {
			bootlog("Locs file %d missing",i);
			return FALSE;
		}
	}
	loctab=new xara_loc_page[loc_hdr.np];
	fread(loctab,sizeof(xara_loc_page),loc_hdr.np,fL);
	maxiloc=loc_hdr.ps;
	npages=loc_hdr.np;
	syslog(LOG_NOTICE,"Loaded page table");

	FILE *fT=fopen(exppath(PARAM_INDEXPATH,"xdbthread"),"rb");
	if (fT==NULL) 
	{
		bootlog("Thread file 0 missing");
		return FALSE;
	}
	fread(&loc_hdr_t,sizeof(loc_hdr_t),1,fT);
	fThread=new FILE* [loc_hdr_t.nf];
	fThread[0]=fT;
	for (int i=1;i<loc_hdr_t.nf;i++) {
		char p[_MAX_PATH];
		sprintf(p,"xdbthread%d",i);
		fThread[i]=fopen(exppath(PARAM_INDEXPATH,p),"rb");
		if (fThread[i]==NULL) {
			bootlog("Thread file %d missing",i);
			return FALSE;
		}
	}
	threadtab=new xara_loc_page[loc_hdr_t.np];
	fread(threadtab,sizeof(xara_loc_page),loc_hdr_t.np,fT);
	maxiloct=loc_hdr_t.ps;
	npagest=loc_hdr_t.np;
	syslog(LOG_NOTICE,"Loaded thread page table");
	thpa=new int[maxiloct];
	FILE* dscf=fopen(exppath(PARAM_INDEXPATH,"dsc.txt"),"r");
	if (dscf==NULL) {
		bootlog("DSC file missing");
		return FALSE;
	}
	fclose(dscf);
	int k=AllocBuffers(0);
	syslog(LOG_NOTICE,"Allocated %d buffers (%d requested)",k,0);
	if (k==0) {
		bootlog("Insufficinet memory for buffering");
		return FALSE;
	}
	GetLocale();
	UErrorCode e=U_ZERO_ERROR;
	brit=NULL;
	if (wbrules) {
		UParseError p;
		brit=new RuleBasedBreakIterator (wbrules,p,e);
		if (e != U_ZERO_ERROR) {
			syslog(LOG_WARNING,"Failed to parse rbbi rules");
			brit=NULL;
		}
	}
	e=U_ZERO_ERROR;
	char* lx=NULL;
	if (strcmp(wblocale,"#def") != 0) lx=wblocale;
	Locale ll(lx);
	if (brit==NULL) brit=BreakIterator::createWordInstance(ll,e);
	assert(U_SUCCESS(e));
	theCollator = Collator::createInstance(ll,e);
	theCollator->setStrength(Collator::PRIMARY);
	assert(U_SUCCESS(e));
	if (!U_SUCCESS(e)) {
		bootlog("Failed to initialise locale %s (ICU err=%d)",wblocale,e);
		return FALSE;
	}
	syslog(LOG_NOTICE,"Created locale %s",wblocale);
	MakeEntityTable();
	syslog(LOG_NOTICE,"Made entity table");
	MakeLocToDocTable();
	syslog(LOG_NOTICE,"Made doc table");
	ntexts=CountDocuments();
	syslog(LOG_DEBUG,"nctab=%d",nctab0);
	MakeDTD();
//	FakeDSC();
	return TRUE;
}


extern "C" void GetDictEntry(FILE *f, sara_long k, xairo_uchar *wd, int maxwd, sara_long *freq, sara_long *forms)
{
	xara_lemma *xl=&ll[k];

	wcsncpy(wd,GetWordListW(xl->word[0]),maxwd);
	*freq=xl->freq;
	*forms=xl->forms;
}	

extern "C" int Lookup(const char* s)
{
	xairo_uchar* w=ISOToUnicode(s);
	int h=hash(w);
	int k=hi[h];
	while (k != -1) {
		xairo_uchar *ww=wl[k].word;
		if (wcscmp(w,ww)==0) return k;
		k=wl[k].next;
	}
	return -1;
}

extern "C" int LookupW(const xairo_uchar* s)
{
	int h=hash(s);
	int k=hi[h];
	while (k != -1) {
		xairo_uchar *ww=wl[k].word;
		if (wcscmp(s,ww)==0) return k;
		k=wl[k].next;
	}
	return -1;
}

extern "C" BOOL LookUpWord(struct lookup *lur, int kf)
{
	int d=LookupW(lur->word);
	if (d==-1) return FALSE;;
	int dd=wl[d].dicptr;
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	fseek(fStreamBase,dd*(sizeof(xara_streambase)+ncont*sizeof(int)),SEEK_SET);
	int i=0; int last=-1;
	while (true) {
		xara_streambase *sb=GetSB(dd+i);
//		fread(&sb,sizeof sb,1,fStreamBase);
		if (sb->w.val!=d) break;
		if (!AKTest(sb)) {i++; continue;}
//		sb.w.con=new int [ncont];
//		fread(sb.w.con,sizeof (int),ncont,fStreamBase);
		wcscpy(lur->sig[i].spell,GetWordListW(sb->w.xval));
		lur->sig[i].type=sb->w.t;
		lur->sig[i].extra=0;
		lur->sig[i].first=sb->first % maxiloc;
		lur->sig[i].ff=sb->first / maxiloc;
		lur->sig[i].acc=sb->acc-1;
		last=sb->first+sb->count;
		i++;
	}
//	fclose(f);
	if (last==-1) {
		return FALSE;
	}
	lur->nsig=i;
	lur->last = last % maxiloc;
	lur->lf = last / maxiloc;
	return TRUE;
}



extern "C" void GetDocPath(sara_ulong k, char *name, char *path)
{
	if (k==0) {
		wcstombs(path,Param(PARAM_HEADER),_MAX_PATH);
		strcpy(name,"header");
	} else if (k==1) {
		wcstombs(path,Param(PARAM_BIB),_MAX_PATH);
		strcpy(name,"bibliography");
	}
	else {
		strcpy(path,GetTextPath(k));
		struct stream *s=MakeAtomStream("document",0,it_global_attribute_value);
		s->cursor=k-2;
		SetPos(s);
		strcpy(name,GetWordList(s->wpos));
		FreeStream(s);
	}
}

extern "C" void GetDocName(sara_ulong k, char *name)
{
	struct stream *s=MakeAtomStream("document",0,it_global_attribute_value);
	s->cursor=k;
	SetPos(s);
	strcpy(name,GetWordList(s->wpos));
	FreeStream(s);
}

extern "C" size_t GetDocBib(sara_ulong k, xairo_uchar *w,int km)
{
	GetNElt(1,L("bibliography"),w,km,k-2);
	return wcslen(w);
/*	int b1, b2; 
	if (k <= 0) return -1;
	k--;
	struct stream *s=MakeAtomStream("BIBSTART",0,it_global_attribute_value);
	if (s==NULL) return -1;
	s->cursor=k;
	if (!SetPos(s)) return -1;
	b1=atoi(GetWordList(s->wpos));
	FreeStream(s);
	s=MakeAtomStream("BIBLENGTH",0,it_global_attribute_value);
	s->cursor=k;
	if (!SetPos(s)) return -1;
	b2=atoi(GetWordList(s->wpos));
	FreeStream(s);
	int kk=b2+1;
	if (b2 > km-1) b2=km-1;
	if (b2 < 0) return kk;
	FILE *f=_wfopen(Param(PARAM_BIB),L("rb"));
	if (f==NULL) return -1;
	fseek(f,b1*2,SEEK_SET);
	fread(w,2,b2,f);
	w[b2]=0;
	fclose(f);
	return kk; */
}





extern "C" int FindDoc(char *name)
{
	struct stream *s=MakeAtomStream("document",0,it_global_attribute_value);
	int i;
	for (i=0;i<CountDocuments();i++) {
		s->cursor=i;
		SetPos(s);
		if (strcmp(name,GetWordList(s->wpos))==0) break;
	}
	FreeStream(s);
	return i;
}
		
extern "C" struct stream *MakeAllPosStream(const xairo_uchar* pos,const xairo_uchar* key)
{
	int k=LookupW(pos);
	if (k==-1) return NULL;
	int dd=wl[k].dicptr;
	int x=LookupW(key);
	if (x==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=GetSB(dd+i);
		if (sb->w.val != k) break;
		if (sb->w.t != it_addkey) continue;
		if (sb->w.name != x) continue;
		struct stream* s=MakeStream(dd+i);
		if (s==NULL) return NULL;
		s->wc=1;
		return s;
	}
	return NULL;
}

extern "C" struct stream *MakePosStream3W(const xairo_uchar *w, const xairo_uchar* pos,const xairo_uchar* key)
{
	int id=-1;
	int k=LookupW(w);
	if (k==-1) return NULL;
	char cn[MAXLTNAME];
	wcstombs(cn,key,MAXLTNAME);
	for (int i=0;i<nctab0;i++) {
		if (strcmp(cn,conttab[i].name)==0)
			id=conttab[i].idx;
	}
	if (id==-1) return NULL;
	int dd=wl[k].dicptr;
	int x=LookupW(pos);
	if (x==-1) return NULL;
	struct stream* s=GrabStream();
	s->type=TY_OR;
	s->wc=1;
	for (int i=0;;i++) {
		xara_streambase* sb=GetSB(dd+i);
		if (sb->w.val != k) break;
		if (sb->w.con[id] != x) continue;
		if (!AKTest(sb)) continue;
		struct stream* ss=MakeStream(dd+i);
		if (ss==NULL) continue;
		s->op[s->ops++]=ss;
	}
	if (s->ops==0) {
		FreeStream(s);
		return NULL;
	}
	return s;
}






extern "C" struct stream *MakeLemmaStream(const xairo_uchar* word)
{
	int k=LookupW(word);
	if (k==-1) return NULL;
//	for (int i=0;i<dicthead.ndix;i++) // wow again - stop doing this!
//		if (ll[i].word==k) break;
//	if (i==dicthead.ndix) return NULL;
	int i=llr[k];
	if (i==-1) return NULL;
	if (ll[i].forms==1) return MakeMultipleStream(ll[i].pt[0]);
	struct stream* s = GrabStream();
	s->type=TY_OR;
	s->ops=0;
	s->wc=1;
	for (int j=0;j<ll[i].forms;j++) {
		struct stream* ss=MakeMultipleStream(ll[i].pt[j]);
		if (ss) s->op[s->ops++]=ss;
	}
	return s;
}

struct stream* MakeMultipleStream(xara_form x)
{
	if (x.count==1) return MakeStream(x.ptr);
	struct stream *s = GrabStream();
	if (s==NULL) return NULL;
	s->type = TY_OR;
	s->ops = 0; // x.count;
	s->wc = 1;
	for (int j = 0; j < x.count; j++)
	{
		struct stream* ss=MakeStream(x.ptr+j);
		if (ss) s->op[s->ops++] = ss;
	}
	return s;
}


struct stream *MakeStream(int k)
{
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	fseek(fStreamBase,k*(sizeof(xara_streambase)+ncont*sizeof(int)),SEEK_SET);
	xara_streambase *sb=GetSB(k);
	if (!AKTest(sb)) return NULL;
//	fread(&sb,sizeof sb,1,fStreamBase);
//	fclose(f);
	struct stream *s1=GrabStream();
	if (s1==NULL) return NULL;
	wcscpy(s1->word, wl[sb->w.val].word);
	wcscpy(s1->xword, wl[sb->w.val].word);
	s1->type = TY_ATOM;
//	s1->ihash = sb.ihash;
//	s1->extra = sb.extra;
	s1->wtype = sb->w.t;
	s1->first = sb->first % maxiloc;
	s1->ff = sb->first / maxiloc;
	int l=sb->first+sb->count;
	s1->last = l % maxiloc;
	s1->lf = l / maxiloc;
	s1->count = -1;
	s1->wc = 1;
	s1->lw = wcslen(s1->word);
	s1->lw0=0;
	return s1;
}	

extern "C" BOOL Form(xairo_uchar* headword,int k,xairo_uchar* form,int* freq)
{
	int kk=LookupW(headword);
	if (kk==-1) return FALSE;
	int i=llr[kk];
//	for (int i=0;i<dicthead.ndix;i++)
//		if (ll[i].word==kk) break;
	if (i==-1) return NULL;
	if (k>=ll[i].forms) return FALSE;
	int p=ll[i].pt[k].ptr;
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	fseek(fStreamBase,p*(sizeof(xara_streambase)+ncont*sizeof(int)),SEEK_SET);
	xara_streambase *sb=GetSB(p);
//	fread(&sb,sizeof sb,1,fStreamBase);
//	fclose(f);
	wcscpy(form, wl[sb->w.val].word);
	if (!AKTest(sb)) {
		*freq=0;
		return TRUE;
	}
	*freq=ll[i].pt[k].freq;
	return TRUE;
}

#define MAXFORMS 1000

//int nf=0;
//xara_form* xb[MAXFORMS];

xara_form* Form(xairo_uchar* headword, int k);

extern "C" BOOL SlowForm(xairo_uchar* headword,int k,xairo_uchar* form,int* freq)
{
	xara_form* f=Form(headword,k);
	if (f==NULL) return FALSE;
	int p=f->ptr;
	xara_streambase* sb=GetSB(p);
	wcscpy(form, wl[sb->w.val].word);
	struct stream* s=MakeMultipleStream(*f);
	if (s==NULL) *freq=0;
	else {
		*freq=XCountStream(s);
		FreeStream(s);
	}
	return TRUE;
}

extern "C" int QuickForms(xairo_uchar* headword)
{
	int k=0;
	int kk=LookupW(headword);
	if (kk==-1) return 0;
	int i=llr[kk];
	if (i==-1) return 0;
	for (int j=0;j<ll[i].forms;j++) {
		int p=ll[i].pt[j].ptr;
		struct stream* s=MakeMultipleStream(ll[i].pt[j]);
		if (s==NULL) continue;
		BOOL b=XAdvanceStream(s);
		FreeStream(s);
		if (b) k++;
	}
	return k;
}

xara_form* Form(xairo_uchar* headword, int k)
{
	int kk=LookupW(headword);
	if (kk==-1) return 0;
	int i=llr[kk];
	if (i==-1) return 0;
	for (int j=0;j<ll[i].forms;j++) {
		int p=ll[i].pt[j].ptr;
		struct stream* s=MakeMultipleStream(ll[i].pt[j]);
		BOOL b=XAdvanceStream(s);
		FreeStreamR(s);
		if (b) {
			if (k==0) return &ll[i].pt[j];
			else k--;
		}
	}
	return NULL;
}



	


int GetPosN(const xairo_uchar* word,int k1)
{
	int kk=LookupW(word);
	if (kk==-1) return 0;
	int dd=wl[kk].dicptr;
	int k=0;
	for (int i=0;;i++) {
		xara_streambase* sb=GetSB(dd+i);
		if (sb->w.val != kk) break;
		if (sb->w.t != it_content) continue;
		if (!AKTest(sb)) continue;
		k++;
	}
	return k;
}

// NB below code returns duplicates

const xairo_uchar* GetPosS(const xairo_uchar* word,int k1,int k)
{
	int kk=LookupW(word);
	if (kk==-1) return L("");
	int dd=wl[kk].dicptr;
	for (int i=0;;i++) {
		xara_streambase* sb=GetSB(dd+i);
		if (sb->w.t != it_content) continue;
		if (!AKTest(sb)) continue;
		if (k==0) {
			int w=sb->w.con[k1];
			return wl[w].word;
		}
		k--;
	}
	return L("");
}

extern "C" int GetPosA(const xairo_uchar* word,int k1,const xairo_uchar** b,int n)
{
	int m=GetPosN(word,k1); int t=0;
	for (int i=0;i<m;i++) {
		const xairo_uchar* p=GetPosS(word,k1,i);
		int j;
		for (j=0;j<t;j++) {
			if (wcscmp(p,b[j])==0) break;
		}
		if (j==t && t < n) {
			b[t]=p;
			t++;
		}
	}
	return t;
}


extern "C" BOOL SubForm(xairo_uchar* headword,int k,int k1, xairo_uchar* form) // better to merge this in
{
	if (k1 >= ncont) return FALSE;
	int kk=LookupW(headword);
	if (kk==-1) return FALSE;
//	for (int i=0;i<dicthead.ndix;i++)
//		if (ll[i].word==kk) break;
	int i=llr[kk];
	if (i==-1) return FALSE;
	if (k>=ll[i].forms) return FALSE;
	int p=ll[i].pt[k].ptr;
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	fseek(fStreamBase,p*(sizeof(xara_streambase)+ncont*sizeof(int)),SEEK_SET);
	xara_streambase *sb=GetSB(p);
//	fread(&sb,sizeof sb,1,fStreamBase);
//	sb.w.con=new int [ncont];
//	fread(sb.w.con,sizeof(int),ncont,fStreamBase);
//	fclose(f);
	wcscpy(form, wl[sb->w.con[k1]].word);
	return TRUE;
}

extern "C" BOOL SlowSubForm(xairo_uchar* headword,int k,int k1, xairo_uchar* form) // better to merge this in
{
	xara_form* f=Form(headword,k);
	if (f==NULL) return FALSE;
	int p=f->ptr;
	xara_streambase *sb=GetSB(p);
	wcscpy(form, wl[sb->w.con[k1]].word);
	return TRUE;
}

extern "C" struct stream *MakeAttStream(const xairo_uchar *elt, const xairo_uchar *attname, xairo_uchar *val, typ it)
{
	int d=LookupW(val);
	if (d==-1) return FALSE;;
	int dd=wl[d].dicptr;
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
	fseek(fStreamBase,dd*(sizeof(xara_streambase)+ncont*sizeof(int)),SEEK_SET);
	struct stream* s1=NULL; int i=0;
	while (true) {
		xara_streambase* sb=GetSB(dd+i++);
//		fread(&sb,sizeof sb,1,fStreamBase);
		if (sb->w.val!=d) break;
		if (!AKTest(sb)) continue;
//		sb.w.con=new int [ncont];
//		fread(sb.w.con,sizeof (int),ncont,fStreamBase);

		if (sb->w.t != it) continue;
	    xairo_uchar* e=GetWordListW(sb->w.element);
		if (wcscmp(elt,e) != 0) continue;
		xairo_uchar* a=GetWordListW(sb->w.name);
		if (wcscmp(attname,a)==0) {
			s1=GrabStream();
			if (s1==NULL) return NULL;
			wcscpy(s1->word,val);
			wcscpy(s1->xword,val);
			s1->type = TY_ATOM;
			s1->wtype = sb->w.t;
			s1->first = sb->first % maxiloc;
			s1->ff = sb->first / maxiloc;
			int l=sb->first+sb->count;
			s1->last = l % maxiloc;
			s1->lf = l / maxiloc;
			s1->count = -1;
			s1->wc = 1;
			s1->lw = wcslen(s1->word);
			s1->lw0=0;
			s1->acc=sb->acc-1;
			break;
		}
	}
	return s1;
}
			
extern "C" struct stream *MakeVarStream(xairo_uchar* elt,xairo_uchar* att,xairo_uchar* val,typ it)
{
	struct stream* s0=MakeAttStream(elt,att,att,it);
	struct stream *s;
	if (s0==NULL) return NULL;
	s = GrabStream();
	s->ops = 1;
	s->op[0]=s0;
	s->type = TY_ID;
	s->kvar=_wtoi(val);
	return s;
}

int cl1(const void* l1,const void* l2)
{
	struct locent1* ll1=(struct locent1*) l1;
	struct locent1* ll2=(struct locent1*) l2;
	return CompareWord(ll1->lcode,ll1->word,ll2->lcode,ll2->word);
}

extern "C" int MergeWords(struct regexpnew* re,xairo_uchar* pre,char* path)
{
	int x1, x2;
	MkLC(pre);
	MkNorm(pre);
   if (!GetLims(fDict, 0, dicthead.ndix - 1,pre, &x1, &x2)) return 0;
   int p=0;
	for (int i=x1;i<=x2;i++) {
		int nf=ll[i].forms;
		for (int j=0;j<nf;j++) {
			xara_form f=ll[i].pt[j];
			for (int k=0;k<f.count;k++) {
				int x=f.ptr+k;
				xara_streambase *sb=GetSB(x);
				if (sb->w.t != it_content) continue;
				if (!pMatch(GetWordListW(sb->w.xval),re)) continue;
				p+=sb->count;
			}
		}
	}
	struct locent1* b=new struct locent1 [p];
	if (b==NULL) return 0;
	int pp=0;
	for (int i=x1;i<=x2;i++) {
		int nf=ll[i].forms;
		for (int j=0;j<nf;j++) {
			xara_form f=ll[i].pt[j];
			for (int k=0;k<f.count;k++) {
				int x=f.ptr+k;
				xara_streambase *sb=GetSB(x);
				if (sb->w.t != it_content) continue;
				xairo_uchar* wx=GetWordListW(sb->w.xval);
				if (!pMatch(wx,re)) continue;
				for (int x=0;x<sb->count;x++) {
					xara_loc l;
					Loc(x+sb->first,&l);
					b[pp].lcode=l.fi;
					b[pp].lw=l.len; 
					b[pp].lw0=0;
//					b[pp].nxt=l.nxt;
					b[pp].pos=l.off;
//					b[pp].pre=l.pre;
					b[pp].sb=l.sb;
					b[pp].word=l.nw;
					b[pp].wc=1;
					pp++;
				}
			}
		}
	}
	qsort(b,p,sizeof (struct locent1),cl1);
	FILE* f=fopen(path,"wb");
	fwrite(b,sizeof (struct locent1),p,f);
	fclose(f);
	return p;
}

extern "C" int MergeWordsA(struct regexpnew *re,char* path,xairo_uchar* att,xairo_uchar* elt,BOOL bG)
{
	xara_streambase *sb;
	int p=LookupTypeW(elt,att); int p0=p;
	int x=sizeof(sb)+ncont*sizeof(int); int pa=0;
	while (p != -1) {
		sb=GetSB(p);
		if (!AKTest(sb)) continue;
		p=sb->next;
		if (wcscmp(GetWordListW(sb->w.name),att)!=0) continue;
		if (bG) {
			if (sb->w.t != it_global_attribute) continue;
		} else {
			if (sb->w.t != it_local_attribute) continue;
			if (wcscmp(GetWordListW(sb->w.element),elt)!=0) continue;
		}
		xairo_uchar* s=GetWordListW(sb->w.xval);
		if (!pMatch(s,re)) continue;
		pa+=sb->count;
	}
	struct locent1* b=new struct locent1 [pa];
	if (b==NULL) return 0;
	int ppa=0; p=p0;
	while (p != -1) {
		sb=GetSB(p);
		if (!AKTest(sb)) continue;
		p=sb->next;
		if (wcscmp(GetWordListW(sb->w.name),att)!=0) continue;
		if (bG) {
			if (sb->w.t != it_global_attribute) continue;
		} else {
			if (sb->w.t != it_local_attribute) continue;
			if (wcscmp(GetWordListW(sb->w.element),elt)!=0) continue;
		}
		xairo_uchar* s=GetWordListW(sb->w.xval);
		if (!pMatch(s,re)) continue;
		for (int x=0;x<sb->count;x++) {
			xara_loc l;
			Loc(x+sb->first,&l);
			b[ppa].lcode=l.fi;
			b[ppa].lw=l.len; 
			b[ppa].lw0=0;
//			b[ppa].nxt=l.nxt;
			b[ppa].pos=l.off;
//			b[ppa].pre=l.pre;
			b[ppa].sb=l.sb;
			b[ppa].word=l.nw;
			b[ppa].wc=1;
			ppa++;
		}
	}
	qsort(b,pa,sizeof (struct locent1),cl1);
	FILE* f=fopen(path,"wb");
	fwrite(b,sizeof (struct locent1),pa,f);
	fclose(f);
	return pa;
}



extern "C" int getFirstFtab()
{
	return dicthead.maxpos;
}

extern "C" BOOL getNextFtab(int* i,xairo_uchar* s,int* f,int* ff)
{
	if (*i==-1) return FALSE;
	xara_lemma l=ll[*i];
	wcscpy(s,GetWordListW(l.word[0]));
	*f=l.freq;
	*ff=l.forms;
	*i=l.nxt;
	return TRUE;
}


extern "C" int xbx, xby;

// This is ghastly! At least use a binary chop for a log(n) solution
extern "C" int GetFirstDictEntry(xairo_uchar *s)
{
	int n=dicthead.ndix;
	int p=wcslen(s);
	int k1=-1;
	int i;
	for (i=0;i<n;i++) {
		xairo_uchar* ss=GetWordListW(ll[i].word[0]);
		int k=theCollator->compare(ss,s,p);
		if (k >= 0 && k1==-1) k1=i;
		if (k > 0) break;
	}
	int k2=i;
	xbx=k1;
	xby=k2;
	return k1;
}

struct _ad 
{
	int w;
	int f;
};

_ad* adict=NULL;

int LookupType(char* element,char* att)
{
	int i;
	for (i=0;i<ktl;i++) {
		xara_type* t=&tl[i];
		if (strcmp(att,GetWordList(t->att))!=0) continue;
		if (strcmp(element,"0")==0 && t->t==it_global_attribute ||
			strcmp(element,GetWordList(t->element))==0 && t->t==it_local_attribute) break;
	}
	if (i==ktl) return -1;
	return tl[i].first;
}

int LookupTypeW(xairo_uchar* element,xairo_uchar* att)
{
	int i;
	for (i=0;i<ktl;i++) {
		xara_type* t=&tl[i];
		if (wcscmp(att,GetWordListW(t->att))!=0) continue;
		if (wcscmp(element,L("0"))==0 && t->t==it_global_attribute ||
			wcscmp(element,GetWordListW(t->element))==0 && t->t==it_local_attribute) break;
	}
	if (i==ktl) return -1;
	return tl[i].first;
}

extern "C" int LookupAtt(char* element,char* att)
{
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
	xara_streambase *sb; int k=0;
	int p0, p;
	p=p0=LookupType(element,att);
	int x=sizeof(sb)+ncont*sizeof(int);
	while (p!=-1) {
		sb=GetSB(p);
		if (!AKTest(sb)) continue;
//		fseek(fStreamBase,p*x,SEEK_SET);
//		if (fread(&sb,sizeof sb,1,fStreamBase)==0) break;
//		sb.w.con=new int [ncont];
		p=sb->next;
//		fread(sb.w.con,sizeof (int),ncont,fStreamBase);
		if (sb->w.t != it_local_attribute && sb->w.t != it_global_attribute
			&& sb->w.t != it_content_attribute) continue;
		char* e=GetWordList(sb->w.element);
		if (sb->w.t == it_local_attribute && strcmp(e,element) != 0) continue;
		char* a=GetWordList(sb->w.name);
		if (strcmp(a,att) != 0) continue;
		k++;
	}
//	fclose(f);
	if (adict) delete [] adict;
	adict=new _ad[k];
//	f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
	k=0;
	p=p0;
	while (p !=-1) {
		sb=GetSB(p);
		if (!AKTest(sb)) continue;
//		fseek(fStreamBase,p*x,SEEK_SET);
//		if (fread(&sb,sizeof sb,1,fStreamBase)==0) break;
		p=sb->next;
//		sb.w.con=new int [ncont];
//		fread(sb.w.con,sizeof (int),ncont,fStreamBase);
		if (sb->w.t != it_local_attribute && sb->w.t != it_global_attribute
			&& sb->w.t != it_content_attribute) continue;
		char* e=GetWordList(sb->w.element);
		if (sb->w.t == it_local_attribute && strcmp(e,element) != 0) continue;
		char* a=GetWordList(sb->w.name);
		if (strcmp(a,att) != 0) continue;
		adict[k].w=sb->w.val;
		adict[k].f=sb->count;
		k++;
	}
//	fclose(f);
	return k;
}

extern "C" void AttDict(int i,char* b,int* f)
{
	strcpy(b,GetWordList(adict[i].w));
	*f=adict[i].f;
}

extern "C" struct stream *MakeFormStream(const xairo_uchar* headword,int k)
{
	int kk=LookupW(headword);
	if (kk==-1) return FALSE;
//	for (int i=0;i<dicthead.ndix;i++)
//		if (ll[i].word==kk) break;
	int i=llr[kk];
	if (i==-1) return NULL;
	if (k>=ll[i].forms) return FALSE;
	return MakeMultipleStream(ll[i].pt[k]);
}

extern "C" int CountDocuments()
{
	struct stream *s=MakeAtomStream("document",0,it_global_attribute_value);
	if (s==NULL) {
		syslog(LOG_NOTICE,"There is no document label stream");
		return 0;
	}
	RewindStream(s);
	int n=CountStream(s);
	FreeStream(s);
	return n;
}

extern "C" long CountWords(sara_ulong cc)
{
	struct stream* s, *s1; 
	s=MakeAtomStream("document",0,it_start);
	s1=MakeAtomStream("document",0,it_end);
	s->cursor=cc;
	s1->cursor=cc;
	SetPos(s);
	SetPos(s1);
	int n=s1->wpos-s->wpos;
	FreeStream(s);
	FreeStream(s1);
	return n;
}


extern "C" BOOL decode(int code,const char* input,int k,XMLCh* output,int* pos,int* len)
{
	struct stream *s0=MakeAtomStream("encoding",0,it_global_attribute_value);
	s0->cursor=code;
	SetPos(s0);
	xairo_uchar* enc=GetWordListW(s0->wpos);
	FreeStream(s0);
// Now puzzle out the encoding. See XML App F.1
	UErrorCode err=U_ZERO_ERROR;
	UConverter* u=ucnv_openU(enc,&err);
	if (err != U_ZERO_ERROR) return FALSE;
	int* os=new int [k+1];
	const char* i0=input;
	xairo_uchar* target=output;
	UChar* t0=target;
	ucnv_toUnicode(u,&target,target+MAXSENTENCE,&input,input+k,os,TRUE,&err);
	int kk=target-t0;
	t0[kk]=0;
	int pos1=*pos+*len;
	int i;
	for (i=0;i<k;i++) {
		if (os[i] >= *pos) break;
	}
	*pos=i;
	for (i=0;i<k;i++) {
		if (os[i] >= pos1) break;
	}
	pos1=i;
	*len=i-*pos;
	delete [] os;
	return TRUE;
}

extern "C" xairo_uchar* Wide(const char* s)
{
	int k=strlen(s);
	xairo_uchar* w=new xairo_uchar [k+1];
	mbstowcs(w,s,k+1);
	return w;
}

extern "C" void DW(xairo_uchar* w)
{
	delete [] w;
}

#define NORM_MODE UNORM_NFKC

extern "C" void MkLC(xairo_uchar* c)
{
	UnicodeString u(c);
	UnicodeString u1=u.foldCase();
	int k=u1.length();
	u1.extract(0,k,c);
	c[k]=0;
}

extern "C" void MkNorm(xairo_uchar* c)
{
	XMLCh* cc=NULL;
	UErrorCode err=U_ZERO_ERROR;
	int k=unorm_normalize(c,-1,NORM_MODE,0,cc,0,&err);
	cc=new XMLCh[k+1];
	err=U_ZERO_ERROR;
	unorm_normalize(c,-1,NORM_MODE,0,cc,k+1,&err);
	assert (U_SUCCESS(err));
	wcscpy(c,cc);
	delete cc;
}

extern "C" int Tokenise(xairo_uchar* in, xairo_uchar* out)
{
	brit->setText(in);
	int i=brit->first(); int i0=0; int k=0;
	int nn=0;
	do {
		bool b=true;
		for (int j=i0;j<i;j++) {
			if (!u_isWhitespace(in[j])) b=false;
		}
		if (!b) {
			wcsncpy(out+k,in+i0,i-i0);
			k+=i-i0;
			out[k]=0;
			k++;
			nn++;
		}
		i0=i;
		i=brit->next();
	}
	while (i!=BreakIterator::DONE);
	out[k]=0;
	return nn;
}



		



int TC(xairo_uchar* s1,xairo_uchar* s2)
{
	UErrorCode success = U_ZERO_ERROR;
	Collator* myCollator = Collator::createInstance(success);
	return myCollator->compare(s1,s2);
}

extern "C" BOOL StrComp(xairo_uchar a,xairo_uchar b,int n)
{
	xairo_uchar s[2], t[2];
	s[0]=a; s[1]=0;
	t[0]=b; t[1]=0;
	UErrorCode success = U_ZERO_ERROR;
	Collator* myCollator = Collator::createInstance(success);
	myCollator->setStrength((Collator::ECollationStrength) n);
	return myCollator->compare(s,t)==0;
}

// prepare for compression by routing odd refs to the loc table through here

/*
bool Loc(int n,xara_loc* l)
{
	FILE* f = fopen(exppath(PARAM_ETCPATH,"xdblocs"),"rb");
	fseek(f, n*sizeof (struct xara_loc),SEEK_SET);
	fread(l,sizeof (struct xara_loc),1,f);
	fclose(f);
	return true;
}
*/

extern "C" void Loc(int,xara_loc*);


int npage = -1;

 
void ReadThreadPage(int n,int* b)
{
	int off=threadtab[n].off;
	int len=threadtab[n].len;
	int nf=threadtab[n].fi;
	unsigned char* c=new unsigned char[len];
	fseek(fThread[nf],off,SEEK_SET);
	fread(c,1,len,fThread[nf]);
	unsigned long d=loc_hdr_t.ps*sizeof(int);
	uncompress((unsigned char*) b,&d,c,len);
	delete [] c;
}


int GetThread(int n)
{
	int np=n / maxiloct;
	int ni=n % maxiloct;
	if (np != npage) {
		ReadThreadPage(np,thpa);
		npage=np;
	}
	return thpa[ni];
}

// nl arg is for colligations
extern "C" xairo_uchar* HeadWord(int n,int nl,BOOL* bC)
{
//	xara_loc l;
	*bC=TRUE;
	int nx=nl?nl-1:nlemma;
//	Loc(n,&l);
	int sb=GetThread(n);
	xara_streambase* s=GetSB(sb);
//	FILE* f=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	fseek(fStreamBase,sb*(sizeof s+ncont*sizeof(int)),SEEK_SET);
//	fread(&s,sizeof s,1,fStreamBase);
//	s.w.con=new int[ncont];
//	fread(s.w.con,sizeof(int),ncont,fStreamBase);
//	fclose(f);
	if (s->w.t==it_enddoc) {
		*bC=FALSE;
		return NULL;
	}
	if (s->w.t != it_content) return NULL;
	int w;
	if (nx < 0) w=s->w.val;
	else w=s->w.con[nx];
	xairo_uchar* ww=GetWordListW(w);
	for (int i=0;i<nctab0;i++) {
		if (conttab[i].idx != nx) continue;
		if (wcscmp(ww,conttab[i].def)==0) return NULL;
	}
	return ww;
}

/*
extern "C" int NextLoc(int n)
{
	xara_loc l;
	Loc(n,&l);
	return l.nxt;
}

// guard against moving across document boundaries


extern "C" int PrevLoc(int n)
{
	xara_loc l;
	Loc(n,&l);
	return l.pre;
}
*/
extern "C" int WN(int n) // don't use this
{
	xara_loc l;
	Loc(n,&l);
	return l.nw;
}

extern "C" int DicFrequency(const xairo_uchar* w)
{
	int k=LookupW(w);
	if (k==-1) return NULL;
	int i=llr[k];
	if (i==-1) return 0;
	return ll[i].freq;
}

/*
The comparison algorithm.
	
  Suppose two entities f0 and f1.
  Let m be the entity to which both are (possibly indirectly) attached.  
  Suppose f1 > f0. Let a(e) be the entity to which e is directly attached.
  Then f1 is not equal to m (f1 > f0 >= m). So a(f1) is attached to m.
  By induction, f we repeatedly apply a to the larger of f0, f1 then the results remain attached to m.
  But a(f1)<f1 so this process must terminate.
  So eventually f0 and f1 are equal. But then f0=f1=m.
*/





struct _etab {
	xairo_uchar* name;
	char path[_MAX_PATH];
	xairo_uchar* enc;
	int file, offset, nw;
};

bool Bound0(const xairo_uchar* lab,int& off,int& count)
{
	int hi0=hash(lab);
	int d=hi[hi0];
	while (d != -1) {
		if (wcscmp(lab,GetWordListW(d))==0) break;
		d=wl[d].next;
	}
	if (d==-1) return false;
	int dp=wl[d].dicptr;
//	FILE* fStreamBase=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
//	fseek(fStreamBase,dp*(sizeof (xara_streambase)+ncont*sizeof(int)),SEEK_SET);
	int i=0;
	while (true) {
		xara_streambase *sb=GetSB(dp+i++); 
//		int* con=new int [ncont];
//		fread(&sb,sizeof (sb),1,fStreamBase);
//		fread(con,sizeof(int),ncont,fStreamBase);
//		delete con;
		if (sb->w.val != d) break;
		if (!AKTest(sb)) continue;
		if (sb->w.t==it_global_attribute_value) {
			off=sb->first;
			count=sb->count;
//			fclose(f);
			return true;
		}
	}
//	fclose(f);
	return false;
}



int netab;
_etab* etab;

void MakeEntityTable()
{
	int o0, o1, o2, o3;
	int c0, c1, c2, c3;
	Bound0(L("ename"),o0,c0);
	Bound0(L("encoding"),o1,c1);
	Bound0(L("path"),o2,c2);
	Bound0(L("attach"),o3,c3);
	netab=c1;
	etab=new _etab[c1];
	xara_loc l; 
//	FILE* f=fopen(exppath(PARAM_ETCPATH,"xdblocs"),"rb");
//	fseek(f,o0*sizeof (xara_loc),SEEK_SET);
	int i;
	for (i=0;i<c1;i++) {
		Loc(o0+i,&l);
//		fread(&l,sizeof l,1,f);
		etab[i].name=GetWordListW(l.nw);
	}
//	fseek(f,o1*sizeof (xara_loc),SEEK_SET);
	for (i=0;i<c1;i++) {
		Loc(o1+i,&l);
//		fread(&l,sizeof l,1,f);
		etab[i].enc=GetWordListW(l.nw);
	}
//	fseek(f,o2*sizeof (xara_loc),SEEK_SET);
	for (i=0;i<c1;i++) {
		Loc(o2+i,&l);
//		fread(&l,sizeof l,1,f);
		strcpy(etab[i].path,GetWordList(l.nw));
	}
//	fseek(f,o3*sizeof (xara_loc),SEEK_SET);
	for (i=0;i<c1;i++) {
		Loc(o3+i,&l);
//		fread(&l,sizeof l,1,f);
		xairo_uchar* w=GetWordListW(l.nw);
		etab[i].file=-1; 
		char* xx=new char[wcslen(w)+1];
		wcstombs(xx,w,wcslen(w)+1);
		sscanf(xx,"%d-%d-%d",&etab[i].file,&etab[i].offset,&etab[i].nw);
		delete [] xx;
	}
//	fclose(f);
}

extern "C" int CompareWord(int loc,int w,int loc1,int w1)
{
	if (loc < loc1) {
		int k=etab[loc1].file;
		if (k==-1) return -1;
		else return CompareWord(loc,w,k,etab[loc1].nw);
	} else if (loc==loc1)
		return w-w1;
	else {
		return -CompareWord(loc1,w1,loc,w);
	}
}

extern "C" int Outer(int loc)
{
	while (etab[loc].file != -1) loc=etab[loc].file;
	return loc;
}

extern "C" int ComparePos(int loc,int p,int loc1,int p1)
{
	if (loc < loc1) {
		int k=etab[loc1].file;
		if (k==-1) return -1;
		else return ComparePos(loc,p,k,etab[loc1].offset);
	} else if (loc==loc1)
		return p-p1;
	else {
		return -ComparePos(loc1,p1,loc,p);
	}
}

extern "C" int dcmp(const xairo_uchar *s1, const xairo_uchar *s2)
{
//	return 	theCollator->compare(s1,s2,wcslen(s2));
	return wcsncmp(s1,s2,wcslen(s2));
}

extern "C" struct regexpnew *ParseRegExp(const xairo_uchar *pattern,BOOL bCS)
{
	UErrorCode pErr=U_ZERO_ERROR;
	RegexPattern* e=RegexPattern::compile(pattern,UREGEX_COMMENTS|(bCS?0:UREGEX_CASE_INSENSITIVE),pErr);
	if (U_FAILURE(pErr)) return NULL;
	regexpnew* n=new regexpnew;
	n->imp=e;
	return n;
}

extern "C" BOOL pMatch(const xairo_uchar* word,struct regexpnew *re)
{
	UErrorCode pE=U_ZERO_ERROR; 
	UnicodeString s(word);
	RegexPattern* e=(RegexPattern*) re->imp;
	RegexMatcher* m=e->matcher(s,pE);
	UBool b=m->matches(pE);
	delete m;
	return b;
}

static xara_streambase* sb; 
int m_x;
int g_kk;

xara_streambase* GetSB(int i)
{
	if (i < g_kk)
		return &sb[i];
	else
		return NULL;
}

bool OpenSB()
{
	fStreamBase=fopen(exppath(PARAM_INDEXPATH,"xdbdict"),"rb");
	if (fStreamBase==NULL) {
		bootlog("The dictionary is missing");
		return false;
	}
	fseek(fStreamBase,0,SEEK_END);
	int k=ftell(fStreamBase);
	fseek(fStreamBase,0,SEEK_SET);
	m_x=sizeof (xara_streambase) + ncont*sizeof (int); // better keep in hdr
	g_kk=k/m_x;
	sb=new xara_streambase [g_kk];
	if (sb==NULL) {
		bootlog("Insufficient memory for dictionary (%d needed)",k);
		return false;
	}
	for (int i=0;i<g_kk;i++) {
        fread(&sb[i],sizeof (xara_streambase),1,fStreamBase);
		sb[i].w.con=new int [ncont];
		fread(sb[i].w.con,sizeof(int),ncont,fStreamBase);
	}
	return true;
}

extern "C" int AKStat(xairo_uchar *key,xairo_uchar* pos,int* d)
{
	*d=0; int i;
	char cn[MAXLTNAME];
	wcstombs(cn,key,MAXLTNAME); 
	for (i=0;i<nctab0;i++) {
		if (strcmp(cn,conttab[i].name)==0)
			break;
	}
	if (i==nctab0) return TRUE;
	int id=conttab[i].idx;
	int k=LookupW(pos);
	if (k==-1) return TRUE;
	for (int j=0;;j++) {
		xara_streambase* sb=GetSB(j);
		if (sb==NULL) break;
		if (sb->w.t != it_content) continue;
		if (sb->w.con[id]==k) *d+=sb->count;
	}
	for (int j=0;j<conttab[i].nf;j++)
		if (conttab[i].fil[j]==k) return FALSE;
	return TRUE;
}

extern "C" void AKFilter(xairo_uchar *key,xairo_uchar* pos)
{
	char cn[MAXLTNAME];
	wcstombs(cn,key,MAXLTNAME); int i; 
	for (i=0;i<nctab0;i++) {
		if (strcmp(cn,conttab[i].name)==0)
			break;
	}
	if (i==nctab0) return;
	int nf=conttab[i].nf++;
	conttab[i].fil[nf]=LookupW(pos);
}

extern "C" void AKClear(xairo_uchar *key)
{
	char cn[MAXLTNAME];
	wcstombs(cn,key,MAXLTNAME); int i;
	for (i=0;i<nctab0;i++) {
		if (strcmp(cn,conttab[i].name)==0)
			break;
	}
	if (i==nctab0) return;
	conttab[i].nf=0;

}

bool AKTest(xara_streambase* sb)
{
	if (sb->w.t != it_content) return true;
	for (int i=0;i<nctab0;i++) {
		for (int j=0;j<conttab[i].nf;j++) {
			if (conttab[i].fil[j]==sb->w.con[conttab[i].idx])
				return false;
		}
	}
	return true;
}

extern "C" BOOL AKExists()
{
	for (int i=0;i<nctab0;i++) {
		if (conttab[i].nf > 0) return true;
	}
	return false;
}

struct _dsc {
	const xairo_uchar* type, *elt, *att, *val;
};

extern "C" int dsccomp(const void* a,const void* b)
{
	_dsc* aa=(_dsc*) a;
	_dsc* bb=(_dsc*) b;
	int k=wcscmp(aa->type,bb->type);
	if (k) return k;
	k=wcscmp(aa->elt,bb->elt);
	if (k) return k;
	k=wcscmp(aa->att,bb->att);
	if (k) return k;
	k=wcscmp(aa->val,bb->val);
	return 0;
}


void FakeDSC()
{
	char dpath[_MAX_PATH];
	strcpy(dpath,etcpath);
	strcat(dpath,corpname);
	strcat(dpath,".dscx");
	FILE* f=fopen(dpath,"w");
	fprintf(f,"ver %d\r\n",100);
	fprintf(f,"# file generated by the xaira server\r\n");
	fprintf(f,"option noposindex\r\n");
	int i=0; int k=0;
	while (true) {
		xara_streambase *sb=GetSB(i++);
		if (sb==NULL) break;
		switch (sb->w.t) {
			case it_local_attribute:
				k++;
				break;
		}
	}
	_dsc* dsc=new _dsc [k];
	i=0; k=0;
	while (true) {
		xara_streambase *sb=GetSB(i++);
		if (sb==NULL) break;
		switch (sb->w.t) {
			case it_local_attribute:
				dsc[k].type=L("av");
				dsc[k].elt=GetWordListW(sb->w.element);
				dsc[k].att=GetWordListW(sb->w.name);
				dsc[k].val=GetWordListW(sb->w.val);
				k++;
				break;
		}
	}
	qsort(dsc,k,sizeof (_dsc),dsccomp);
	const xairo_uchar* e0=L("");
	const xairo_uchar* a0=L("");

	for (int i=0;i<k;i++) {
		if (wcscmp(e0,dsc[i].elt) != 0) {
			fwprintf(f,L("elt %s e b\r\n"),dsc[i].elt);
			e0=dsc[i].elt;
			a0=L("");
		}
		if (wcscmp(a0,dsc[i].att) != 0) {
			fwprintf(f,L("att %s CDATA 0\r\n"),dsc[i].att);
			a0=dsc[i].att;
		}
		fwprintf(f,L("val %s\r\n"),dsc[i].val);
	}
	fclose(f);
    
}




extern "C" const xairo_uchar* GetTaxName(int i)
{
	return taxtab[i].name;
}

extern "C" int GetNTax()
{
	return ntax;
}

extern "C" int GetTaxClasses(const xairo_uchar* name ,int* texts)
{
	struct stream *s;
	texts[0]=texts[1]=0;

	s=MakeAttStream(L("0"),L("catref"),(xairo_uchar*) name,it_global_attribute_value);
	if (s==NULL) return 0;
	int k=CountStream(s); int i=2;
	while (AdvanceStream(s)) {
		texts[i++]=atoi(GetWordList(s->wpos))+1;
	}
	FreeStream(s);
	return k;
}


