#include "Corpinc.h"

#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/sax2/DefaultHandler.hpp>
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/util/TransService.hpp>
#include <xercesc/util/TranscodingException.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/validators/common/Grammar.hpp>
#include <unicode/brkiter.h>
#include <unicode/rbbi.h>
#include <unicode/uchar.h>
#include <unicode/unorm.h>
#include <unicode/coll.h>
#include <unicode/regex.h>
#include <zlib.h>

using namespace XERCES_CPP_NAMESPACE;

extern "C" void FlattenStream(struct stream *);

char dtdpath[_MAX_PATH];

struct stream* BuildXMLQuery(DOMElement*,char*);
enum _tag {TAG_SEQ,TAG_OR,TAG_AND,TAG_PROD,TAG_BPROD,TAG_NEG,TAG_LEMMA,
	TAG_FORM,TAG_POS,TAG_PHRASE,TAG_WORD,TAG_ELEMENT,
	TAG_PATTERN,TAG_SCOPE,TAG_MAX};

struct tagtab {
	char* name;
	int arity;
} tags[TAG_MAX] =
{
	{"seq",0},
	{"or",0},
	{"and",0},
	{"prod",0},
	{"bprod",2},
	{"neg",1},
	{"lemma",1},
	{"form",1},
	{"pos",2},
	{"phrase",0},
	{"word",1},
	{"element",0},
	{"pattern",1},
	{"scope",2}
};

class EH:public ErrorHandler
{
public:
	EH(xairo_uchar*);
	void  warning (const SAXParseException &); 
	void  error (const SAXParseException &);
	void  fatalError (const SAXParseException &); 
	void  resetErrors (); 
	xairo_uchar* buffer;
	int errloc;
};

EH::EH(xairo_uchar* c):buffer(c),errloc(-1)
{
}

void EH::warning(const SAXParseException &exp)
{
	if (buffer[0]==0) wcscpy(buffer,exp.getMessage());
}

void EH::error(const SAXParseException &exp)
{
	if (buffer[0]==0) wcscpy(buffer,exp.getMessage());
	if (errloc=-1) errloc=exp.getColumnNumber();
}

void EH::fatalError(const SAXParseException &exp)
{
	if (buffer[0]==0) wcscpy(buffer,exp.getMessage());
}

void EH::resetErrors()
{
	buffer[0]=0;
}

xairo_uchar buffer[1000];

extern "C" struct stream *BuildXMLQuery(const xairo_uchar* xql,char* err)
{
	xairo_uchar eb[200]; eb[0]=0; 
	XercesDOMParser d; EH eh(eb);
//	d.loadGrammar(dtd,Grammar::DTDGrammarType,true);
	swprintf(buffer,L("<!DOCTYPE cql SYSTEM \"%S\">"),dtdpath);
	int ke=wcslen(buffer);
	wcscat(buffer,xql);	
	MemBufInputSource is((const unsigned char* const) buffer,wcslen(buffer)*2,L("xql-client")); 
	is.setEncoding(L("UTF-16"));
	d.setErrorHandler(&eh);
	d.setValidationScheme(XercesDOMParser::Val_Always);
	d.setCreateCommentNodes(false);
	d.setIncludeIgnorableWhitespace(false);
	d.parse(is);
	if (d.getErrorCount()) {
		sprintf(err,"Parser error: %S",eb);
		int p=eh.errloc-ke;
		return NULL;
	}
	DOMDocument* doc=d.getDocument();
	DOMElement* r=doc->getDocumentElement();
	if (wcscmp(r->getTagName(),L("cql")) != 0) {
		strcpy(err,"Bad document type");
		return NULL;
	}
	DOMNodeList* l=r->getChildNodes();
	if (l->getLength() != 1) {
		strcpy(err,"<cql> tag must have just one child");
		return NULL;
	}
	DOMNode* n=l->item(0);
	if (n->getNodeType() != DOMNode::ELEMENT_NODE) 
	{
		strcpy(err,"<cql> content is not a query");
        return NULL;
	}
	return BuildXMLQuery((DOMElement*) n,err);
}


struct stream* BuildXMLQuery(DOMElement* e,char* err)
{
	const xairo_uchar* tag=e->getTagName(); int it;
	char* ttag=new char[wcslen(tag)+1];
	wcstombs(ttag,tag,wcslen(tag)+1);
	for (it=0;it<TAG_MAX;it++) {
		if (strcmp(ttag,tags[it].name)==0) break;
	}
	if (it==TAG_MAX) {
		sprintf(err,"Unknown tag %s",ttag);
		return NULL;
	}
	delete [] ttag;
	DOMNodeList* l=e->getChildNodes();
	if (tags[it].arity && tags[it].arity != l->getLength()) {
		sprintf(err,"Wrong number of arguments for %S",tag);
		return NULL;
	}
	switch (it) {
		case TAG_PHRASE:
			{
				const xairo_uchar* s=e->getTextContent();
				xairo_uchar ss[MAXPHRASE]; //range check
				BOOL bCase=wcscmp(e->getAttribute(L("case")),L("yes"))==0;
				int n=Tokenise((xairo_uchar*) s,ss);
				xairo_uchar* sx=ss;
				if (n==1) 
					if (wcscmp(sx,L("_"))==0) {
						struct stream* st=GrabStream();
						if (st==NULL) {
							strcpy(err,"No more streams");
							return NULL;
						}
						st->type=TY_ALL;
						st->wtype=ID_WORD;
						return st; 
					} else
						return MakeAllPOSStreamW(sx,TRUE,bCase);
				else 
				{
					struct stream* st=GrabStream();
					if (st==NULL) {
						strcpy(err,"No more streams");
						return NULL;
					}
					st->type=TY_SEQ;
					st->ops=n;
					if (n>=MAXOPS) {
						strcpy(err,"Too many substreams");
						FreeStream(st);
						return NULL;
					}
					int jj=0;
					while (*sx) {
						if (wcscmp(sx,L("_"))==0) {
							st->op[jj]=GrabStream();
							if (st->op[jj]==NULL) {
								strcpy(err,"No more streams");
							}
							st->op[jj]->type=TY_ALL;
							st->op[jj]->wtype=ID_WORD;
						} else
							st->op[jj]=MakeAllPOSStreamW(sx,TRUE,bCase);
						if (st->op[jj]==NULL) {
							FreeStream(st);
							return NULL;
						}
						jj++;
						sx+=wcslen(sx)+1;
					}
					FlattenStream(st);
					return st;
				}
			}
		case TAG_ELEMENT:
			{
				BOOL bE=wcscmp(e->getAttribute(L("end")),L("yes"))==0;
				const xairo_uchar* name=e->getAttribute(L("name"));
				int n=l->getLength();
				if (n>=MAXOPS) {
					strcpy(err,"Too many substreams");
					return NULL;
				}
				const xairo_uchar** attnames=new const xairo_uchar* [n];
				const xairo_uchar** attvals=new const xairo_uchar* [n];
				const xairo_uchar** attextras=new const xairo_uchar* [n];
				int i;
				for (i=0;i<n;i++) {
					DOMNode* nd=l->item(i);
					if (nd->getNodeType() != DOMNode::ELEMENT_NODE) break;
					DOMElement* ne=(DOMElement*) nd;
					if (wcscmp(ne->getTagName(),L("attribute")) != 0) break;
					attnames[i]=ne->getAttribute(L("name"));
					BOOL bVar=wcscmp(ne->getAttribute(L("var")),L("yes"))==0;
					BOOL bPatt=wcscmp(ne->getAttribute(L("pattern")),L("yes"))==0;
					if (bVar) 
						attextras[i]=L("_");
					else if (bPatt)
						attextras[i]=L("*");
					else 
						attextras[i]=ne->getAttribute(L("value"));
					attvals[i]=nd->getTextContent();
				}
				if (i<n) {
					strcpy(err,"Unrecognised attribute specifier");
					return NULL;
				}
				if (bE) 
					return MakeAtomStreamW((xairo_uchar*) name, 0, it_end);
				else
					return MakeSGMLStream2W((xairo_uchar*) name, n, (xairo_uchar**) attnames, (xairo_uchar**) attvals,(xairo_uchar**) attextras);
			}
		case TAG_PATTERN:
			{
				BOOL bCase=wcscmp(e->getAttribute(L("case")),L("yes"))==0;
				const xairo_uchar* p=e->getTextContent();
// change to parse expression here for better error handling
				return MakeRegExpStreamW((xairo_uchar*) p,bCase); // entry from XCQL only
			}
		case TAG_FORM:
			{
				int i=_wtoi(e->getAttribute(L("index")));
				const xairo_uchar* w=e->getTextContent();
          	    return MakeFormStream(w,i);
			}
		case TAG_POS:
			{
				DOMNode* n=l->item(0);
				if (n->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					strcpy(err,"Bad word arg in pos");
					return NULL;
				}
				DOMElement* ne=(DOMElement*) n;
				if (wcscmp(ne->getTagName(),L("word")) != 0) 
				{
					strcpy(err,"Bad word arg in pos");
					return NULL;
				}
				const xairo_uchar* word=ne->getTextContent();
				DOMNode* nn=l->item(1);
				if (nn->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					strcpy(err,"Bad postag arg in pos");
					return NULL;
				}
				DOMElement* nne=(DOMElement*) nn;
				if (wcscmp(nne->getTagName(),L("poscode")) != 0) 
				{
					strcpy(err,"Bad postag arg in pos");
					return NULL;
				}
				const xairo_uchar* key=nne->getAttribute(L("key"));
				const xairo_uchar* tag=nne->getTextContent();
				if (wcscmp(word,L("_"))==0) {
					return MakeAllPosStream(tag,key);
				} else 
					return MakePosStream3W(word, tag, key);
			}
		case TAG_LEMMA:
			{
				const xairo_uchar *w = e->getTextContent();
				return MakeLemmaStream(w);
			}
		case TAG_NEG:
			{
				DOMNode* n=l->item(0);
				if (n->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					strcpy(err,"Bad arg in neg");
					return NULL;
				}
				struct stream* st1=BuildXMLQuery((DOMElement*) n,err);
				if (st1 == NULL) 
					return NULL;
				struct stream* st=GrabStream();
				if (st==NULL) {
					strcpy(err,"No more streams");
					return NULL;
				}
				st->type = TY_NEG;
				st->ops = 1;
				st->op[0] = st1;
				return st;
			}
		case TAG_SCOPE:
			{
				DOMNode* n1=l->item(0);
				if (n1->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					strcpy(err,"Bad first arg in scope");
					return NULL;
				}
				struct stream* st1=BuildXMLQuery((DOMElement*) n1,err);
				if (st1 == NULL) 
					return NULL;
				DOMNode* n2=l->item(1);
				if (n2->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					strcpy(err,"Bad second arg in scope");
					return NULL;
				}
				DOMElement* ne2=(DOMElement*) n2;
				int tp, tn; struct stream *s1, *s2;
				if (wcscmp(ne2->getTagName(),L("element"))==0) {
					BOOL bE=wcscmp(e->getAttribute(L("end")),L("yes"))==0;
					const xairo_uchar* name=ne2->getAttribute(L("name"));
					if (bE) {
						s1 = MakeAtomStreamW((xairo_uchar*) name, 0, it_end);
						s2 = MakeAtomStreamW((xairo_uchar*) name, 0, it_start);
						tp = LT_XSGML;
					} else {
		                s1 = BuildXMLQuery(ne2,err);                     
						s2 = MakeAtomStreamW((xairo_uchar*) name, 0, it_end);
						tp = LT_SGML;
					}
					if (s1==NULL || s2==NULL) return NULL;
					tn=0;
				} else if (wcscmp(ne2->getTagName(),L("span"))==0) {
					tn=_wtoi(ne2->getAttribute(L("size")));
					s1=s2=NULL;
					tp = LT_INT;
				} else {
					strcpy(err,"Bad second argument in scope");
					return NULL;
				}
				if (st1->type == TY_PROD || st1->type == TY_BPROD) {
					st1->limit1 = s1;
					st1->limit2 = s2;
					st1->limtype = tp;
					st1->limint = tn;
					return st1;
				} else if (tp == LT_INT)
					return st1; // limit has no effect in this case
				else {
					struct stream *st=GrabStream();
					if (st==NULL) {
						strcpy(err,"No more streams");
						return NULL;
					}
					st->type = TY_LIMIT;
					st->op[0] = st1;
					st->ops = 1;
					st->limit1 = s1;
					st->limit2 = s2;
					st->limtype = tp;
					return st;
				}
			}
		case TAG_AND: 
		case TAG_OR:
		case TAG_PROD:
		case TAG_BPROD:
		case TAG_SEQ:
			{
				int n=l->getLength(); int i;
				if (n==1) {
					DOMNode* n=l->item(0);
					if (n->getNodeType() != DOMNode::ELEMENT_NODE) break;
					return BuildXMLQuery((DOMElement*) n,err);
				}
				if (n>=MAXOPS) {
					strcpy(err,"Too many substreams");
					return NULL;
				}
				struct stream* st=GrabStream();
				if (st==NULL) {
					strcpy(err,"No more streams");
					return NULL;
				}
				switch (it) {
					case TAG_AND:
						st->type=TY_AND;
						break;
					case TAG_OR:
						st->type=TY_OR;
						break;
					case TAG_PROD:
						st->type=TY_PROD;
						break;
					case TAG_BPROD:
						st->type=TY_BPROD;
						break;
					case TAG_SEQ:
						st->type=TY_SEQ;
						break;
				}
				st->ops=n;
				for (i=0;i<n;i++) {
					DOMNode* n=l->item(i);
					if (n->getNodeType() != DOMNode::ELEMENT_NODE) break;
					st->op[i]=BuildXMLQuery((DOMElement*) n,err);
					if (st->op[i]==NULL && it != TAG_OR) break;
				}
				if (i < n) {
					FreeStream(st);
					return NULL;
				}
				FlattenStream(st);
				return st;
			}



 	}
	return NULL;
}

const char* dtd_txt[]={
"<!ENTITY % cqlop \"seq | or | and | prod | bprod | neg | lemma | form | pos | phrase | element | pattern | scope\">\n",
"<!ELEMENT cql (%cqlop;)>\n",
"<!ELEMENT seq (%cqlop;)*>\n",
"<!ELEMENT or (%cqlop;)*>\n", 
"<!ELEMENT and (%cqlop;)*>\n", 
"<!ELEMENT prod (%cqlop;)*>\n", 
"<!ELEMENT bprod (%cqlop;)*>\n", 
"<!ELEMENT neg (%cqlop;)>\n", 
"<!ELEMENT lemma (#PCDATA)>\n",
"<!ELEMENT form (#PCDATA)>\n",
"<!ATTLIST form\n",
"	index CDATA #REQUIRED>\n",
"<!ELEMENT pos (word,poscode)>\n",
"<!ELEMENT phrase (#PCDATA)>\n",
"<!ATTLIST phrase\n"
"	case CDATA \"no\">\n",
"<!ELEMENT element (attribute*)>\n",
"<!ATTLIST element\n"
"	end CDATA \"no\"\n"
"	name CDATA #REQUIRED>\n",
"<!ELEMENT pattern (#PCDATA)>\n",
"<!ATTLIST pattern\n"
"	case CDATA \"no\">\n",
"<!ELEMENT scope ((%cqlop;),(element|span))>\n",
"<!ELEMENT word (#PCDATA)>\n",
"<!ELEMENT poscode (#PCDATA)>\n",
"<!ATTLIST poscode\n",
"	key CDATA #REQUIRED>\n",
"<!ELEMENT attribute (#PCDATA)>\n",
"<!ATTLIST attribute\n",
"	name CDATA #REQUIRED\n",
"	var CDATA \"no\"\n",
"	pattern CDATA \"no\"\n",
"	value CDATA #IMPLIED>\n",
"<!ELEMENT span EMPTY>\n",
"<!ATTLIST span\n",
"	size   CDATA  #REQUIRED >\n"};


#define NDTDT ((sizeof dtd_txt)/sizeof (char*))


void MakeDTD()
{
	strcpy(dtdpath,_tempnam(tmppath,"cqldtd"));
	FILE* f=fopen(dtdpath,"w");
	for (int i=0;i<NDTDT;i++) {
		fwrite(dtd_txt[i],1,strlen(dtd_txt[i]),f);
	}
	fclose(f);
}
