#include "SServer.h"
#include "DataReader.h"

#define SWAP4(X) X=((X & 0xFF)<<24)|((X & 0xFF00)<<8)|((X & 0xFF0000)>>8)|((X & 0xFF000000)>>24)
#define SWAP2(X) X=((X & 0xFF)<<8)|((X & 0xFF00)>>8)

XAIRO_NAMESPACE_BEGIN

CDataReader::CDataReader(int n,FILE* f):m_file(f),m_n(n) 
{
	
}

bool CDataReader::bRev=false;

CDataReader::~CDataReader(void)
{
}

int CDataReader::ReadThread(int n,int* b,int len)
{
	unsigned char* c=new unsigned char[len];
	fread(c,1,len,m_file);
	unsigned long dn=n*sizeof(int);
	uncompress((unsigned char*) b,&dn,c,len);
	delete [] c;
	if (bRev) {
		for (int i=0;i<n;i++) {
              SWAP4(b[i]);
 		}
	}
	return n;
}

int CDataReader::ReadWLHash(int n,int* d)
{
	int k=fread(d,sizeof(int),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
            SWAP4(d[i]);
		}
	}
	return k;
}

int CDataReader::ReadAccel(int n,int* d)
{
	int k=fread(d,sizeof(int),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i]);
		}
	}
	return k;
}

int CDataReader::ReadWChars(int n,xairo_uchar* d)
{
	int k=fread(d,sizeof(xairo_uchar),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP2(d[i]);
		}
	}
	return k;
}

int CDataReader::ReadWLHeader(int n,xara_wl_header* d)
{
	int k=fread(d,sizeof(xara_wl_header),n,m_file);
	if (d->sig==DBSIGR) bRev=true;
	if (bRev) {
		SWAP4(d->sig);
		SWAP4(d->version);
		SWAP4(d->size);
		SWAP4(d->hashsize);
		SWAP4(d->places);
		SWAP4(d->shift);
		SWAP4(d->cont);
		SWAP4(d->lemma);
		SWAP4(d->ntax);

	}
	return k;
}

int CDataReader::ReadLemmaHeader(int n,xara_lemma_header* d)
{
	int k=fread(d,sizeof(xara_lemma_header),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
			SWAP4(d[i].maxpos);
			SWAP4(d[i].ndix);
			SWAP4(d[i].nu);
			SWAP4(d[i].nfacets);
		}
	}
	return k;
}

int CDataReader::ReadLemma(int n,int nf,xara_lemma* d)
{
	int k=0;
	for (int i=0;i<n;i++) {
		fread(&d[i],sizeof(int),3,m_file);
		fread(&d[i].word[0],sizeof(int),nf,m_file);
		if (bRev) {
			SWAP4(d[i].freq);
			SWAP4(d[i].forms);
			SWAP4(d[i].nxt);
			for (int j=0;j<nf;j++)
				SWAP4(d[i].word[j]);
		}
		k++;
	}
	return k;
}

int CDataReader::ReadForm(int n,xara_form* d)
{
	int k=fread(d,sizeof(xara_form),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
			SWAP4(d[i].ptr);
			SWAP4(d[i].count);
			SWAP4(d[i].freq);
		}
	}
	return k;
}

int CDataReader::ReadWLHeaderAddkey(int n,xara_wl_header_addkey* d)
{
	int k=fread(d,sizeof(xara_wl_header_addkey),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i].name);
               SWAP4(d[i].bLemma);
               SWAP4(d[i].def);
		}
	}
	return k;
}

int CDataReader::ReadWLHeaderTax(int n,xara_wl_header_tax* d)
{
	int k=fread(d,sizeof(xara_wl_header_tax),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i].name);
               SWAP4(d[i].nc);
		}
	}
	return k;
}

int CDataReader::ReadWLFile(int n,xara_wl_file* d)
{
	int k=fread(d,sizeof(xara_wl_file),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i].len);
               SWAP4(d[i].next);
               SWAP4(d[i].dicptr);
		}
	}
	return k;
}

int CDataReader::ReadType(int n,xara_type* d)
{
	int k=fread(d,sizeof(xara_type),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
			int x=d[i].t;
            SWAP4(x);
			d[i].t=(typ) x;
            SWAP4(d[i].element);
            SWAP4(d[i].att);
            SWAP4(d[i].first);
		}
	}
	return k;
}

int CDataReader::ReadAhdr(int n,xara_ahdr* d)
{
	int k=fread(d,sizeof(xara_ahdr),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i].nt);
               SWAP4(d[i].na);
		}
	}
	return k;
}

int CDataReader::ReadLocHdr(int n,xara_loc_hdr* d)
{
	int k=fread(d,sizeof(xara_loc_hdr),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i].ps);
               SWAP4(d[i].strength);
               SWAP4(d[i].nf);
               SWAP4(d[i].np);
               SWAP4(d[i].maxpage);
 		}
	}
	return k;
}

int CDataReader::ReadLocPage(int n,xara_loc_page* d)
{
	int k=fread(d,sizeof(xara_loc_page),n,m_file);
	if (bRev) {
		for (int i=0;i<n;i++) {
               SWAP4(d[i].fi);
               SWAP4(d[i].off);
               SWAP4(d[i].len);
		}
	}
	return k;
}

int CDataReader::ReadLoc(int n,xara_loc* b,int len)
{
	unsigned char* c=new unsigned char[len];
	fread(c,1,len,m_file);
	unsigned long dn=n*sizeof(xara_loc);
	uncompress((unsigned char*) b,&dn,c,len);
	delete [] c;
	if (bRev) {
		for (int i=0;i<n;i++) {
              SWAP4(b[i].fi);
              SWAP4(b[i].nw);
              SWAP4(b[i].off);
              SWAP4(b[i].len);
              SWAP4(b[i].sb);
              SWAP4(b[i].seq);
              SWAP4(b[i].parent);
 		}
	}
	return n;
}

int CDataReader::ReadStreamBase(int n,int ncont,xara_streambase* d)
{
	int i=0;
	for (;i<n;i++) {
		int delta=sizeof(void *)-sizeof(int); // 64 bit kludge
        if (fread(&d[i],sizeof (xara_word)-delta,1,m_file)==0) break;
        if (fread(&d[i].first,sizeof (xara_streambase)
				-sizeof(xara_word),1,m_file)==0) break;
		if (bRev) {
			SWAP4(d[i].w.val);
			int x=d[i].w.t;
			SWAP4(x);
			d[i].w.t=(typ) x;
			SWAP4(d[i].w.uri);
			SWAP4(d[i].w.name);
			SWAP4(d[i].w.euri);
			SWAP4(d[i].w.element);
			SWAP4(d[i].w.xval);
			SWAP4(d[i].w.nco);
			SWAP4(d[i].first);
			SWAP4(d[i].count);
			SWAP4(d[i].next);
			SWAP4(d[i].acc);
		}
		d[i].w.con=new int [ncont];
		fread(d[i].w.con,sizeof(int),ncont,m_file);
		if (bRev) {
			for (int j=0;j<ncont;j++) {
				SWAP4(d[i].w.con[j]);
			}
		}
	}
	return i;
}


void CDataReader::Seek(int n)
{
	assert(m_n != 0);
	fseek(m_file,n*m_n,SEEK_SET);
}

void CDataReader::Swap8(__int64& x)
{
	__int64 mask=0xffffffff;
	mask <<=32;
	__int64 x0=(unsigned int) (x & 0xffffffff);
	__int64 x1=(unsigned int) ((x & mask) >> 32);
	SWAP4(x0);
	SWAP4(x1);
	x = x1 | (x0 << 32);
}


XAIRO_NAMESPACE_END

