#include "SServer.h"
#include "Dictionary.h"
#include "WordList.h"
#include "Parameters.h"
#include "DataReader.h"

XAIRO_NAMESPACE_BEGIN

CDictionary::CDictionary(CXairoServer* s):m_ll(0),m_llr(0),m_server(s)
{
}

CDictionary::~CDictionary()
{
	for (int i=0;i<m_dicthead.ndix;i++) {
		xara_lemma* xl=(xara_lemma*) &m_ll[i*m_ns];
		delete [] xl->pt;
	}
	delete [] m_ll;
	delete [] m_llr;
}

bool CDictionary::GetLimits(const xairo_uchar* w,int& x1,int& x2)
{
	return GetLimits(w,0,m_ndix,x1,x2);
}

/*int CDictionary::GetIndex(int w)
{
	return m_llr[w];
}
*/
bool CDictionary::GetLimits(const xairo_uchar* patt,int i1,int i2,int &x1,int& x2)
{
   const xairo_uchar* s;
   if (i1 > i2)
		return false;
   int mid = (i1 + i2) / 2;
   int c;
   if (mid<m_ndix) {
		xara_lemma* l=LookupEntry(mid);
		s=m_server->GetWL()->GetWordW(l->word[0]);
		c = wcsncmp(s, patt, wcslen(patt));
   } else 
	   c=-1;
   if (c < 0)
      return GetLimits(patt, mid + 1, i2, x1, x2);
   else if (c == 0) {
      if (i1 == i2) {
            x1 = i1;
            x2 = i1;
         return true;
      } else if (i1 == mid) {
		  int z1, z2;
            x1 = i1;
            x2 = GetLimits(patt,i2, i2, z1, z2) ? i2 : i1;
         return true;
      } else {
		  int z;
		  GetLimits(patt,i1,mid,x1,z);
		  GetLimits(patt,mid,i2,z,x2);
          return true;
      }
   } else
      return GetLimits(patt,i1, mid - 1, x1, x2);
}


xara_lemma* CDictionary::LookupEntry(const xairo_uchar* w)
{
	int d=m_server->GetWL()->LookupW(w);
	if (d==-1) return NULL;
	int dd=m_llr[d];
	if (dd==-1) return NULL;
	return LookupEntry(dd);
}

xara_lemma* CDictionary::LookupEntry2(int nw,UnicodeString* w)
{
	int d=m_server->GetWL()->LookupW(w[0].getTerminatedBuffer());
	if (d==-1) return NULL;
	int dd=m_llr[d];
	if (dd==-1) return NULL;
	while (dd<m_dicthead.ndix) {
		xara_lemma* wl=(xara_lemma*) &m_ll[dd*m_ns];
		if (wl->word[0] != d) break;
		int j=0;
		for (j=1;j<nw;j++) {
			if (m_server->GetWL()->LookupW(w[j].getTerminatedBuffer()) != wl->word[j]) break;
		}
		if (j==nw) return wl;
		dd++;
	}
	return NULL;
}

int CDictionary::WordIndex(const xairo_uchar* w)
{
	int d=m_server->GetWL()->LookupW(w);
	if (d==-1) return -1;
	int dd=m_llr[d];
	return dd;
}

bool CDictionary::Load(int n)
{
	xairo_uchar nm[_MAX_PATH];
	swprintf(nm,L("xdblemma%d"),n+1);
	const xairo_uchar* path=m_server->GetParam()->exppath(PARAM_INDEXPATH,nm);
	FILE* f=_wfopen(path,L("rb"));
	if (f==NULL) {
		return false;
	}
	CDataReader* d=new CDataReader(0,f);
	d->ReadLemmaHeader(1,&m_dicthead);
	if (m_ll) delete [] m_ll;
	if (m_llr) delete [] m_llr;
	m_ns=sizeof(xara_lemma)+(m_dicthead.nfacets-1)*sizeof(int);
	m_ll=new char[m_ns*m_dicthead.ndix];
	m_llr=new int [m_server->GetWL()->Count()];
	for (int i=0;i<m_server->GetWL()->Count();i++) m_llr[i]=-1;
	for (int i=0;i<m_dicthead.ndix;i++) {
		xara_lemma* xl=(xara_lemma*) &m_ll[i*m_ns];
		d->ReadLemma(1,m_dicthead.nfacets,xl);
		if (m_llr[xl->word[0]]==-1) m_llr[xl->word[0]]=i;
		int j=xl->forms;
		xl->pt=new xara_form[j];
		d->ReadForm(j,xl->pt);
	}
	fclose(f);
	m_ndix=m_dicthead.ndix;
	return true;
}

int CDictionary::GetMaxFreq()
{
	return m_dicthead.maxpos;
}

int CDictionary::GetNextFreq(int i)
{
	xara_lemma* xl=(xara_lemma*) &m_ll[i*m_ns];
	return xl->nxt;
}

xara_lemma* CDictionary::LookupEntry(int i)
{
	return 	(xara_lemma*) &m_ll[i*m_ns];

}

int CDictionary::GetHeadwordIndex(xara_word* w)
{
	int n=m_nf;
	int* c=new int [n];
	for (int i=0;i<n;i++) {
		if (m_f[i]<0) c[i]=w->val;
		else c[i]=w->con[m_f[i]];
	}
	int d=c[0];
	int dd=m_llr[d];
	if (dd==-1) return -1;
	while (dd<m_dicthead.ndix) {
		xara_lemma* wl=(xara_lemma*) &m_ll[dd*m_ns];
		if (wl->word[0] != d) break;
		int j=0;
		for (j=1;j<n;j++) {
			if (c[j] != wl->word[j]) break;
		}
		if (j==n) return dd;
		dd++;
	}
	return -1;
}

XAIRO_NAMESPACE_END
