#include "SServer.h"
#include "EntityManager.h"
#include "StreamFactory.h"
#include "WordList.h"
#include "Parameters.h"
#include "XairoSolveError.h"

XAIRO_NAMESPACE_BEGIN

CEntityManager::CEntityManager(CXairoServer* s):m_server(s)
{
}

CEntityManager::~CEntityManager(void)
{
}

bool CEntityManager::Load()
{
	CStreamFactory* sf=m_server->GetSF();
	CAtomStream* s=(CAtomStream*) sf->MakeAtomAStream(L("ename"),XAIRA_URI,it_global_attribute_value);
	if (s==NULL) {
		m_server->BootLog(L("No text entities found"));		
		return false;
	}
	m_nents=s->Count();
	m_ents=new _ent [m_nents];
	int t=-1; int i=0;
	while (s->Advance(t,true)) {
		m_ents[i++].name=m_server->GetWL()->GetWordW(s->m_nw);
		t=s->m_seq+1;
	}
	sf->FreeStream(s);
	s=(CAtomStream*) sf->MakeAtomAStream(L("encoding"),XAIRA_URI,it_global_attribute_value);
	t=-1; i=0;
	while (s->Advance(t,true)) {
		m_ents[i++].enc=m_server->GetWL()->GetWordW(s->m_nw);
		t=s->m_seq+1;
	}
	sf->FreeStream(s);
	s=(CAtomStream*) sf->MakeAtomAStream(L("path"),XAIRA_URI,it_global_attribute_value);
	t=-1; i=0;
	while (s->Advance(t,true)) {
		m_ents[i++].path=m_server->GetWL()->GetWordW(s->m_nw);
		t=s->m_seq+1;
	}
	sf->FreeStream(s);
	s=(CAtomStream*) sf->MakeAtomAStream(L("attach"),XAIRA_URI,it_global_attribute_value);
	t=-1; i=0;
	while (s->Advance(t,true)) {
		const xairo_uchar* x=m_server->GetWL()->GetWordW(s->m_nw);
		char* xx=new char[wcslen(x)+1];
		wcstombs(xx,x,wcslen(x)+1);
		if (strcmp(xx,"/")==0) m_ents[i].file=-1;
		else
		sscanf(xx,"%d-%d-%d",&m_ents[i].file,&m_ents[i].offset,&m_ents[i].nw);
		t=s->m_seq+1; i++;
	}
	sf->FreeStream(s);
	return true;
}

const xairo_uchar* CEntityManager::GetTextPath(int i)
{
	if (i==0) return m_server->GetParam()->exppath(PARAM_ROOTPATH,m_server->GetParam()->Param(PARAM_HEADER));
	else if (i==1) return m_server->GetParam()->exppath(PARAM_ROOTPATH,m_server->GetParam()->Param(PARAM_BIB));
	else return m_server->GetParam()->exppath(PARAM_TEXTPATH,m_ents[i].path);
}

const xairo_uchar* CEntityManager::TextExtract(int fi,int pos0,int len,int* k1,int *k2)
{
	const xairo_uchar* path=GetTextPath(fi);
	FILE* f=_wfopen(path,L("rb"));
	if (f==NULL) throw new CXairoError(CXairoIError::ERR_NOTEXT,L("Missing text"));
	char* x=new char [len];
	fseek(f,pos0,SEEK_SET);
	fread(x,1,len,f);
	fclose(f);
	xairo_uchar* w=decode(fi,x,len,k1,k2);
	delete [] x;
	return w;
}

int CEntityManager::GetOuterEntity(int i)
{
	int ii;
	while (i != -1) {
		ii=i;
		i=m_ents[i].file;
	}
	return ii;
}

using namespace XERCES_CPP_NAMESPACE;


xairo_uchar* CEntityManager::decode(int code,const char* input,int k,int* pos,int* pos1)
{
	UErrorCode err=U_ZERO_ERROR;
	const xairo_uchar* enc=m_ents[code].enc;
	UConverter* u=ucnv_openU(enc,&err);
	if (err != U_ZERO_ERROR) throw new CXairoError(CXairoIError::ERR_ENCODING,L("Cannot make decoder"));
	xairo_uchar* target=new xairo_uchar [k];
	UChar* t0=target;
	int* os=new int [k+1];
	const char* i0=input;
	ucnv_toUnicode(u,&target,target+k,&input,input+k,os,TRUE,&err);
	int kk=target-t0;
	xairo_uchar* xx=new xairo_uchar [kk+1];
	int i;
	for (i=0;i<kk;i++) {
		xx[i]=t0[i];
	}
	xx[kk]=0;
	delete [] t0;
	for (i=0;i<k;i++) {
		if (os[i] >= *pos) break;
	}
	*pos=i;
	for (i=0;i<kk;i++) {
		if (os[i] >= *pos1) break;
	}
	*pos1=i;
	delete [] os;
	return xx;
}

/*m_cs=mynew unsigned char [k];
	m_b=input-i0;
	os[m_kk]=m_b;
	for (int i=0;i<m_kk;i++) {
		m_cs[i]=os[i+1]-os[i];
	}
	mydelete [] os;
	return true;

// must be changed to use ICU transcoders
xairo_uchar* CEntityManager::decode(int code,const char* input,int k,int* pos,int* pos1)
{
	const xairo_uchar* enc=m_ents[code].enc;
	XMLTransService *s=XMLPlatformUtils::fgTransService;
	XMLTransService::Codes c=XMLTransService::Ok; 
	XMLTranscoder* t=s->makeNewTranscoderFor(enc,c,k,XMLPlatformUtils::fgMemoryManager);
	if (c!=XMLTransService::Ok) return false;
	unsigned char *cs=new unsigned char [k]; 
	unsigned int b;
	int kk=0; int i;
	xairo_uchar* output=new xairo_uchar [k];
	try {
		kk=t->transcodeFrom((unsigned char*) input,k,output,k,b,cs);
	}
	catch (TranscodingException& e)
	{
		const xairo_uchar* w=e.getMessage();
		return NULL;
	}
	xairo_uchar* xx=new xairo_uchar [kk+1];
	for (i=0;i<kk;i++) {
		xx[i]=output[i];
	}
	xx[kk]=0;
	delete [] output;
	int x=0; 
	for (i=0;i<k;i++) {
		if (x >= *pos) break;
		x+=cs[i];
	}
	*pos=i;
	 x=0;
	for (i=0;i<k;i++) {
		if (x >= *pos1) break;
		x+=cs[i];
	}
	*pos1=i;
	return xx;
}
*/
XAIRO_NAMESPACE_END
