#include "SServer.h"
#include "Grammar.h"
#include "WordList.h"
#include "Parameters.h"
#include "Dictionary.h"
#include "SB.h"
#include "StreamFactory.h"
#include "Stream.h"
#include "XairoCorpusTextList.h"
#include "XairoCorpusText.h"
#include "XairoSolveError.h"

XAIRO_NAMESPACE_BEGIN


CXairoGrammar::CXairoGrammar(CXairoServer* s)
{
	m_server=s;
}

CXairoGrammar::~CXairoGrammar()
{
}

class GHandle : public DOMErrorHandler {
public:
	GHandle():bErr(false) {}
	bool handleError(const DOMError &);	
	bool bErr; XMLFileLoc row, col;
	const xairo_uchar* msg;
	const xairo_uchar* uri;
};

bool GHandle::handleError(const DOMError& e)
{
	msg=e.getMessage();
	DOMLocator* l=e.getLocation();
	row=l->getLineNumber();
	col=l->getColumnNumber();
	uri=_wcsdup(l->getURI());
	bErr=true;
	return false;
}

void CXairoGrammar::InitAddkeys()
{
	CWordList* wl=m_server->GetWL();
	int *x=new int [1];
	x[0]=-1;
	m_server->GetDict(-1)->SetFacets(1,x);
	m_naddkeys=wl->m_ncont;
	m_addkeys=new Addkey* [m_naddkeys];
	xara_wl_header_addkey* h=wl->m_cont;
	for (int i=0;i<m_naddkeys;i++) {
		const xairo_uchar* name=wl->GetWordW(h[i].name);
		DOMNodeList* l=m_doc->getElementsByTagName(L("addkeys"));
		DOMElement* e=(DOMElement*) l->item(0);
		DOMNodeList* ll=e->getElementsByTagName(L("addkey"));
		int m=ll->getLength();
		DOMElement* v=NULL;
		for (int j=0;j<m;j++) {
			e=(DOMElement*) ll->item(j);
			if (wcscmp(e->getAttribute(L("name")),name)==0)
			{v=e; break;}
		}
		m_addkeys[i]=new Addkey(name,i,wl->GetWordW(h[i].def),h[i].bLemma != 0,v,m_server);
		int *y=new int [1];
		y[0]=i;
		m_server->GetDict(i)->SetFacets(1,y);
	}
	DOMNodeList* l=m_doc->getElementsByTagName(L("lemmascheme"));
	xairo_uchar** ln=new xairo_uchar*[l->getLength()];
	int il=0;
	for (int i=0;i<(int) l->getLength();i++) {
		DOMElement* ln0=(DOMElement*) l->item(i);
		ln[i]=_wcsdup(ln0->getAttribute(L("name")));
		DOMNodeList* ll=ln0->getElementsByTagName(L("key"));
		int nf=ll->getLength();
		int* f=new int [nf];
		for (int j=0;j<nf;j++) {
			DOMElement* ek=(DOMElement*) ll->item(j);
			const xairo_uchar* kn=ek->getAttribute(L("name"));
			int k;
			for (k=0;k<m_naddkeys;k++) {
				if (wcscmp(kn,m_addkeys[k]->getName())==0) break;
			}
			f[j]=k;
		}
		m_server->GetDict(m_naddkeys+il++)->SetFacets(nf,f);
	}
	wl->SetLS(l->getLength(),ln);
}

const CXairoGrammar::Addkey* CXairoGrammar::getAddkeyById(const xairo_uchar* id) const
{
	for (int i=0;i<m_naddkeys;i++) {
		Addkey* a=m_addkeys[i];
		const xairo_uchar* x=a->getId();
		if (wcscmp(x,id)==0) return a;
	}
	return NULL;
}

bool CXairoGrammar::Init()
{
	DOMImplementationLS* di=(DOMImplementationLS*) DOMImplementation::getImplementation();
	m_parser=di->createLSParser(DOMImplementationLS::MODE_SYNCHRONOUS,NULL);
	const xairo_uchar* gfile=m_server->GetParam()->exppath(PARAM_INDEXPATH,L("xgrammar.xml"));
	GHandle eh;
	DOMConfiguration* dcfg=m_parser->getDomConfig();
	dcfg->setParameter(XMLUni::fgXercesSchema,true);
	dcfg->setParameter(XMLUni::fgDOMNamespaces,true);
//    dcfg->setParameter(L("schema-type"),L("http://www.w3.org/2001/XMLSchema"));
	dcfg->setParameter(L("validate"), true);
	xairo_uchar sl[_MAX_PATH*2];
	wcscpy(sl,L("http://xaira.org/ns/1.0 ")); 
	const char* p=m_server->GetDataPath();
	for (int i=0;i<(int) strlen(p);i++) {
		xairo_uchar x[2];
		if (p[i]==32)
			wcscat(sl,L("%20"));
		else {
			x[0]=p[i];
			x[1]=0;
			wcscat(sl,x);
		}
	}
	wcscat(sl,L("/xgrammar.xsd"));
	dcfg->setParameter(XMLUni::fgXercesSchemaExternalSchemaLocation,sl);
//	dcfg->setParameter(L("schema-location"),sl);
	dcfg->setParameter(L("error-handler"),&eh);
	m_doc=m_parser->parseURI(gfile);
	if (eh.bErr) { // need uri here
		m_server->BootLog(L("Parser error in %s at %d,%d: %s\n"),eh.uri,(unsigned) eh.row,(unsigned) eh.col,eh.msg);		
		return false;
	}
	InitAddkeys();
	return true;
}

int CXairoGrammar::getElementCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("elements"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("element"));
	return ll->getLength();
}

const CXairoGrammar::Element* CXairoGrammar::getElement(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("elements"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("element"));
	DOMElement* el=(DOMElement*) ll->item(i);
	void * v;
	v=el->getUserData(L("cache"));
	if (v==NULL) {
		v=new Element(el,m_server);
		el->setUserData(L("cache"),v,NULL);
	}
	return (Element*) v;
}

const xairo_uchar* CXairoGrammar::Partition::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Element::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Attribute::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Addkey::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::LemmaScheme::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Region::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Label::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Codebook::getId() const
{
	return m_ptr->getAttribute(L("id"));
}

const xairo_uchar* CXairoGrammar::Element::getName() const
{
	return m_ptr->getAttribute(L("name"));
}

const CXairoGrammar::Namespace* CXairoGrammar::Element::getNamespace() const
{
	return new Namespace(m_ptr->getAttribute(L("ns")));
}

bool CXairoGrammar::Element::inBib() const
{
	return wcscmp(m_ptr->getAttribute(L("bib")),L("yes"))==0;
}

const xairo_uchar* CXairoGrammar::Element::getGloss() const
{
	return m_ptr->getAttribute(L("gloss"));
}

int CXairoGrammar::Element::getAttributeCount() const
{
	DOMNodeList* ll=m_ptr->getElementsByTagName(L("attribute"));
	return ll->getLength();
}

int CXairoGrammar::getGlobalAttributeCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("elements"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("globals")); // globals or global?
	if (ll->getLength()==0) return 0;
	DOMElement* ee=(DOMElement*) ll->item(0);
	DOMNodeList* lll=ee->getElementsByTagName(L("attribute"));
	return lll->getLength();
}

const CXairoGrammar::Attribute* CXairoGrammar::getGlobalAttribute(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("elements"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("globals"));
	if (ll->getLength()==0) return 0;
	DOMElement* ee=(DOMElement*) ll->item(0);
	DOMNodeList* lll=ee->getElementsByTagName(L("attribute"));
	DOMElement* a=(DOMElement*) lll->item(i);
	void * v;
	v=a->getUserData(L("cache"));
	if (v==NULL) {
		v=new Attribute(a,L("0"),m_server);
		a->setUserData(L("cache"),v,NULL);
	}
	return (Attribute*) v;
}

const CXairoGrammar::Attribute *CXairoGrammar::Element::getAttribute(int i) const
{
	DOMNodeList* ll=m_ptr->getElementsByTagName(L("attribute"));
	DOMElement* a=(DOMElement*) ll->item(i);
	void * v;
	v=a->getUserData(L("cache"));
	if (v==NULL) {
		v=new Attribute(a,getName(),m_server);
		a->setUserData(L("cache"),v,NULL);
	}
	return (Attribute*) v;
}

const xairo_uchar* CXairoGrammar::Attribute::getName() const
{
	return m_ptr->getAttribute(L("name"));
}

const CXairoGrammar::Namespace* CXairoGrammar::Attribute::getNamespace() const
{
	return new Namespace(m_ptr->getAttribute(L("ns")));
}

bool CXairoGrammar::Attribute::inBib() const
{
	return wcscmp(m_ptr->getAttribute(L("bib")),L("yes"))==0;
}

const xairo_uchar* CXairoGrammar::Attribute::getGloss() const
{
	return m_ptr->getAttribute(L("gloss"));
}

CXairoIGrammar::Attribute::ip CXairoGrammar::Attribute::getIP() const
{
	const xairo_uchar* nip=m_ptr->getAttribute(L("ip"));
	if (wcscmp(nip,L("ID"))==0)
		return ip_id;
	else if (wcscmp(nip,L("IDREF"))==0)
		return ip_idref;
	else if (wcscmp(nip,L("CDATA"))==0)
		return ip_cdata;
	else if (wcscmp(nip,L("MULTIDREFS"))==0)
		return ip_tax;
	else return ip_none;
}



const xairo_uchar* CXairoGrammar::Attribute::getTarget() const
{
	return m_ptr->getAttribute(L("target"));
}


CXairoGrammar::Codebook::Codebook(DOMElement* e,CXairoServer* s):m_ptr(e),m_server(s)
{
}

const CXairoGrammar::Codebook* CXairoGrammar::Attribute::getCodebook() const
{
	DOMElement* cb=NULL;
	DOMNodeList* cl=m_ptr->getElementsByTagName(L("codebook"));
	if (cl->getLength() > 0) 
		cb=(DOMElement*) cl->item(0);
	if (cb==NULL) return NULL;
	void * v=cb->getUserData(L("cache"));
	if (v==NULL) {
		v=new Codebook(cb,m_server);
		cb->setUserData(L("cache"),v,NULL);
	}
	return (Codebook*) v;
}

const xairo_uchar* CXairoGrammar::Codebook::lookup(const xairo_uchar* name) const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("code"));
	for (int i=0;i<(int) l->getLength();i++) {
		DOMElement* code=(DOMElement*) l->item(i);
		if (wcscmp(name,code->getAttribute(L("name")))==0) 
		return code->getAttribute(L("gloss"));
	}
	return L("");
}

int CXairoGrammar::Codebook::getCount() const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("code"));
	return l->getLength();
}

void CXairoGrammar::Codebook::getPair(int i,const xairo_uchar** n1,const xairo_uchar** n2) const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("code"));
	DOMElement* code=(DOMElement*) l->item(i);
	*n1=code->getAttribute(L("name"));
	*n2=code->getAttribute(L("gloss"));
}

CXairoGrammar::Label::Label(DOMElement* e,CXairoServer* s):m_ptr(e),m_server(s)
{
}

const xairo_uchar* CXairoGrammar::Label::getElement() const
{
	return m_ptr->getAttribute(L("element"));
}

const xairo_uchar* CXairoGrammar::Label::getAttribute() const
{
	return m_ptr->getAttribute(L("attribute"));
}

int CXairoGrammar::getLabelCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("labels"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("label"));
	return ll->getLength();
}

const CXairoGrammar::Label* CXairoGrammar::getLabel(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("labels"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("label"));
	DOMElement* el=(DOMElement*) ll->item(i);
	void * v;
	v=el->getUserData(L("cache"));
	if (v==NULL) {
		v=new Label(el,m_server);
		el->setUserData(L("cache"),v,NULL);
	}
	return (Label*) v;
}
int CXairoGrammar::getWTagCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("wtags"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("wtag"));
	return ll->getLength();
}

const xairo_uchar* CXairoGrammar::getWTag(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("wtags"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("wtag"));
	DOMElement* el=(DOMElement*) ll->item(i);
	return el->getAttribute(L("id"));
}

int CXairoGrammar::getNamespaceCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("namespaces"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("namespace"));
	return ll->getLength();
}


const CXairoGrammar::Namespace* CXairoGrammar::getNamespace(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("namespaces"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("namespace"));
	DOMElement* ns=(DOMElement*)  ll->item(i);
	void * v;
	v=ns->getUserData(L("cache"));
	if (v==NULL) {
		v=new Namespace(ns->getAttribute(L("uri")));
		ns->setUserData(L("cache"),v,NULL);
	}
	return (Namespace*) v;
}

CXairoGrammar::Partition::Partition(const xairo_uchar* id,CXairoServer* s)
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
	m_bLoaded=false;
	m_texts=NULL;
}

CXairoGrammar::Partition::Partition(DOMElement* e,CXairoServer* s):m_ptr(e),m_server(s)
{
	m_bLoaded=false;
	m_texts=NULL;
}

const xairo_uchar* CXairoGrammar::Partition::getName() const
{
	return m_ptr->getAttribute(L("taxid"));
}

const xairo_uchar* CXairoGrammar::Partition::getDescription() const
{
	return m_ptr->getAttribute(L("desc"));
}

int CXairoGrammar::Partition::getCount() const
{
	DOMNodeList* ml=m_ptr->getChildNodes();
	return ml->getLength();
}

int CXairoGrammar::Partition::getIndex() const
{
	int i=0;
	DOMNode* n=m_ptr;
	while (n=n->getPreviousSibling()) i++;
	return i;
}


const CXairoGrammar::Partition* CXairoGrammar::Partition::getDivision(int i) const
{
	DOMNodeList* ml=m_ptr->getChildNodes();
	DOMElement* e=(DOMElement*) ml->item(i);
	void * v;
	v=e->getUserData(L("cache"));
	if (v==NULL) {
		v=new Partition(e,m_server);
		e->setUserData(L("cache"),v,NULL);
	}
	return (Partition*) v;
}

const CXairoGrammar::Partition* CXairoGrammar::getRootPartition() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("taxonomy"));
	if (l->getLength()==0) return NULL;
	DOMNode* n=l->item(0);
	void * v;
	v=n->getUserData(L("cache"));
	if (v==NULL) {
		v=new Partition((DOMElement*) n,m_server);
		n->setUserData(L("cache"),v,NULL);
	}
	return (Partition*) v;
}

void CXairoGrammar::Partition::Load()
{
	if (m_bLoaded) return;
	m_bLoaded=true;
	m_bHasClasses=false;
	CAtomStream* ds=m_server->GetDocumentStream();
	int n=ds->Count();
	CStreamFactory* sf=m_server->GetSF();
	CStream* s=sf->MakeAtomCStream(L("catref"),L("0"),XAIRA_URI,getName(),it_global_attribute_value);
	if (s==NULL) {
		return;
	}
	int k=0; int i=2;
	m_texts=new int[n];
	for (int ii=0;ii<n;ii++)
		m_texts[ii]=-1;
	while (s->Advance(k,true)) {
		m_texts[i++]=_wtoi(m_server->GetWL()->GetWordW(s->m_nw));
		k=s->m_seq+1;
	}
	sf->FreeStream(s);
	m_bHasClasses=true;
}

bool CXairoGrammar::Partition::hasClasses() const
{
	if (!m_bLoaded) ((CXairoGrammar::Partition*) this)->Load();
	return m_bHasClasses;
}

int CXairoGrammar::Partition::Classify(int i) const
{
	if (!m_bLoaded) ((CXairoGrammar::Partition*) this)->Load();
	return m_texts[i];
}

const CXairoGrammar::Partition* CXairoGrammar::Partition::GetParent(int& k) const
{
	DOMElement* e=(DOMElement*) m_ptr->getParentNode();
	DOMNodeList* l=e->getChildNodes();
	int i;
	for (i=0;i<(int) l->getLength();i++) {
		if (l->item(i)==m_ptr) break;
	}
	k=i;
	void * v;
	v=e->getUserData(L("cache"));
	if (v==NULL) {
		v=new Partition(e,m_server);
	e->setUserData(L("cache"),v,NULL);
	}
	return (Partition*) v;
}

bool CXairoGrammar::Partition::Test(int t,int p) const
{
	if (!m_bLoaded) ((CXairoGrammar::Partition*) this)->Load();
	return m_texts[t]==p;
}

int CXairoGrammar::Partition::getWordCount() const
{
	int ip=0;
	const CXairoGrammar::Partition* pp=GetParent(ip);
	if (pp==NULL) return 0;
	int nw=0; int k=0; int id=0;
	int sd=0;
	CXairoCorpusTextList l(m_server);
	for (int i=0;i<l.getCount();i++) {
		CXairoGrammar::Partition* ppp=(CXairoGrammar::Partition*) l.getCorpusText(i)->classify(pp);
		if (ppp && ppp->m_ptr==m_ptr) 
			k+=l.getCorpusText(i)->getWordCount();
	}
	return k;
}

const xairo_uchar* CXairoGrammar::Addkey::getAttribute() const
{
	return m_ptr->getAttribute(L("attribute"));
}

const CXairoGrammar::Codebook* CXairoGrammar::Addkey::getCodebook() const
{
	DOMElement* cb=NULL;
	DOMNodeList* cl=m_ptr->getElementsByTagName(L("codebook"));
	if (cl->getLength() > 0) 
		cb=(DOMElement*) cl->item(0);
	if (cb==NULL) return NULL;
	void * v=cb->getUserData(L("cache"));
	if (v==NULL) {
		v=new Codebook(cb,m_server);
		cb->setUserData(L("cache"),v,NULL);
	}
	return (Codebook*) v;
}

int CXairoGrammar::getRegionCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("regions"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("region"));
	return ll->getLength();
}

const xairo_uchar* CXairoGrammar::Region::getName() const
{
	return m_ptr->getAttribute(L("name"));
}

const xairo_uchar* CXairoGrammar::Region::getGloss() const
{
	return m_ptr->getAttribute(L("desc"));
}

int CXairoGrammar::Region::getValueCount() const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("value"));
	return l->getLength();
}

const xairo_uchar* CXairoGrammar::Region::getValue(int i) const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("value"));
	DOMElement* e=(DOMElement*) l->item(i);
	return e->getAttribute(L("name"));
}

const CXairoGrammar::Addkey* CXairoGrammar::Region::getAddkey() const
{
	const xairo_uchar* x=m_ptr->getAttribute(L("addkey"));
	int n=_wtoi(x);
	return m_server->GetGrammarImpl()->getAddkey(n);
}

const CXairoGrammar::Region* CXairoGrammar::getRegion(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("regions"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("region"));
	DOMElement* ls=(DOMElement*)  ll->item(i);
	void * v;
	v=ls->getUserData(L("cache"));
	if (v==NULL) {
		v=new Region(ls,m_server);
		ls->setUserData(L("cache"),v,NULL);
	}
	return (Region*) v;
}


int CXairoGrammar::getLemmaSchemeCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("lemmaschemes"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("lemmascheme"));
	return ll->getLength();
}

const xairo_uchar* CXairoGrammar::LemmaScheme::getName() const
{
	return m_ptr->getAttribute(L("name"));
}

const xairo_uchar* CXairoGrammar::LemmaScheme::getGloss() const
{
	return m_ptr->getAttribute(L("desc"));
}

int CXairoGrammar::LemmaScheme::getFacetCount() const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("key"));
	return l->getLength();
}

const xairo_uchar* CXairoGrammar::LemmaScheme::getFacet(int i) const
{
	DOMNodeList* l=m_ptr->getElementsByTagName(L("key"));
	DOMElement* e=(DOMElement*) l->item(i);
	return e->getAttribute(L("name"));
}

const CXairoGrammar::LemmaScheme* CXairoGrammar::getLemmaScheme(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("lemmaschemes"));
	if (l->getLength()==0) return NULL;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("lemmascheme"));
	DOMElement* ls=(DOMElement*)  ll->item(i);
	if (ls==NULL) return NULL;
	void * v;
	v=ls->getUserData(L("cache"));
	if (v==NULL) {
		v=new LemmaScheme(ls,m_server);
		ls->setUserData(L("cache"),v,NULL);
	}
	return (LemmaScheme*) v;
}



// change this to use an ID assigned when the grammar is defined 

const xairo_uchar* CXairoGrammar::gId(DOMElement* n) const
{
	return n->getAttribute(L("id"));
}

DOMElement* CXairoGrammar::gNode(const xairo_uchar* w) const
{
	DOMElement* e=m_doc->getElementById(w);
	if (e==NULL) throw(new CXairoError(CXairoIError::ERR_GRAMMARID,L("No grammar element for node id")));
	return e;
}

struct _fct {
	const xairo_uchar* n;
	wchar_t c;
} fctab[]= {
	{L("amp"),'&'},
	{L("lt"),'<'},
	{L("gt"),'>'},
	{L("apos"),'\''},
	{L("quot"),'"'}};

#define NSE (sizeof(fctab)/sizeof(fctab[0]))

xairo_uchar CXairoGrammar::EvalCharEntity(const xairo_uchar* w) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("chartable"));
	if (l->getLength()!=0) {
		DOMElement* tab=(DOMElement*) l->item(0);
		const xairo_uchar* trad=tab->getAttribute(L("radix"));
		int radix=10;
		if (trad) radix=_wtoi(trad);
		l=tab->getElementsByTagName(L("char"));
		for (int i=0;i<(int) l->getLength();i++) {
			DOMElement* e=(DOMElement*) l->item(i);
			if (wcscmp(e->getAttribute(L("name")),w)==0) {
				const xairo_uchar* v=e->getAttribute(L("code"));
				return (xairo_uchar)  wcstoul(v,NULL,radix);
			}
		}
	}
	for (int i=0;i<NSE;i++) {
		if (wcscmp(fctab[i].n,w)==0)
			return fctab[i].c;
	}
	return 0;
}

CXairoGrammar::Codebook::Codebook(const xairo_uchar* id,CXairoServer* s)
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
}

CXairoGrammar::Label::Label(const xairo_uchar* id,CXairoServer* s)
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
}

CXairoGrammar::Attribute::Attribute(const xairo_uchar* id,const xairo_uchar* e,CXairoServer* s) 
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_elt=e; 
	m_server=s;
}

/*
CXairoGrammar::Addkey::Addkey(const xairo_uchar* id,CXairoServer* s) 
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
}
*/

CXairoGrammar::LemmaScheme::LemmaScheme(const xairo_uchar* id,CXairoServer* s) 
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
}

CXairoGrammar::Region::Region(const xairo_uchar* id,CXairoServer* s) 
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
}

CXairoGrammar::Element::Element(const xairo_uchar* id,CXairoServer* s) 
{
	m_ptr=(DOMElement*) s->GetGrammarImpl()->gNode(id); 
	m_server=s;
}

int CXairoGrammar::getCodePageCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("codepages"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("charset"));
	return ll->getLength();
}

const xairo_uchar* CXairoGrammar::getCodePage(int i) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("codepages"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("charset"));
	e=(DOMElement*) ll->item(i);
	return e->getAttribute(L("name"));
}

int CXairoGrammar::getCharacterEntityCount() const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("chartable"));
	if (l->getLength()==0) return 0;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("char"));
	return ll->getLength();
}

void CXairoGrammar::getCharacterEntity(int i,const xairo_uchar** c,int* v) const
{
	DOMNodeList* l=m_doc->getElementsByTagName(L("chartable"));
	if (l->getLength()==0) return;
	DOMElement* e=(DOMElement*) l->item(0);
	DOMNodeList* ll=e->getElementsByTagName(L("char"));
	e=(DOMElement*) ll->item(i);
	*c=e->getAttribute(L("name"));
	*v=wcstoul(e->getAttribute(L("code")),NULL,16);
}

const char* CXairoGrammar::getDSC() const
{
	const xairo_uchar* dscfile=m_server->GetParam()->exppath(PARAM_INDEXPATH,L("dsc.txt"));
	int kd=wcslen(dscfile);
	char* df=new char [kd+1];
	wcstombs(df,dscfile,kd+1);
	FILE* f=fopen(df,"rb");
	if (f==NULL) throw new CXairoError(CXairoIError::ERR_UNKNOWN,L("No DSC file"));
	fseek(f,0,SEEK_END);
	int t=ftell(f);
	fseek(f,0,SEEK_SET);
	char* c=new char [t+1];
	fread(c,t,1,f);
	fclose(f);
	c[t]=0;
	return c;
}



/*

// DTD for xgrammar.xml file

<xgrammar>
<wtags/>
<addkeys>
<addkey name="xxx" gloss="">
<codebook/><addkey/>
<elements>
<global/>
<element name="xxx" ns="yyy" gloss="ggg">
<attribute name="xxx" gloss="ggg" ip="CDATA">
<codebook/></attribute></element>
</elements>
<chartable>
<char name="nnn" val="vvv"/>
</chartable>
<coderanges/>
</xgrammar>

*/

XAIRO_NAMESPACE_END
