#include "SServer.h"
#include "KDB.h"
#include "XairoSolveError.h"

XAIRO_NAMESPACE_BEGIN


CKDB::CKDB():m_n(0)
{
}

CKDB::~CKDB(void)
{
}

bool CKDB::Init(CXairoServer* s)
{
	m_server=s;
	strcpy(m_tmppath,s->GetTempPath());
	strcat(m_tmppath,"/");
	char path[_MAX_PATH];
	sprintf(path,"%sx.tmp",m_tmppath);
	FILE* f=fopen(path,"wb");
	if (f==NULL) {
		m_server->BootLog(L("Cannot write temporary directory"));
		return false;
	}
	fclose(f);
	remove(path);
	m_nkeys=0;
	m_keys=NULL;
	return true;
}

CDBKey* CKDB::getKey(int n,int l)
{
//	CDBKey* k=m_keys[n];
//	if (k->bNull) {
//		m_server->Debug("Throw!");
//		throw(CXairoError(CXairoIError::ERR_NULLKEY,L("Null key accessed")));
//	}
//	return k;
	CDBKey* k=new CDBKey;
	k->bNull=false;
	k->n=n;
	k->l=l;
	return k;
}

void CKDB::deletekey(CDBKey* key)
{
	if (key->bNull) throw(CXairoError(CXairoIError::ERR_NULLKEY,L("Null key accessed")));
	char path[_MAX_PATH];
	if (key->f) fclose(key->f);
//	sprintf(path,"%sx%07d.tmp",m_tmppath,key->n);
	keypath(key,path);
	remove(path);
	key->bNull=true;
	delete key;
}

int CKDB::record(CDBKey* key,void* v,unsigned c,int k)
{
	if (key->bNull) throw(CXairoError(CXairoIError::ERR_NULLKEY,L("Null key accessed")));
	return fwrite(v,c,k,key->f);
}

void CKDB::closekey(CDBKey* key)
{
	if (key->bNull) throw(CXairoError(CXairoIError::ERR_NULLKEY,L("Null key accessed")));
	fclose(key->f);
	key->f=NULL;
//	char path[_MAX_PATH];
//	FILE* f;
//	sprintf(path,"%sx%07d.tmp",m_tmppath,key->n);
//	key->f=fopen(path,"rb");
}


CDBKey* CKDB::newkey(int l,const char* p)
{
	CDBKey* k=new CDBKey;
	k->l=l;
	k->pre=strdup(p);
//	int j;
//	for (j=0;j<m_nkeys;j++) {
//		if (m_keys[j]->bNull) break;
//	}
//	if (j>=m_nkeys) {
//		int t=m_nkeys+NKEYS;
//		CDBKey** nk=new CDBKey*[t];
//		for (int jj=0;jj<m_nkeys;jj++)
//			nk[jj]=m_keys[jj];
//		for (int jj=0;jj<NKEYS;jj++)
//			nk[m_nkeys+jj]=new CDBKey;
//		if (m_keys) delete [] m_keys;
//		j=m_nkeys;
//		m_nkeys=t;
//		m_keys=nk;
//	}
//	int id=j;
//	CDBKey* k=m_keys[j];
	char path[_MAX_PATH]; FILE* f; int i;
	for (i=m_n;;i++) {
		sprintf(path,"%s%s%07d.tmp",m_tmppath,p,i);
		if (f=fopen(path,"rb")) {fclose(f);} else break;
	}
	f=fopen(path,"wb");
	if (f==NULL) throw new CXairoError(CXairoIError::ERR_UNKNOWN,L("Cannot create db file"));
//	k->id=id;
	k->n=i;
	k->f=f;
	k->l=l;
	k->bNull=false;
//	k->pid=m_server->getPid();
	m_n=i+1;

	return k;
}

int CKDB::countrecords(CDBKey* k)
{
	if (k->count >=0) return k->count;
	if (k->bNull) throw(CXairoError(CXairoIError::ERR_NULLKEY,L("Null key accessed")));
	char path[_MAX_PATH];
//	FILE* f=k->f;
//	sprintf(path,"%sx%07d.tmp",m_tmppath,k->n);
	keypath(k,path);
	FILE* f=fopen(path,"rb");
	fseek(f,0,SEEK_END);
	int kk=ftell(f);
	fclose(f);
	return k->count=kk/k->l;
}

bool CKDB::recorded(CDBKey* key,int k,void* v)
{
	fop(key);
	bool b=frecorded(key,k,v);
	fcl(key);
	return b;
}

void CKDB::fop(CDBKey* k)
{
	char path[_MAX_PATH];
//	FILE* f=key->f;
//	sprintf(path,"%sx%07d.tmp",m_tmppath,k->n);
	keypath(k,path);
	k->f=fopen(path,"rb");
}

void CKDB::fcl(CDBKey* k)
{
	fclose(k->f);
	k->f=NULL;
}

bool CKDB::frecorded(CDBKey* key,int k,void* v)
{
	FILE* f=key->f;
	if (fseek(f,k*key->l,SEEK_SET)==0) {
		int p=fread(v,key->l,1,f);
		return p==1;
	} 
	return false;
}




const xairo_uchar* CKDB::getId(CDBKey* k)
{
	return CXairoServer::MakeId(L("<kdb><n>%d</n><s>%d</s></kdb>"),k->n,k->l);
}

void CKDB::keypath(CDBKey* k,char* p)
{
	sprintf(p,"%s%s%07d.tmp",m_tmppath,k->pre,k->n);
}

XAIRO_NAMESPACE_END
