#include "SServer.h"
#include "RegularExpression.h"
#include "WordList.h"
#include "Dictionary.h"
#include "AttributeDictionary.h"
#include "XairoWordList.h"
#include "XairoAttributeValueList.h"
#include "XairoAddkeyValueList.h"
#include "SB.h"
#include "Locs.h"
#include "Thread.h"
#include "Grammar.h"
#include "language.h"
#include "XairoSolution.h"
#include "Stream.h"
#include "StreamFactory.h"
#include "XairoSolveError.h"
#include "EntityManager.h"
#include "XairoCorpusTextList.h"
#include "XairoCorpusTextFilter.h"
#include "Parameters.h"
#include "Accelerators.h"
#include "XairoExtendedForm.h"
#include "XairoAddkeyFilter.h"
#include "BootMonitor.h"
#ifdef _WIN32
#include "../XairaVersion.h"
#endif

#ifdef _WIN32
extern "C" const char* WindowsTempPath();
__declspec(thread) int g_thrid;
#include <crtdbg.h>
#else
static int g_pid;
#endif

#ifdef _DEBUG
#define mynew ::new(_NORMAL_BLOCK,__FILE__,__LINE__)
#define mydelete ::delete
#else
#define mynew new
#define mydelete delete
#endif

XAIRO_NAMESPACE_BEGIN

#ifdef __WIN32
int YourAllocHook( int allocType, void *userData, size_t size, int blockType, 
   long requestNumber, const unsigned char *filename, int lineNumber)
{
//	_ASSERT(size < 100000);
	return TRUE;
}
#endif


CXairoServer* CXairoServer::getServer()
{
	CXairoServer* s=mynew CXairoServer;
//	s->Init(params);
	return s;
}

CXairoServer::CXairoServer(void)
{
	m_accels=NULL;
	m_language=NULL;
	m_locs=NULL;
	m_thread=NULL;
	m_sb=NULL;
	m_dict=NULL;
	m_attdict=NULL;
	m_wl=NULL;
	m_param=NULL;
}

CXairoServer::~CXairoServer(void)
{
	mydelete m_accels;
	mydelete m_language;
	mydelete m_locs;
	mydelete m_thread;
	mydelete m_sb;
	int nd=m_wl->GetNAddkeys()+m_wl->GetNLemmas();
	for (int i=0;i<nd+1;i++) {
		mydelete m_dict[i];
	}
	mydelete [] m_dict;
	mydelete m_attdict;
	mydelete m_wl;
	mydelete m_param;
}



CXairoIWordList* CXairoServer::getWordList(int np,const xairo_uchar** pattern,const xairo_uchar* ls)
{
	int ak=GetWL()->lemma_index(ls);
	CRegularExpression **re=mynew CRegularExpression *[np];;
	for (int x=0;x<np;x++) {
		UnicodeString us=GetLanguage()->MkNorm(pattern[x]);
		re[x] = mynew CRegularExpression(us.getTerminatedBuffer(),false);
		if (!re[x]->Parse()) throw mynew CXairoError(CXairoIError::ERR_REGEXP,L("Badly formed regexp"));
	}
	UnicodeString pre;
	pre=re[0]->Predict(10);
	int k1=-1; int k2=-1;
	if (!GetDict(ak)->GetLimits(pre.getTerminatedBuffer(),k1,k2)) return NULL;
	const xairo_uchar* w;
	CDBKey* key=m_kdb.newkey(sizeof (CXairoWord::_ff),"wl"); int p=0;
	for (int i=k1;i<=k2;i++) {
		xara_lemma* l=GetDict(ak)->LookupEntry(i);
		int x;
		for (x=0;x<np;x++) {
			w=GetWL()->GetWordW(l->word[x]);
			if (!re[x]->Match(w)) break;
		}
		if (x<np) continue;
		CXairoWord::_ff f;
		f.i=i;
		f.f=l->freq;
		m_kdb.record(key,&f,sizeof f,1);
		p++;
		TestInterrupt();
	}
	m_kdb.closekey(key);
	for (int x=0;x<np;x++)
		mydelete re[x];
	mydelete [] re;
	if (p==0) {
		m_kdb.deletekey(key);
		return NULL;
	} else
		return mynew CXairoWordList(key,ak,this);
}

CXairoIWordList* CXairoServer::getFrequencyList(int np,const xairo_uchar** pattern,int maxn,int minf,int maxf,const xairo_uchar* ls)
{
	int ak=GetWL()->lemma_index(ls);
	CRegularExpression **re=mynew CRegularExpression *[np];;
	for (int x=0;x<np;x++) {
		UnicodeString us=GetLanguage()->MkNorm(pattern[x]);
		re[x] = mynew CRegularExpression(us.getTerminatedBuffer(),false);
		if (!re[x]->Parse()) throw mynew CXairoError(CXairoIError::ERR_REGEXP,L("Badly formed regexp"));
	}
	const xairo_uchar* w;
	CDBKey* key=m_kdb.newkey(sizeof (CXairoWord::_ff)); int p=0;
	if (maxf < 0) maxf=0x7fffffff;
	for (int i=GetDict(ak)->GetMaxFreq();i != -1;i=GetDict(ak)->GetNextFreq(i)) {
		xara_lemma* l=GetDict(ak)->LookupEntry(i);
		if (l->freq > maxf) continue;
		if (l->freq < minf) break;
		int x;
		for (x=0;x<np;x++) {
			w=GetWL()->GetWordW(l->word[x]);
			if (!re[x]->Match(w)) break;
		}
		if (x<np) continue;
		CXairoWord::_ff f;
		f.i=i;
		f.f=l->freq;
		m_kdb.record(key,&f,sizeof f,1);
		p++;
		if (p >= maxn && maxn > 0) break;
		TestInterrupt();
	}
	m_kdb.closekey(key);
	for (int x=0;x<np;x++)
		mydelete re[x];
	mydelete [] re;
	if (p==0) {
		m_kdb.deletekey(key);
		return 0;
	} else
		return mynew CXairoWordList(key,ak,this);
}


int CXairoServer::ffc(const void* v,const void* vv)
{
	CXairoWord::_ff* s=(CXairoWord::_ff*) v;
	CXairoWord::_ff* ss=(CXairoWord::_ff*) vv;
	return ss->f-s->f;
}

#define MAXFF 1000

CXairoIWordList* CXairoServer::getFilteredFrequencyList(int np,const xairo_uchar** pattern,int maxn,int minf,int maxf,const xairo_uchar* ls,const CXairoICorpusTextFilter* f,const CXairoIAddkeyFilter* ff)
{
	int ak=GetWL()->lemma_index(ls);
	CRegularExpression **re=mynew CRegularExpression *[np];;
	for (int x=0;x<np;x++) {
		UnicodeString us=GetLanguage()->MkNorm(pattern[x]);
		re[x] = mynew CRegularExpression(us.getTerminatedBuffer(),false);
		if (!re[x]->Parse()) throw mynew CXairoError(CXairoIError::ERR_REGEXP,L("Badly formed regexp"));
	}
	UnicodeString pre;
	pre=re[0]->Predict(10);
	int k1=-1; int k2=-1;
	if (!GetDict(ak)->GetLimits(pre.getTerminatedBuffer(),k1,k2)) return NULL;
	const xairo_uchar* w;
	CDBKey* key=m_kdb.newkey(sizeof (struct CXairoWord::_ff)); int p=0;
	if (maxf < 0) maxf=0x7fffffff;
	int npf=0;
	int* pf=NULL;
	if (f) npf=((CXairoCorpusTextFilter*) f)->OnOffMap(pf); 
	for (int i=k1;i<=k2;i++) 
	{
		CXairoWord::_ff fff;
		fff.i=i;
		fff.f=0;
		CXairoWord wd(ak,fff,this);
		int x;
		for (x=0;x<np;x++) {
			w=wd.getNorm(x);
			if (!re[x]->Match(w)) break;
		}
		if (x<np) continue;
		int freq=wd.getFilteredFrequencyMap(npf,pf,ff);
		if (freq > maxf) continue;
		if (freq < minf) continue;
		CXairoWord::_ff recf;
		recf.i=i;
		recf.f=freq;
		m_kdb.record(key,&recf,sizeof recf,1);
		p++;
		if (p > MAXFF) {
			throw new CXairoError(CXairoIError::ERR_LIMIT,L("Too many forms found"));
		}
//		if (p >= maxn && maxn > 0) break;
		TestInterrupt();
	}
	m_kdb.closekey(key);
	for (int x=0;x<np;x++)
		mydelete re[x];
	mydelete [] re;
// Now sort.
	int n=m_kdb.countrecords(key);
	CXairoWord::_ff* sa=new CXairoWord::_ff [n];
	for (int i=0;i<n;i++) {
		m_kdb.recorded(key,i,&sa[i]);
	}
	qsort(sa,n,sizeof (struct CXairoWord::_ff),CXairoServer::ffc);
	CDBKey* key1=m_kdb.newkey(sizeof (struct CXairoWord::_ff));
	if (maxn > 0 && n > maxn) n=maxn; 
	for (int i=0;i<n;i++) {
		m_kdb.record(key1,&sa[i],sizeof(sa[i]),1);
	}
	m_kdb.deletekey(key);
	m_kdb.closekey(key1);
	if (n==0) {
		m_kdb.deletekey(key1);
		return 0;
	} else
		return mynew CXairoWordList(key1,ak,this);
}

bool CXairoServer::Init(const CXairoIParameters* xp)
{
	m_bm=xp->bm;
	m_inth=xp->inth;
	strncpy(m_temppath,xp->temppath,_MAX_PATH);
	int k=strlen(m_temppath);
	if (m_temppath[k-1]=='/') m_temppath[k-1]=0;
	strncpy(m_datapath,xp->datapath,_MAX_PATH);
    XMLPlatformUtils::Initialize();
	m_param=mynew CParameters(this);
	char fp[_MAX_PATH];
	_fullpath(fp,xp->parameter_file,_MAX_PATH);
	if (!m_param->Read(fp)) return false;
	m_wl=mynew CWordList(this);
	if (!m_wl->Load()) return false;
	int nd=m_wl->GetNAddkeys()+m_wl->GetNLemmas();
	m_dict=mynew CDictionary* [nd+1];
	for (int i=0;i<nd+1;i++) {
		m_dict[i]=mynew CDictionary(this);
        if (!m_dict[i]->Load(i-1)) return false;
	}
	m_attdict=mynew CAttributeDictionary(m_wl,this);
	if (!m_attdict->Load()) return false;
	m_sb=mynew CSB(m_wl,this);
	m_sb->Load();
	m_locs=mynew CLocs(this);
	if (!m_locs->Load()) return false;
	m_thread=mynew CThread(this);
	if (!m_thread->Load()) return false;
	m_accels=mynew CAccelerators(this);
	if (!m_accels->Load()) return false;
	CXairoGrammar* g=mynew CXairoGrammar(this);
	if (!g->Init()) return false;
	m_grammar=g;
	m_sf=mynew CStreamFactory(this);
//	m_sf->MakeDTD();
	m_em=mynew CEntityManager(this);
	if (!m_em->Load()) return false;
	m_ds=(CAtomStream*) m_sf->MakeAtomAStream(L("document"),XAIRA_URI,it_global_attribute_value);
	m_es=(CAtomStream*) m_sf->MakeAtomAStream(L(""),L(""),it_enddoc);
	m_us=(CAtomStream*) m_sf->MakeAtomAStream(L("unit"),XAIRA_URI,it_global_attribute_value);
	m_ss=(CAtomStream*) m_sf->MakeAtomAStream(L("scope"),XAIRA_URI,it_global_attribute_value);
	m_sl=(CAtomStream*) m_sf->MakeAtomAStream(L("locale"),XAIRA_URI,it_global_attribute_value);
	MakeDocTab();
	const char* locale="";
	if (m_sl) {
		xara_loc* l=m_sl->SetCursor(0);
		if (l) {
			locale=m_wl->GetWord(l->nw);
		}
	}
	m_language=mynew CLanguage(this);
	if (!m_language->Load(locale)) return false;
	if (!m_kdb.Init(this)) return false;
//	_CrtSetAllocHook(YourAllocHook);
 	return true;
}

void CXairoServer::TestInterrupt()
{
	if (m_inth==NULL) return;
	bool b=m_inth->testInterrupt();
	if (b)
		throw mynew CXairoError(CXairoIError::ERR_INT,L("Interrupt"));
}

CXairoISolution* CXairoServer::solve(const xairo_uchar* xql)
{
	return solveInDivision(xql,NULL);
}

CXairoISolution* CXairoServer::solveInDivision(const xairo_uchar* xql,const CXairoGrammar::Partition* part)
{
	int ip=0;
	const CXairoGrammar::Partition* p2=NULL;
	CAtomStream* sd=GetDocumentStream();
	if (part) p2=part->GetParent(ip);
	const xairo_uchar* err=NULL; CXairoError* e=NULL;
	CStream* s=m_sf->BuildXMLQuery(xql,err);
	if (err) {
		e=mynew CXairoError(CXairoIError::ERR_CQL,err);
		throw(e);
	}
	if (s==NULL) return NULL;
	int k=-1;
	CDBKey* key=m_kdb.newkey(sizeof(xara_loc),"sol");
	int p=0;
	int seqd=0;
	bool bInc=true;
	while (s->InitVars(),s->Advance(k,true)) {
		if (part) {
            if (s->m_seq > seqd) {
				sd->Advance(s->m_seq,true);
				seqd=sd->m_seq;
				bInc=p2->Test(sd->GetCursor()-1,ip);
			}
			if (!bInc) {
				k=s->m_seq+1;
				continue;
			}
		}
		xara_loc l;
		l.fi=s->m_fi;
		l.len=s->m_len;
		l.nw=s->m_nseq;  // not nw, which isn't needed to extract hits
		l.off=s->m_off;
		l.seq=s->m_seq;
		l.sb=s->m_nw; // sigh, yes it is
		m_kdb.record(key,&l,sizeof(l),1);
		k=s->m_seq+1;
		p++;
		TestInterrupt();
	}
	m_kdb.closekey(key);
	if (p==0) {
		m_kdb.deletekey(key);
		return NULL;
	} else 
		return mynew CXairoSolution(key,part,this);
}


 
CXairoICorpusTextList* CXairoServer::getCorpusTextList()
{
	return mynew CXairoCorpusTextList(this);
}

void CXairoServer::BootLog(const xairo_uchar* format,...)
{
    va_list(args); xairo_uchar msg[1000];
    if (m_bm == NULL) return;
    va_start(args, format);
    _vsnwprintf(msg,1000,format,args);
 	m_bm->Log(msg);
}

const xairo_uchar* CXairoServer::getVersion()
{
	xairo_uchar v[100]; char vv[100];
#ifndef _WIN32
	strncpy(vv,"Xaira-server version " VERSION " built " __DATE__,100);
#else
	_snprintf(vv,100,"Xaira-server version %d.%d.%d.%d built %s",XV1,XV2,XV3,XV4,__DATE__);
#endif
	mbstowcs(v,vv,100);
	return _wcsdup(v);
}

int CXairoServer::getWordCount()
{
	return GetDict(-1)->GetNUsages();
}

struct _s {int i;int f;};

int CXairoServer::fc(const void* v,const void* vv)
{
	_s* s=(_s*) v;
	_s* ss=(_s*) vv;
	return ss->f-s->f;
}
	
CXairoIAttributeValueList* CXairoServer::getAttributeValueList(const CXairoIGrammar::Attribute* a)
{
	xara_streambase *sb; int k=0;
	int p; int t=0;
	p=GetAttDict()->Lookup(a->getName(),a->getElt(),a->getNamespace()->getUrl());
	int p0=p; 
	while (p0 != -1) {
		sb=GetSB()->GetSB(p0);
		p0=sb->next;
		t++;
	}
	if (t==0) return NULL;
	_s* s=mynew _s [t]; 
	if (s==NULL) throw mynew CXairoError(CXairoIError::ERR_MEMORY,L("Out of memory sorting values"));
	int tt=0;
	p0=p; 
	while (p0 != -1) {
		sb=GetSB()->GetSB(p0);
		s[tt].i=p0;
		s[tt].f=sb->count;
		p0=sb->next;
		tt++;
	}
	qsort(s,tt,sizeof(_s),fc);
	CDBKey* key=GetDB()->newkey(sizeof (int)); 
	for (int i=0;i<t;i++) {
		GetDB()->record(key,&s[i].i,sizeof (int),1);
	}
	GetDB()->closekey(key);
	mydelete [] s;
	if (t==0) {
		GetDB()->deletekey(key);
		return 0;
	} else
		return mynew CXairoAttributeValueList(key,a->getCodebook(),this);
}

CXairoIAddkeyValueList* CXairoServer::getAddkeyValueList(const CXairoIGrammar::Addkey* a)
{
	int nd=((CXairoGrammar::Addkey*) a)->GetDic();
	CDictionary* d=GetDict(nd);
	int t=d->Count();
	if (t==0) {
		return NULL;
	} else
		return mynew CXairoAddkeyValueList(nd,this);
}

const xairo_uchar* CXairoServer::getCorpusName()
{
	return GetParam()->Param(PARAM_NAME);
}

const xairo_uchar* CXairoServer::getDefaultLanguage()
{
	return _wcsdup(UnicodeString(GetLanguage()->GetDefault()).getTerminatedBuffer());
}

#define MAXID 100

const xairo_uchar* CXairoServer::MakeId(const xairo_uchar* fmt,int id)
{
#ifdef _WIN32
	int x=_scwprintf(fmt,id);
	xairo_uchar* w=mynew xairo_uchar [x+1];
	swprintf(w,fmt,id);
	return w;
#else
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* c=mynew char [MAXID+1];
	snprintf(c,MAXID+1,f2,id);
	mydelete [] f2;
	xairo_uchar* w=mynew xairo_uchar [strlen(c)+1];
	mbstowcs(w,c,strlen(c)+1);
	mydelete [] c;
	return w;
#endif
}


const xairo_uchar* CXairoServer::MakeId(const xairo_uchar* fmt,int id,int idd)
{
#ifdef _WIN32
	int x=_scwprintf(fmt,id,idd);
	xairo_uchar* w=mynew xairo_uchar [x+1];
	swprintf(w,fmt,id,idd);
	return w;
#else
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* c=mynew char [MAXID+1];
	snprintf(c,MAXID+1,f2,id,idd);
	mydelete [] f2;
	xairo_uchar* w=mynew xairo_uchar [strlen(c)+1];
	mbstowcs(w,c,strlen(c)+1);
	mydelete [] c;
	return w;
#endif
}

const xairo_uchar* CXairoServer::MakeId(const xairo_uchar* fmt,int id,int idd,int iddd)
{
#ifdef _WIN32
	int x=_scwprintf(fmt,id,idd,iddd);
	xairo_uchar* w=mynew xairo_uchar [x+1];
	swprintf(w,fmt,id,idd,iddd);
	return w;
#else
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* c=mynew char [MAXID+1];
	snprintf(c,MAXID+1,f2,id,idd,iddd);
	mydelete [] f2;
	xairo_uchar* w=mynew xairo_uchar [strlen(c)+1];
	mbstowcs(w,c,strlen(c)+1);
	mydelete [] c;
	return w;
#endif
}

int CXairoServer::ScanId(const xairo_uchar* fmt,const xairo_uchar* id)
{
#ifdef _WIN32
	int k;
	swscanf(id,fmt,&k);
	return k;
#else
	int k;
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* i2=mynew char[wcslen(id)+1];
	wcstombs(i2,id,wcslen(id)+1);
	sscanf(i2,f2,&k);
	mydelete [] f2; 
	mydelete [] i2;
	return k;
#endif
}

void CXairoServer::ScanId(const xairo_uchar* fmt,const xairo_uchar* id,int* k)
{
#ifdef _WIN32
	swscanf(id,fmt,k);
	return;
#else
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* i2=mynew char[wcslen(id)+1];
	wcstombs(i2,id,wcslen(id)+1);
	sscanf(i2,f2,k);
	mydelete [] f2; 
	mydelete [] i2;
	return;
#endif
}

void CXairoServer::ScanId(const xairo_uchar* fmt,const xairo_uchar* id,int* k,int* kk)
{
#ifdef _WIN32
	swscanf(id,fmt,k,kk);
	return;
#else
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* i2=mynew char[wcslen(id)+1];
	wcstombs(i2,id,wcslen(id)+1);
	sscanf(i2,f2,k,kk);
	mydelete [] f2; 
	mydelete [] i2;
	return;
#endif
}

void CXairoServer::ScanId(const xairo_uchar* fmt,const xairo_uchar* id,int* k,int* kk,int* kkk)
{
#ifdef _WIN32
	swscanf(id,fmt,k,kk,kkk);
	return;
#else
	char* f2=mynew char[wcslen(fmt)+1];
	wcstombs(f2,fmt,wcslen(fmt)+1);
	char* i2=mynew char[wcslen(id)+1];
	wcstombs(i2,id,wcslen(id)+1);
	sscanf(i2,f2,k,kk,kkk);
	mydelete [] f2; 
	mydelete [] i2;
	return;
#endif
}

CXairoIExtendedForm* CXairoServer::getUsage(unsigned int i)
{
	int sb=GetThread()->GetThread(i);
	return mynew CXairoExtendedForm(sb,this);
}

CXairoIAddkeyFilter* CXairoServer::getAddkeyFilter(const xairo_uchar* akey,int nvals,const xairo_uchar** vals)
{
	CDBKey* key=GetDB()->newkey(sizeof(int));
	int x=GetWL()->LookupW(akey);
	GetDB()->record(key,&x,sizeof(int),1);
	for (int i=0;i<nvals;i++) {
		int x=GetWL()->LookupW(vals[i]);
		GetDB()->record(key,&x,sizeof(int),1);
	}
	GetDB()->closekey(key);
	return new CXairoAddkeyFilter(key,this);
}

void* CXairoServer::deserialize(const xairo_uchar* c)
{
	if (wcsncmp(c,L("sol"),3)==0)
		return mynew CXairoSolution(c,this);
	else if (wcsncmp(c,L("ctf"),3)==0)
		return mynew CXairoCorpusTextFilter(c,this);
	else 
		return NULL;

}


void CXairoServer::MakeDocTab()
{
	m_nd=m_ds->Count();
	m_doctab=mynew _doc [m_nd];
	for (int i=0;i<m_nd;i++) {
		xara_loc* l=m_ds->SetCursor(i);
		m_doctab[i].seq=l->seq;
	}
}

int CXairoServer::SeqToDoc(int s)
{
	for (int i=0;i<m_nd;i++) {
		if (s < m_doctab[i].seq) return i-1;
	}
	return m_nd-1;
}
	
void CXairoServer::Debug(const char* msg,...)
{
	time_t t;
	time(&t);
	const char* b=ctime(&t);
	char path[_MAX_PATH];
	strcpy(path,GetTempPath());
	strcat(path,"/xaira_dbg.txt");
	FILE* f=fopen(path,"a");
	fprintf(f,b);
	fprintf(f,": ");
	va_list list; 
	va_start(list,msg);
	vfprintf(f,msg,list);
	va_end(list);
	fprintf(f,"\n",msg);
	fclose(f);
}






XAIRO_NAMESPACE_END
