#pragma once

XAIRO_NAMESPACE_BEGIN

enum stream_type {TY_NULL,TY_ATOM,TY_OR,TY_NEG,TY_ALL,TY_SEQ,TY_FILE,TY_AND,TY_PROD,
	TY_LIMIT,TY_SUBCORPUS,TY_ID,TY_IDC,TY_BPROD,TY_CASE,TY_MEM};

enum lim_type {LT_SGML,LT_NULL,LT_INT,LT_XSGML};

#define MAXOPS 100
#define NVARS 16

struct CStreamVar {
   bool      bSet;
   int       val;
   CStream *setter;
   CStream* stream;
   CStreamVar():bSet(false),val(0),setter(NULL) {}
};

class CStream
{
	friend class CStreamFactory;
public:
// by design, properties set by the stream factory are public. Properties used in Advance 
// must be private
	CStream(CXairoServer*);
	~CStream(void);
	virtual bool Advance(int,bool,int=-1)=0;
	virtual bool Flatten() {return true;}
	virtual bool IsWildcard() {return false;}
	virtual bool LimitStream(lim_type,CStream*,CStream*,int) {return false;}
	virtual int GetParent() {return -1;}
	CStreamVar* GetVar(int i) {return &(m_vars)[i];}
	xara_loc* GetLocBySeq(int);

	void InitVars();
	int m_seq, m_nseq;
	int m_nw, m_wc;
	int m_fi, m_off;
	int m_len;
private:
	CStreamVar* m_vars;
protected:
	CXairoServer* m_server;
	bool gtr(int,int,bool);
};

class CAtomStream:public CStream
{
public:
	CAtomStream(CXairoServer* s):CStream(s),m_cache_b(false) {}
	int m_first, m_last;
	typ m_wtype;
	const xairo_uchar* m_word, *m_xword;
	bool Advance(int,bool,int=-1);
	void SetPos(int);
	int Count() {return m_last-m_first;}
	xara_loc* SetCursor(int);
	xara_loc* SetCursor() {return SetCursor(m_c);}
	int GetCursor();
	int Count(const CXairoGrammar::Partition*);
	int Count(int,int*);
	int CountMap(int,int*);
	int m_accel;
	int GetParent();
private:
	bool AdvanceForwards(int,int);
	bool AdvanceBackwards(int,int);
	int m_c;
	bool m_cache_b;
	int m_cache_i, m_cache_seq, m_cache_target;
};

class CMulStream:public CStream
{
public:
	CMulStream(CXairoServer* s):CStream(s) {}
	int m_ops;
	CStream** m_op;
	virtual bool Advance(int,bool,int=-1)=0;
protected:
	void CopyPos(int);
};

class COrStream:public CMulStream
{
public:
	COrStream(CXairoServer* s):CMulStream(s) {}
	bool Advance(int,bool,int);
};

class CAndStream:public CMulStream
{
public:
	CAndStream(CXairoServer* s):CMulStream(s) {}
	bool Advance(int,bool,int=-1);
	int GetParent();
};

class CSeqStream:public CMulStream
{
public:
	CSeqStream(CXairoServer* s):CMulStream(s) {}
	bool Advance(int,bool,int=-1);
	bool Flatten();
};

class CKeyStream:public CStream
{
public:
	CKeyStream(CXairoServer* s):CStream(s) {}
	CDBKey *m_key;
	bool Advance(int,bool,int=-1);
private:
	bool AdvanceForwards(int);
	bool AdvanceBackwards(int);
	void SetPos(xara_loc*);
};

class CMemStream:public CStream
{
public:
	CMemStream(CXairoServer* s):CStream(s) {}
	int m_count;
	xara_loc* m_buffer;
	bool Advance(int,bool,int=-1);
private:
	bool AdvanceForwards(int);
	bool AdvanceBackwards(int);
	void SetPos(xara_loc*);
};

class CAllStream:public CStream
{
public:
	CAllStream(CXairoServer* s):CStream(s) {}
	bool Advance(int,bool,int=-1);
	bool IsWildcard() {return true;}
};

class CNegStream:public CStream
{
public:
	CNegStream(CXairoServer* s):CStream(s) {}
	CStream* m_op;
	bool Advance(int,bool,int=-1);
};

class CLimitStream:public CStream
{
public:
	CLimitStream(CXairoServer* s):CStream(s) {}
	CStream* m_op;
	CStream *m_limit1;
	CAtomStream *m_limit2;
	int m_nlimit;
	lim_type m_limtype;
	bool Advance(int,bool,int=-1);
};

class CProdStream:public CMulStream
{
public:
	CProdStream(CXairoServer* s):CMulStream(s) {}
	CStream *m_limit1, *m_limit2;
	lim_type m_limtype;
	int m_limint;
	bool LimitStream(lim_type,CStream*,CStream*,int);
	bool Advance(int,bool,int=-1);
};

class CBProdStream:public CMulStream
{
public:
	CBProdStream(CXairoServer* s):CMulStream(s) {}
	CStream *m_limit1, *m_limit2;
	lim_type m_limtype;
	int m_limint;
	bool LimitStream(lim_type,CStream*,CStream*,int);
	bool Advance(int,bool,int=-1);
};

class CIdStream:public CStream
{
public:
	CIdStream(CXairoServer* s):CStream(s) {}
	CStream* m_op;
	int m_kvar;
	bool Advance(int,bool,int=-1);
	const xairo_uchar* m_elt;
	const xairo_uchar* m_att;
	const xairo_uchar* m_url;
	const xairo_uchar* m_aurl;

};

class CIDCStream:public CStream
{
public:
	CIDCStream(CXairoServer* s):CStream(s) {}
	CStream* m_op;
	int m_kval;
	bool Advance(int,bool,int=-1);
};

XAIRO_NAMESPACE_END
