#include "SServer.h"
#include "StreamFactory.h"
#include "Stream.h"
#include "WordList.h"
#include "SB.h"
#include "Dictionary.h"
#include "Grammar.h"
#include "language.h"
#include "Locs.h"
#include "RegularExpression.h"
#include "AttributeDictionary.h"
#include "Accelerators.h"
#include "XairoSolveError.h"

XAIRO_NAMESPACE_BEGIN

const char* TempPath();


//"<!ENTITY % cqlop \"seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope\">\n\

static const char* dtd_txt=
"<!ELEMENT cql (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)>\n\
<!ELEMENT seq (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)*>\n\
<!ELEMENT or (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)*>\n\
<!ELEMENT and (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)*>\n\
<!ELEMENT prod (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)*>\n\
<!ELEMENT bprod (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)*>\n\
<!ELEMENT neg (seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope)>\n\
<!ELEMENT lemma (#PCDATA)>\n\
<!ATTLIST lemma\n\
	ls CDATA \"\">\n\
<!ELEMENT pos (word,poscode*)>\n\
<!ELEMENT phrase (#PCDATA)>\n\
<!ATTLIST phrase\n\
	case CDATA \"no\">\n\
<!ELEMENT element (attribute*)>\n\
<!ATTLIST element\n\
	end CDATA \"no\"\n\
	name CDATA #IMPLIED\n\
	ns CDATA \"\">\n\
<!ELEMENT pattern (#PCDATA)>\n\
<!ATTLIST pattern\n\
	case CDATA \"no\">\n\
<!ELEMENT scope ((seq | or | and | prod | bprod | neg | lemma | pos | phrase | element | pattern | scope),(element|span))>\n\
<!ELEMENT word (#PCDATA)>\
<!ELEMENT poscode (#PCDATA)>\n\
<!ATTLIST poscode\n\
	key CDATA #REQUIRED>\n\
<!ELEMENT attribute (#PCDATA)>\n\
<!ATTLIST attribute\n\
	name CDATA #REQUIRED\n\
	ns  CDATA \"\"\n\
	var CDATA \"no\"\n\
	pattern CDATA \"no\"\n\
	value CDATA #IMPLIED>\n\
<!ELEMENT span EMPTY>\n\
<!ATTLIST span\n\
	size   CDATA  #REQUIRED >\n";


//#define NDTDT ((sizeof dtd_txt)/sizeof (char*))


CStreamFactory::CStreamFactory(CXairoServer* s):m_server(s)
{

}

CStreamFactory::~CStreamFactory()
{
}



//extern "C" void FlattenStream(struct stream *);


enum _tag {TAG_SEQ,TAG_OR,TAG_AND,TAG_PROD,TAG_BPROD,TAG_NEG,TAG_LEMMA,TAG_LEMMAS,
	TAG_FORM,TAG_POS,TAG_PHRASE,TAG_WORD,TAG_ELEMENT,
	TAG_PATTERN,TAG_SCOPE,TAG_BPRODX,TAG_MAX};

// now we have the DTD there is no need for the arity check

static struct tagtab {
	const xairo_uchar* name;
	int arity;
} tags[TAG_MAX] =
{
	{L("seq"),0},
	{L("or"),0},
	{L("and"),0},
	{L("prod"),0},
	{L("bprod"),0},
	{L("neg"),1},
	{L("lemma"),1},
	{L("lemmas"),0},
	{L("form"),1},
	{L("pos"),0},
	{L("phrase"),0},
	{L("word"),1},
	{L("element"),0},
	{L("pattern"),1},
	{L("scope"),2},
	{L("bprodx"),2}
};

class ErrH:public DOMErrorHandler
{
public:
	ErrH():bErr(false) {}
	bool handleError(const DOMError &);	
	bool bErr; XMLFileLoc row, col;
	const xairo_uchar* msg;
	const xairo_uchar* uri;
};


bool ErrH::handleError(const DOMError& e)
{
	msg=_wcsdup(e.getMessage());
	DOMLocator* l=e.getLocation();
	row=l->getLineNumber();
	col=l->getColumnNumber();
	uri=_wcsdup(l->getURI());
	bErr=true;
	return false;
}


//#define DOCTYPE L("<!DOCTYPE cql SYSTEM \"%S\">")
#define DOCTYPEI L("<!DOCTYPE cql [%S]>")


CStream* CStreamFactory::BuildXMLQuery(const xairo_uchar* xql,const xairo_uchar*& err)
{
//	xairo_uchar* buffer=new xairo_uchar [wcslen(xql)+wcslen(DOCTYPE)+strlen(m_dtdpath)+10];
	xairo_uchar* buffer=new xairo_uchar [wcslen(xql)+wcslen(DOCTYPEI)+strlen(dtd_txt)+10];
	DOMImplementationLS* di=(DOMImplementationLS*) DOMImplementation::getImplementation();
	DOMLSParser* parser=di->createLSParser(DOMImplementationLS::MODE_SYNCHRONOUS,NULL);
	ErrH eh;
//	d.loadGrammar(dtd,Grammar::DTDGrammarType,true);
//	swprintf(buffer,DOCTYPEI,dtd_txt);
//	int ke=wcslen(buffer);
	wcscpy(buffer,xql);	
	MemBufInputSource is((const unsigned char* const) buffer,wcslen(buffer)*2,L("cql")); 
	is.setEncoding(L("UTF-16"));
	DOMConfiguration* dcfg=parser->getDomConfig();
	dcfg->setParameter(L("error-handler"),&eh);
	dcfg->setParameter(XMLUni::fgXercesSchema,true);
	dcfg->setParameter(XMLUni::fgDOMNamespaces,true);
//	d.setValidationScheme(XercesDOMParser::Val_Always);
//	parser->setFeature(XMLUni::fgDOMNamespaces,true);
//    parser->setFeature(XMLUni::fgXercesSchema,true);
	dcfg->setParameter(L("validate"), true);
	xairo_uchar sl[_MAX_PATH*2]; sl[0]=0;
//	wcscpy(sl,L("http://xaira.org/ns/1.0 ")); 
	const char* p=m_server->GetDataPath();
	for (int i=0;i<(int) strlen(p);i++) {
		xairo_uchar x[2];
		if (p[i]==32)
			wcscat(sl,L("%20"));
		else {
			x[0]=p[i];
			x[1]=0;
			wcscat(sl,x);
		}
	}
	wcscat(sl,L("/cql.xsd"));
//	dcfg->setParameter(L("schema-type"),L"http://www.w3.org/2001/XMLSchema");
	dcfg->setParameter(XMLUni::fgXercesSchemaExternalNoNameSpaceSchemaLocation,sl);
//	dcfg->setParameter(L("schema-location"),sl);
	dcfg->setParameter(L("comments"),false);
	dcfg->setParameter(L("element-content-whitespace"),false);
	DOMDocument* doc=NULL;
	doc=parser->parse(new Wrapper4InputSource(&is));
	if (eh.bErr) { // need uri here
		xairo_uchar m[1000];
		_snwprintf(m,1000,L("Parse error in %s at %d,%d: %s\n"),eh.uri,(int) eh.row,(int) eh.col,eh.msg);		
		err=_wcsdup(m);
		return NULL;
	}
	DOMElement* r=doc->getDocumentElement();
	if (wcscmp(r->getTagName(),L("cql")) != 0) {
		err=BXQerr(L("Bad document type"));
		return NULL;
	}
	DOMNodeList* l=r->getChildNodes();
	if (l->getLength() != 1) {
		err=L("<cql> tag must have just one child");
		return NULL;
	}
	DOMNode* n=l->item(0);
	if (n->getNodeType() != DOMNode::ELEMENT_NODE) 
	{
		err=BXQerr(L("<cql> content is not a query"));
        return NULL;
	}
	m_vars=new CStreamVar [NVARS];
	CStream* res=BuildXMLQuery((DOMElement*) n,err);
	delete parser;
	return res;
}

const xairo_uchar* CStreamFactory::BXQerr(const xairo_uchar* fmt,...)
{
#define MAXERR 1000
	xairo_uchar* s=new xairo_uchar [MAXERR];
    va_list(args); 
    va_start(args, fmt);
    _vsnwprintf(s,MAXERR,fmt,args);
	return s;
}

CStream* CStreamFactory::BuildXMLQuery(DOMElement* e,const xairo_uchar*& err)
{
	const xairo_uchar* tag=e->getTagName(); _tag it;
	int ix;
	for (ix=0;ix<TAG_MAX;ix++) {
		if (wcscmp(tag,tags[ix].name)==0) break;
	}
	if (ix==TAG_MAX) {
		err=BXQerr(L("Unknown tag %s"),tag);
		return NULL;
	}
	it = (_tag) ix;
	DOMNodeList* l=e->getChildNodes();
	if (tags[it].arity && tags[it].arity != l->getLength()) {
		err=BXQerr(L("Wrong number of arguments for %s"),tag);
		return NULL;
	}
	switch (it) {
		stream_type sty;
		case TAG_PHRASE:
			{
				const xairo_uchar* s=e->getTextContent();
				bool bCase=wcscmp(e->getAttribute(L("case")),L("yes"))==0;
				UnicodeString us=m_server->GetLanguage()->MkNorm(s);
				UnicodeString* ss; 
				int n=XTokenise(us,ss);
				int ns=0;
				if (n==0) {
					err=BXQerr(L("Empty phrase"));
					return NULL;
				}
				if (n==1) 
					if (ss[ns]==L("_")) {
						CStream* st=(CAllStream*) GrabStream(TY_ALL);
						if (st==NULL) {
							err=BXQerr(L("No more streams"));
							return NULL;
						}
//						st->m_wtype=it_content;
						return st; 
					} else
						return MakeWordStream(ss[ns].getTerminatedBuffer(),bCase);
				else 
				{
					CSeqStream* st=(CSeqStream*) GrabStream(TY_SEQ);
					if (st==NULL) {
						err=BXQerr(L("No more streams"));
						return NULL;
					}
					st->m_ops=n;
					if (n>=MAXOPS) {
						err=BXQerr(L("Too many substreams"));
						FreeStream(st);
						return NULL;
					}
					st->m_op=new CStream* [n];
					int jj=0;
					while (ns < n) {
						if (ss[ns]==L("_")) {
							st->m_op[jj]=(CAllStream*) GrabStream(TY_ALL);
							if (st->m_op[jj]==NULL) {
								err=BXQerr(L("No more streams"));
							}
//							st->m_op[jj]->wtype=it_content;
						} else
							st->m_op[jj]=MakeWordStream(ss[ns].getTerminatedBuffer(),bCase);
						if (st->m_op[jj]==NULL) {
							FreeStream(st);
							return NULL;
						}
						jj++;
						ns++;
					}
					if (!st->Flatten()) {
						err=BXQerr(L("Illegal sequence"));
						return NULL;
					}
					return st;
				}
			}
		case TAG_ELEMENT:
			{
				bool bE=wcscmp(e->getAttribute(L("end")),L("yes"))==0;
				const xairo_uchar* name=e->getAttribute(L("name"));
				if (wcslen(name)==0) name=L("0");
				const xairo_uchar* ns=e->getAttribute(L("ns"));
				int n=l->getLength();
				if (n>=MAXOPS) {
					err=BXQerr(L("Too many substreams"));
					return NULL;
				}
				_attstream* as=new _attstream [n];
				int i;
				for (i=0;i<n;i++) {
					DOMNode* nd=l->item(i);
					if (nd->getNodeType() != DOMNode::ELEMENT_NODE) break;
					DOMElement* ne=(DOMElement*) nd;
					if (wcscmp(ne->getTagName(),L("attribute")) != 0) break;
					as[i].name=ne->getAttribute(L("name"));
					as[i].ns=ne->getAttribute(L("ns"));
					as[i].bVar=wcscmp(ne->getAttribute(L("var")),L("yes"))==0;
					as[i].bPatt=wcscmp(ne->getAttribute(L("pattern")),L("yes"))==0;
					as[i].extra=ne->getAttribute(L("value"));
					as[i].value=nd->getTextContent();
				}
				if (i<n) {
					err=BXQerr(L("Unrecognised attribute specifier"));
					return NULL;
				}
				if (bE) 
					return MakeAtomStream(name, ns, it_end);
				else
					return MakeTagStream(name, ns, n, as);
			}
		case TAG_PATTERN:
			{
				bool bCase=wcscmp(e->getAttribute(L("case")),L("yes"))==0;
				const xairo_uchar* p=e->getTextContent();
// change to parse expression here for better error handling
				return MakeRegExpStream((xairo_uchar*) p,bCase); // entry from XCQL only
			}
//		case TAG_FORM: // withdrawn - use pos
//			{
//				int i=_wtoi(e->getAttribute(L"index"));
//				const xairo_uchar* w=e->getTextContent();
//      	    return MakeFormStream(w,i);
//		}
		case TAG_POS:
			{
				DOMNode* n=l->item(0);
				if (n->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					err=BXQerr(L("Bad word arg in pos"));
					return NULL;
				}
				DOMElement* ne=(DOMElement*) n;
				if (wcscmp(ne->getTagName(),L("word")) != 0) 
				{
					err=BXQerr(L("Bad word arg in pos"));
					return NULL;
				}
				const xairo_uchar* word=ne->getTextContent();
				const xairo_uchar** keys=new const xairo_uchar* [l->getLength()-1];
				const xairo_uchar** tags=new const xairo_uchar* [l->getLength()-1];
				int jj=0;
				for (unsigned j=1;j<l->getLength();j++) {
					DOMNode* nn=l->item(j);
					if (nn->getNodeType() != DOMNode::ELEMENT_NODE) 
					{
						err=BXQerr(L("Bad postag arg in pos"));
						return NULL;
					}
					DOMElement* nne=(DOMElement*) nn;
					if (wcscmp(nne->getTagName(),L("poscode")) != 0) 
					{
						err=BXQerr(L("Bad postag arg in pos"));
						return NULL;
					}
					keys[jj]=nne->getAttribute(L("key"));
					tags[jj]=nne->getTextContent();
					jj++;
				}		
				CStream* ss;
				if (wcscmp(word,L("_"))==0) {
					if (jj==0) {
						err=BXQerr(L("In pos the word _ must have at least one tag"));
						return NULL;
					}
					ss=MakeAllPosStream(jj,tags,keys);
				} else {
					UnicodeString us=m_server->GetLanguage()->MkNorm(word);
					ss=MakePosStream(us.getTerminatedBuffer(), jj,tags, keys);
				}
				delete [] keys; delete [] tags;
				return ss;
			}
		case TAG_LEMMA:
			{
				const xairo_uchar* w;
				w=e->getTextContent();
				UnicodeString us=m_server->GetLanguage()->MkNorm(w);
				const xairo_uchar* ls=e->getAttribute(L("ls"));
				return MakeLemmaStream(us.getTerminatedBuffer(),ls);
			}
		case TAG_LEMMAS:
			{
				DOMNodeList* nl=e->getElementsByTagName(L("word"));
				int nw=nl->getLength();
				UnicodeString* ws=new UnicodeString [nw];
				for (int j=0;j<nw;j++) {
					DOMElement* ee=(DOMElement*) nl->item(j);
					ws[j]=ee->getTextContent();
					ws[j]=m_server->GetLanguage()->MkNorm(ws[j]);
				}
				const xairo_uchar* ls=e->getAttribute(L("ls"));
				return MakeLemmasStream(nw,ws,ls);
			}
		case TAG_NEG:
			{
				DOMNode* n=l->item(0);
				if (n->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					err=BXQerr(L("Bad arg in neg"));
					return NULL;
				}
				CStream* st1=BuildXMLQuery((DOMElement*) n,err);
				if (st1 == NULL) 
					return NULL;
				CNegStream* st=(CNegStream*) GrabStream(TY_NEG);
				if (st==NULL) {
					err=BXQerr(L("No more streams"));
					return NULL;
				}
				st->m_op = st1;
				return st;
			}
		case TAG_SCOPE:
			{
				DOMNode* n1=l->item(0);
				if (n1->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					err=BXQerr(L("Bad first arg in scope"));
					return NULL;
				}
				CStream* st1=BuildXMLQuery((DOMElement*) n1,err);
				if (st1 == NULL) 
					return NULL;
				DOMNode* n2=l->item(1);
				if (n2->getNodeType() != DOMNode::ELEMENT_NODE) 
				{
					err=BXQerr(L("Bad second arg in scope"));
					return NULL;
				}
				DOMElement* ne2=(DOMElement*) n2;
				lim_type tp; int tn; CStream *s1; CAtomStream *s2;
				if (wcscmp(ne2->getTagName(),L("element"))==0) {
					bool bE=wcscmp(ne2->getAttribute(L("end")),L("yes"))==0;
					const xairo_uchar* name=ne2->getAttribute(L("name"));
					const xairo_uchar* ns=ne2->getAttribute(L("ns"));
					if (bE) {
						s1 = MakeAtomStream(name, ns, it_end);
						s2 = (CAtomStream*) MakeAtomStream(name, ns, it_start);
						tp = LT_XSGML;
					} else {
		                s1 = BuildXMLQuery(ne2,err);                     
						s2 = (CAtomStream*) MakeAtomStream(name, ns, it_end);
						tp = LT_SGML;
					}
					if (s1==NULL || s2==NULL) return NULL;
					tn=0;
				} else if (wcscmp(ne2->getTagName(),L("span"))==0) {
					tn=_wtoi(ne2->getAttribute(L("size")));
					s1=s2=NULL;
					tp = LT_INT;
				} else {
					err=BXQerr(L("Bad second argument in scope"));
					return NULL;
				}
				if (st1->LimitStream(tp,s1,s2,tn))
					return st1;
				else {
					CLimitStream* st=(CLimitStream*) GrabStream(TY_LIMIT);
					st->m_limtype=tp;
					st->m_limit1=s1;
					st->m_limit2=s2;
					st->m_nlimit=tn;
					st->m_op=st1;
                    return st;
				}
			}
		case TAG_AND: 
			sty=TY_AND; goto cmn;
		case TAG_OR:
			sty=TY_OR; goto cmn;
		case TAG_PROD:
			sty=TY_PROD; goto cmn;
		case TAG_BPROD:
			sty=TY_BPROD; goto cmn;
		case TAG_BPRODX:
			{
				int n=l->getLength(); int i;
				CMulStream* st=(CMulStream*) GrabStream(TY_OR);
				if (st==NULL) {
					err=BXQerr(L("No more streams"));
					return NULL;
				}
				st->m_ops=2;
				st->m_op=new CStream* [2];
				CMulStream* st0=(CMulStream*) GrabStream(TY_PROD);
				st->m_op[0]=st0;
				st0->m_ops=n;
				st0->m_op=new CStream* [n];
				for (i=0;i<n;i++) {
					DOMNode* n=l->item(i);
					if (n->getNodeType() != DOMNode::ELEMENT_NODE) break;
					st0->m_op[i]=BuildXMLQuery((DOMElement*) n,err);
					if (st0->m_op[i]==NULL) break;
				}
				if (i < n) {
					FreeStream(st);
					return NULL;
				}
				CMulStream* st1=(CMulStream*) GrabStream(TY_PROD);
				st->m_op[1]=st1;
				st1->m_ops=n;
				st1->m_op=new CStream* [n];
				for (i=0;i<n;i++) {
					DOMNode* no=l->item(n-i-1);
					if (no->getNodeType() != DOMNode::ELEMENT_NODE) break;
					st1->m_op[i]=BuildXMLQuery((DOMElement*) no,err);
					if (st1->m_op[i]==NULL) break;
				}
				if (i < n) {
					FreeStream(st);
					return NULL;
				}
				if (!st->Flatten()) {
					err=BXQerr(L("Illegal sequence"));
					return NULL;
				}
			return st;
			}


		case TAG_SEQ:
			sty=TY_SEQ;
cmn:
			{
				int n=l->getLength(); int i;
				if (n==1) {
					DOMNode* n=l->item(0);
					if (n->getNodeType() != DOMNode::ELEMENT_NODE) break;
					return BuildXMLQuery((DOMElement*) n,err);
				}
				if (n>=MAXOPS) {
					err=BXQerr(L("Too many substreams"));
					return NULL;
				}
				CMulStream* st=(CMulStream*) GrabStream(sty);
				if (st==NULL) {
					err=BXQerr(L("No more streams"));
					return NULL;
				}
				st->m_ops=n;
				if (n>=MAXOPS) {
					err=BXQerr(L("Too many substreams"));
					FreeStream(st);
					return NULL;
				}
				st->m_op=new CStream* [n];
				for (i=0;i<n;i++) {
					DOMNode* n=l->item(i);
					if (n->getNodeType() != DOMNode::ELEMENT_NODE) break;
					st->m_op[i]=BuildXMLQuery((DOMElement*) n,err);
					if (st->m_op[i]==NULL && it != TAG_OR) break;
				}
				if (i < n) {
					FreeStream(st);
					return NULL;
				}
				if (!st->Flatten()) {
					err=BXQerr(L("Illegal sequence"));
					return NULL;
				}
				return st;
			}
 	}
	return NULL;
}


/*
void CStreamFactory::MakeDTD()
{
	strcpy(m_dtdpath,TempPath());
	strcat(m_dtdpath,"cql.dtd");
	FILE* f=fopen(m_dtdpath,"w");
	for (int i=0;i<NDTDT;i++) {
		fwrite(dtd_txt[i],1,strlen(dtd_txt[i]),f);
	}
	fclose(f);
}
*/

CStream* CStreamFactory::GrabStream(stream_type st)
{
	CStream* s;
	switch (st) {
		case TY_ATOM:
			s=new CAtomStream(m_server);
			break;
		case TY_OR:
			s=new COrStream(m_server);
			break;
		case TY_NEG:
			s=new CNegStream(m_server);
			break;
		case TY_ALL:
			s=new CAllStream(m_server);
			break;
		case TY_SEQ:
			s=new CSeqStream(m_server);
			break;
		case TY_FILE:
			s=new CKeyStream(m_server);
			break;
		case TY_MEM:
			s=new CMemStream(m_server);
			break;
		case TY_AND:
			s=new CAndStream(m_server);
			break;
		case TY_PROD:
			s=new CProdStream(m_server);
			break;
		case TY_LIMIT:
			s=new CLimitStream(m_server);
			break;
		case TY_ID:
			s=new CIdStream(m_server);
			break;
		case TY_IDC:
			s=new CIDCStream(m_server);
			break;
		case TY_BPROD:
			s=new CBProdStream(m_server);
			break;
	}
	if (s==NULL) throw new CXairoError(CXairoIError::ERR_MEMORY,L("No memory for streams"));
	s->m_vars=m_vars;
	return s;
}


	

void CStreamFactory::FreeStream(CStream* s)
{
	delete s;
}


CStream* CStreamFactory::MakeWordStream(const xairo_uchar* w,bool bCase)
{
	UnicodeString us=m_server->GetLanguage()->MkLC(w);
	us=m_server->GetLanguage()->MkNorm(us);
	int d=m_server->GetWL()->LookupW(us.getTerminatedBuffer());
	int xd=m_server->GetWL()->LookupW(w);
	if (d==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	if (dd==-1) return NULL;
	int k=0;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.t != it_content) continue;
		if (bCase) {
			if (sb->w.xval==xd) k++;
		}
		else k++;
	}
	if (k==0) return NULL;
	int* sbl=new int [k];
	k=0;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.t != it_content) continue;
		if (bCase) {
			if (sb->w.xval==xd) sbl[k++]=dd+i;
		}
		else sbl[k++]=dd+i;
	}
	CStream* s=MakeStream(k,sbl);
	delete [] sbl;
	return s;
}

CStream *CStreamFactory::MakeAtomStream(const xairo_uchar* w,const xairo_uchar* ns, typ t)
{
	int d=m_server->GetWL()->LookupW(w);
	if (d==-1) return NULL;
	int dns=m_server->GetWL()->LookupW(ns);
	if (dns==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	if (dd==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.euri != dns) continue;
		if (sb->w.t != t) continue;
		int tt=dd+i;
		return MakeStream(1,&tt);
	}
	return NULL;
}

CStream *CStreamFactory::MakeAtomAStream(const xairo_uchar* w,const xairo_uchar* ns, typ t)
{
	int d=m_server->GetWL()->LookupW(w);
	if (d==-1) return NULL;
	int dns=m_server->GetWL()->LookupW(ns);
	if (dns==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	if (dd==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.uri != dns) continue;
		if (sb->w.t != t) continue;
		int t=dd+i;
		return MakeStream(1,&t);
	}
	return NULL;
}

CStream *CStreamFactory::MakeAtomBStream(const xairo_uchar* w,const xairo_uchar* e,const xairo_uchar* ns, typ t)
{
	int d=m_server->GetWL()->LookupW(w);
	if (d==-1) return NULL;
	int de=m_server->GetWL()->LookupW(e);
	if (de==-1) return NULL;
	int dns=m_server->GetWL()->LookupW(ns);
	if (dns==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	if (dd==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.element != de) continue;
		if (sb->w.euri != dns) continue;
		if (sb->w.t != t) continue;
		int t=dd+i;
		return MakeStream(1,&t);
	}
	return NULL;
}

CStream *CStreamFactory::MakeAtomBStream(const xairo_uchar* w,const xairo_uchar* ns, typ t)
{
	int d=m_server->GetWL()->LookupW(w);
	if (d==-1) return NULL;
	int dns=m_server->GetWL()->LookupW(ns);
	if (dns==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	if (dd==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.euri != dns) continue;
		if (sb->w.t != t) continue;
		int t=dd+i;
		return MakeStream(1,&t);
	}
	return NULL;
}

CStream *CStreamFactory::MakeAtomCStream(const xairo_uchar* w,const xairo_uchar* e,const xairo_uchar* ns, const xairo_uchar* facet,typ t)
{
	int d=m_server->GetWL()->LookupW(facet);
	if (d==-1) return NULL;
	int da=m_server->GetWL()->LookupW(w);
	if (da==-1) return NULL;
	int de=m_server->GetWL()->LookupW(e);
	if (de==-1) return NULL;
	int dns=m_server->GetWL()->LookupW(ns);
	if (dns==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	if (dd==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != d) break;
		if (sb->w.element != de) continue;
		if (sb->w.uri != dns) continue;
		if (sb->w.t != t) continue;
		if (sb->w.name != da) continue;
		int t=dd+i;
		return MakeStream(1,&t);
	}
	return NULL;
}

CStream* CStreamFactory::MakeLemmaStream(const xairo_uchar* w,const xairo_uchar* key)
{
	int ak=m_server->GetWL()->lemma_index(key);
//	int d=m_server->GetWL()->LookupW(w);
//	if (d==-1) return NULL; 
	xara_lemma* l=m_server->GetDict(ak)->LookupEntry(w);
//	CWord* wd=m_server->GetDict(ak)->GetEntry(w);
	if (l==NULL) return NULL;
	int k=0;
	int fc=l->forms;
	for (int i=0;i<fc;i++) {
		xara_form* f=&l->pt[i];
		k+=f->count;
	}
	if (k==0) return NULL;
	int *x=new int[k]; 
	k=0;
	for (int i=0;i<fc;i++) {
		xara_form* f=&l->pt[i];
		for (int j=0;j<f->count;j++)
			x[k++]=f->ptr+j;
	}
	CStream* s=MakeStream(k,x);
	delete [] x;
	return s;
}

CStream* CStreamFactory::MakeLemmasStream(int n,UnicodeString* w,const xairo_uchar* key)
{
	int ak=m_server->GetWL()->lemma_index(key);
//	int d=m_server->GetWL()->LookupW(w);
//	if (d==-1) return NULL; 
	xara_lemma* l=m_server->GetDict(ak)->LookupEntry2(n,w);
//	CWord* wd=m_server->GetDict(ak)->GetEntry(w);
	if (l==NULL) return NULL;
	int k=0;
	int fc=l->forms;
	for (int i=0;i<fc;i++) {
		xara_form* f=&l->pt[i];
		k+=f->count;
	}
	if (k==0) return NULL;
	int *x=new int[k]; 
	k=0;
	for (int i=0;i<fc;i++) {
		xara_form* f=&l->pt[i];
		for (int j=0;j<f->count;j++)
			x[k++]=f->ptr+j;
	}
	CStream* s=MakeStream(k,x);
	delete [] x;
	return s;
}

CStream* CStreamFactory::MakeAllPosStream(int j,const xairo_uchar** tag,const xairo_uchar** key)
{
	if (j==1) return MakeAllPos1Stream(*tag,*key);
	CAndStream* s=(CAndStream*) GrabStream(TY_AND);
	s->m_ops=j;
	s->m_op=new CStream* [j];
	for (int i=0;i<j;i++)
		s->m_op[i]=MakeAllPos1Stream(tag[i],key[i]);
	return s;
}

CStream* CStreamFactory::MakeAllPos1Stream(const xairo_uchar* tag,const xairo_uchar* key)
{
	int k=m_server->GetWL()->LookupW(tag);
	if (k==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(k);
	int x=m_server->GetWL()->LookupW(key);
	if (x==-1) return NULL;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != k) break;
		if (sb->w.t != it_addkey) continue;
		if (sb->w.name != x) continue;
		int t=dd+i;
		CStream* s=MakeStream(1,&t);
		if (s==NULL) return NULL;
//		s->wc=1; // why do this? It will be deduced from the type.
		return s;
	}
	return NULL;
}

CStream* CStreamFactory::MakePosStream(const xairo_uchar* w,int j,const xairo_uchar** pos,const xairo_uchar** key)
{
	int id=-1;
	int k=m_server->GetWL()->LookupW(w);
	if (k==-1) return NULL;
	int *ix=new int[j];
	int *wx=new int[j];
	for (int i=0;i<j;i++) {
		ix[i]=m_server->GetWL()->addkey_index(key[i]);
		if (ix[i]<0) return NULL;
		wx[i]=m_server->GetWL()->LookupW(pos[i]);
		if (wx[i]==-1) return NULL;
	}
	int dd=m_server->GetWL()->GetDictPtr(k);
	int p=0;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != k) break;
		if (sb->w.t != it_content) continue;
		bool bOK=true;
		for (int ii=0;ii<j;ii++) {
			if (sb->w.con[ix[ii]] != wx[ii]) 
				bOK=false;
		}
		if (bOK) p++;
	}
	if (p==NULL) return NULL;
	int* xx=new int[p]; p=0;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val != k) break;
		if (sb->w.t != it_content) continue;
		bool bOK=true;
		for (int ii=0;ii<j;ii++) {
			if (sb->w.con[ix[ii]] != wx[ii]) 
				bOK=false;
		}
		if (bOK) xx[p++]=dd+i;
	}
	CStream* s=MakeStream(p,xx);
	delete [] ix; delete [] wx;
	delete [] xx;
	return s;
}


CStream* CStreamFactory::MakeRegExpStream(const xairo_uchar *patt,bool bCS)
{
	int p; xara_loc* b;
	p=EvalRegexpM(patt,bCS,&b);
	return MakeMemStream(p,b);
}

CStream* CStreamFactory::MakeAttributeRegexpStream(const xairo_uchar* e,const xairo_uchar* ns,const xairo_uchar* a,const xairo_uchar* ans,const xairo_uchar *patt)
{
	int p; xara_loc* b;
	p=EvalAttributeRegexpM(e,ns,a,ans,patt,&b);
	return MakeMemStream(p,b);
//	CDBKey k;
//	if (!EvalAttributeRegexp(e,ns,a,ans,patt,k)) return false;
//	return MakeKeyStream(k);
}
 
CStream* CStreamFactory::MakeKeyStream(CDBKey* k)
{
	CKeyStream* s=(CKeyStream*) GrabStream(TY_FILE);
	if (s==NULL) return NULL;
	s->m_key=k;
	return s;
}

CStream* CStreamFactory::MakeMemStream(int p,xara_loc* b)
{
	CMemStream* s=(CMemStream*) GrabStream(TY_MEM);
	if (s==NULL) return NULL;
	s->m_count=p;
	s->m_buffer=b;
	return s;
}

CStream* CStreamFactory::MakeTagStream(const xairo_uchar* w,const xairo_uchar* ns,int n,_attstream* a)
{
	CStream *s0=NULL; int k=0;
	if (wcscmp(w,L("0")) != 0) { // better use a flag here
		s0=MakeAtomStream(w,ns,it_start);
		if (s0) k++;
	}
	int i0=-1;
	int i;
	for (i=0;i<n;i++) {
		if (i0 >= 0 && wcscmp(a[i].name,a[i0].name)==0) continue;
		else if (i0>=0) {
			k++;
		}
		i0=i;
	}
	if (n) k++;
	if (k==0) return NULL;
	CAndStream* s=(CAndStream*) GrabStream(TY_AND);
//	s->wtype = it_start;
	s->m_ops=k;
	s->m_op=new CStream* [k];
	k=0;
	if (s0) {
		s->m_op[0]=s0;
		k++;
	}
	i0=-1;
	for (i=0;i<n;i++) {
		if (i0 >= 0 && wcscmp(a[i].name,a[i0].name)==0) continue;
		else if (i0>=0) {
			s->m_op[k++]=MakeMultiAttributeStream(w,ns,i-i0,&a[i0]);
		}
		i0=i;
	}
	if (n) s->m_op[k++]=MakeMultiAttributeStream(w,ns,i-i0,&a[i0]);
	return s;
}

CStream* CStreamFactory::MakeMultiAttributeStream(const xairo_uchar* w,const xairo_uchar* ns,int n,_attstream* a)
{
	if (n==1) 
		return MakeAttributeStream(w,ns,&a[0]);
	COrStream* s=(COrStream*) GrabStream(TY_OR);
	s->m_ops=n;
	s->m_op=new CStream* [n];
	for (int i=0;i<n;i++) {
		s->m_op[i]=MakeAttributeStream(w,ns,&a[i]);
	}
	return s;
}

CStream* CStreamFactory::MakeAttributeStream(const xairo_uchar* w,const xairo_uchar* ns,_attstream* a)
{
	if (a->bPatt)
		return MakeAttributeRegexpStream(w,ns,a->name,a->ns,a->value);
	if (a->bVar)
		return MakeAttributeVarStream(w,ns,a->name,a->ns,a->value);
	if (a->extra && a->extra[0])
		return MakeAttributeValStream(w,ns,a->name,a->ns,a->value,a->extra);
	return MakeAttrStream(w,ns,a->name,a->ns,a->value,false);
}

// Used in u-who evaluation
CStream* CStreamFactory::MakeAttributeVarStream(const xairo_uchar* w,const xairo_uchar* ns,const xairo_uchar* a,const xairo_uchar* ans,const xairo_uchar* v)
{
	CStream* s0=MakeAttrStream(w,ns,a,ans,a,true); // use lookup stream
	if (s0==NULL) return NULL;
	CIdStream* s = (CIdStream*) GrabStream(TY_ID);
	s->m_op=s0;
	s->m_kvar=_wtoi(v); // must improve var naming!
	s->m_elt=_wcsdup(w);
	s->m_url=_wcsdup(ns);
	s->m_att=_wcsdup(a);
	s->m_aurl=_wcsdup(ans);
	return s;
}

CStream* CStreamFactory::MakeAttributeIdStream(const xairo_uchar* w,const xairo_uchar* ns,const xairo_uchar* a,const xairo_uchar* ans,int v)
{
	const xairo_uchar* s=m_server->GetWL()->GetWordW(v);
	CStream* s0=MakeAttrStream(w,ns,a,ans,s,false); // use lookup stream
	return s0;
}

// Handle MULTIDREFs
CStream* CStreamFactory::MakeAttributeValStream(const xairo_uchar* w,const xairo_uchar* ns,const xairo_uchar* a,const xairo_uchar* ans,const xairo_uchar* v,const xairo_uchar* e)
{
	CStream* s0=MakeAttrStream(w,ns,a,ans,v,true);
	if (s0==NULL) return NULL;
	CIDCStream* s = (CIDCStream*) GrabStream(TY_IDC);
	s->m_op=s0;
	s->m_kval=m_server->GetWL()->LookupW(e);
	return s;
}

// I think this is daft - there will be at most one stream . The Xpath engine needs
// access to it.
CStream* CStreamFactory::MakeAttrStream(const xairo_uchar* w,const xairo_uchar* ns,const xairo_uchar* a,const xairo_uchar* ans,const xairo_uchar* v,bool bVal)
{
	typ it; bool bG=wcscmp(w,L("0"))==0 || ans[0];
	if (bG)
		it=bVal?it_global_attribute_value:it_global_attribute;
	else
		it=bVal?it_local_attribute_value:it_local_attribute;
	int d=m_server->GetWL()->LookupW(v);
	if (d==-1) return NULL;
	int dd=m_server->GetWL()->GetDictPtr(d);
	int de=m_server->GetWL()->LookupW(w);
	int dens=m_server->GetWL()->LookupW(ns);
	int da=m_server->GetWL()->LookupW(a);
	int dans=m_server->GetWL()->LookupW(ans);
	int p=0;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val!=d) break;
		if (sb->w.t != it) continue;
		if (sb->w.name==da && sb->w.uri==dans && (bG || sb->w.element==de && sb->w.euri==dens)) 
			p++;
	}
	int* x=new int[p];
	p=0;
	for (int i=0;;i++) {
		xara_streambase* sb=m_server->GetSB()->GetSB(dd+i);
		if (sb->w.val!=d) break;
		if (sb->w.t != it) continue;
		if (sb->w.name==da && sb->w.uri==dans && (bG || sb->w.element==de && sb->w.euri==dens)) 
			x[p++]=dd+i;
	}
	return MakeStream(p,x);
}

CStream* CStreamFactory::MakeStream(int n,int* x)
{
	if (n==1) return Make1Stream(*x);
	COrStream* s=(COrStream*) GrabStream(TY_OR);
	s->m_ops=n;
	s->m_op=new CStream* [n];
	for (int i=0;i<n;i++) {
		s->m_op[i]=Make1Stream(x[i]);
	}
	return s;
}

CAtomStream* CStreamFactory::Make1Stream(int n)
{
	xara_streambase *sb=m_server->GetSB()->GetSB(n);
	CAtomStream *s=(CAtomStream*) GrabStream(TY_ATOM);
	if (s==NULL) return NULL;
	s->m_word=m_server->GetWL()->GetWordW(sb->w.val);
	s->m_xword=m_server->GetWL()->GetWordW(sb->w.xval);
	s->m_wtype = sb->w.t;
	s->m_first = sb->first;
	s->m_last = sb->first+sb->count;
	s->m_accel = sb ->acc-1;
	if (s->m_accel >= 0) m_server->GetAccelerators()->SetLast(s->m_accel,s->m_last);
	return s;
}

bool CStreamFactory::EvalRegexp(const xairo_uchar* w,bool bCase,CDBKey*& key)
{
	xara_loc* b;
	int p=EvalRegexpM(w,bCase,&b);
	key=m_server->GetDB()->newkey(sizeof (xara_loc));
	m_server->GetDB()->record(key,b,sizeof (xara_loc),p);
	m_server->GetDB()->closekey(key);
	delete [] b;
	return true;
}

int CStreamFactory::EvalRegexpM(const xairo_uchar* w,bool bCase,xara_loc** bx)
{
	CRegularExpression re(w,bCase);
	if (!re.Parse()) throw new CXairoError(CXairoIError::ERR_REGEXP,L("Badly formed regexp"));
	UnicodeString pre=re.Predict(10);
	pre=m_server->GetLanguage()->MkLC(pre);
	pre=m_server->GetLanguage()->MkNorm(pre);
	int x1=-1; int x2=-1;
	m_server->GetDict(-1)->GetLimits(pre.getTerminatedBuffer(),x1,x2);
	int p=0;
	for (int i=x1;i<=x2;i++) {
		xara_lemma* w; 
		w=m_server->GetDict(-1)->LookupEntry(i);
		for (int j=0;j<w->forms;j++) {
			xara_form* f=&w->pt[j];
//			m_server->GetDict(-1)->GetForm(w->GetNorm(),j,m,f,n);
			for (int k=0;k<f->count;k++) {
				int x=f->ptr+k;
				xara_streambase *sb=m_server->GetSB()->GetSB(x);
				if (sb->w.t != it_content) continue;
				if (!re.Match(m_server->GetWL()->GetWordW(sb->w.xval))) continue;
				p+=sb->count;
			}
		}
	}
	xara_loc* b=new xara_loc [p];
	if (b==NULL) return 0;
	int pp=0;
	for (int i=x1;i<=x2;i++) {
		xara_lemma* w; 
		w=m_server->GetDict(-1)->LookupEntry(i);
		for (int j=0;j<w->forms;j++) {
			xara_form* f=&w->pt[j];
//			m_server->GetDict(-1)->GetForm(w,j,m,f, n);
			for (int k=0;k<f->count;k++) {
				int x=f->ptr+k;
				xara_streambase *sb=m_server->GetSB()->GetSB(x);
				if (sb->w.t != it_content) continue;
				if (!re.Match(m_server->GetWL()->GetWordW(sb->w.xval))) continue;
				for (int x=0;x<sb->count;x++) {
					xara_loc *l=m_server->GetLocs()->GetLoc(sb->first+x);
					b[pp++]=*l;
				}
			}
		}
	}
	qsort(b,p,sizeof (xara_loc),compare_locs);
	*bx=b;
	return p;
}

int CStreamFactory::EvalAttributeRegexpM(const xairo_uchar* e,const xairo_uchar* ens,const xairo_uchar* a,const xairo_uchar* ans,const xairo_uchar* patt,xara_loc** bx)
{
 	CRegularExpression re(patt,true);
	if (!re.Parse()) throw new CXairoError(CXairoIError::ERR_REGEXP,L("Badly formed regexp"));
	const xairo_uchar* ns=(wcscmp(e,L("0"))==0)?ans:ens;
	int k=m_server->GetAttDict()->Lookup(a,e,ns);
	if (k==-1) return false;
	int p=0;
	while (k != -1) {
		xara_streambase *sb=m_server->GetSB()->GetSB(k);
		const xairo_uchar* s=m_server->GetWL()->GetWordW(sb->w.val);
		if (re.Match(s))
			p+=sb->count;
		k=sb->next;
	}
	xara_loc* b=new xara_loc [p];
	if (b==NULL) return 0;
	int pp=0;
	k=m_server->GetAttDict()->Lookup(a,e,ns);
	while (k != -1) {
		xara_streambase *sb=m_server->GetSB()->GetSB(k);
		const xairo_uchar* s=m_server->GetWL()->GetWordW(sb->w.val);
		if (re.Match(s)) {
			for (int x=0;x<sb->count;x++) {
				xara_loc *l=m_server->GetLocs()->GetLoc(sb->first+x);
				b[pp++]=*l;
			}
		}
		k=sb->next;
	}
	qsort(b,p,sizeof (xara_loc),compare_locs);
	*bx=b;
	return p;
}

int CStreamFactory::compare_locs(const void* l1,const void* l2)
{
	xara_loc* ll1=(xara_loc*) l1;
	xara_loc* ll2=(xara_loc*) l2;
	return ll1->seq-ll2->seq;
}

// This used to flatten seq streams but in xql I think that is unnecessary

int CStreamFactory::XTokenise(UnicodeString us,UnicodeString*& ss)
{
	return XTokenise(us,0,NULL,ss);
}

UnicodeString CStreamFactory::NoEsc(UnicodeString s)
{
	int x=s.indexOf("[[");
	int y=s.indexOf("]]");
	int z=x; char c='[';
	if (y != -1 && (x == -1 || y < x)) {
		z=y;
		c=']';
	}
	if (z==-1) 
		return s;
	UnicodeString ss;
	s.extract(0,y,ss);
	ss+=c;
	UnicodeString t;
	s.extract(z+2,s.length()-z-2,t);
	return ss+NoEsc(t);
}



int CStreamFactory::XTokenise(UnicodeString us, int m,UnicodeString* in,UnicodeString*& ss)
{
	if (us.isEmpty()) {
		ss=in;
		return m;
	}
	int k=0; 
	while (true) {
		k=us.indexOf('[',k);
		if (k==-1) break;
		if (k+1 < us.length() && us[k+1] == '[') 
			k+=2;
		else 
			break;
	}
	UnicodeString x; UnicodeString *x1, *x2;
	if (k==-1) {
		x=us;
	}
	else us.extract(0,k,x);
	x=NoEsc(x);
	int n1=m_server->GetLanguage()->Tokenise(x,x1);
	int n2=0; UnicodeString res;
	if (k != -1) {
		int kk=k;
		while (true) {
			kk=us.indexOf(']',kk);
			if (kk == -1) break;
			if (kk+1 < us.length() && us[kk+1]==']')
				kk+=2;
			else
				break;
		}
		UnicodeString y;
		if (kk==-1) {
			us.extract(k+1,us.length()-k-1,y);
//			res="";
		} else {
			us.extract(k+1,kk-k-1,y);
			us.extract(kk+1,us.length()-1,res);
		}
		y=NoEsc(y);
		int m=0; int km=0;
		while (TRUE) {
			m=y.indexOf(' ',m);
			if (m==-1) break;
			km++; m++;
		}
		n2=km+1;
		x2=new UnicodeString [km+1];
		km=0; m=0;
		while (TRUE) {
			int mm=y.indexOf(' ',m);
			if (mm==-1) break;
			y.extract(m,mm-m,x2[km]);
			km++; m=mm+1;
		}
		y.extract(m,y.length()-m,x2[km]);
	}
	UnicodeString* w=new UnicodeString[m+n1+n2];
	for (int i=0;i<m;i++)
		w[i]=in[i];
	for (int i=0;i<n1;i++)
		w[m+i]=x1[i];
	for (int i=0;i<n2;i++)
		w[m+n1+i]=x2[i];
	return XTokenise(res,m+n1+n2,w,ss);
}



XAIRO_NAMESPACE_END






