#pragma once

//#include <xercesc/dom/DOM.hpp>

//
using namespace XERCES_CPP_NAMESPACE;

#include "Stream.h"

XAIRO_NAMESPACE_BEGIN

struct _attstream {
	const xairo_uchar* name;
	const xairo_uchar* ns;
	const xairo_uchar* value;
	const xairo_uchar* extra;
	bool bVar;
	bool bPatt;
};

class CStreamFactory
{
public:
	CStreamFactory(CXairoServer* s);
	~CStreamFactory();
	CStream* BuildXMLQuery(const xairo_uchar*,const xairo_uchar*&);
	CStream *MakeAtomStream(const xairo_uchar*,const xairo_uchar*, typ);
	CStream *MakeAtomAStream(const xairo_uchar*,const xairo_uchar*, typ);
	CStream* MakeAtomBStream(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,typ);
	CStream* MakeAtomBStream(const xairo_uchar*,const xairo_uchar*,typ);
	CStream* MakeAtomCStream(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,typ);
	CStream* MakeAttributeIdStream(const xairo_uchar* ,const xairo_uchar* ,const xairo_uchar* ,const xairo_uchar*,int v);
	CAtomStream* Make1Stream(int);
	void MakeDTD();
	void FreeStream(CStream*);
private:
	CStream* BuildXMLQuery(DOMElement* e,const xairo_uchar*& err);
	CStream* GrabStream(stream_type);
	void FlattenStream(CStream*);
	char m_dtdpath[_MAX_PATH];
	int Tokenise(const xairo_uchar*,xairo_uchar*);
	CXairoServer* m_server;
	bool EvalRegexp(const xairo_uchar*,bool,CDBKey*&);
	int EvalRegexpM(const xairo_uchar*,bool bCase,xara_loc**);
	int EvalAttributeRegexpM(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,xara_loc**);
	static int compare_locs(const void* l1,const void* l2);
	const xairo_uchar* BXQerr(const xairo_uchar*,...);
	CStreamVar* m_vars;
	int XTokenise(UnicodeString, UnicodeString*&);
	int XTokenise(UnicodeString, int,UnicodeString*,UnicodeString*&);
	UnicodeString NoEsc(UnicodeString);

private:
	CStream *MakeWordStream(const xairo_uchar*,bool); // bad name - means ignore pos
	CStream* MakeTagStream(const xairo_uchar*,const xairo_uchar*, int, _attstream*);
	CStream* MakeMultiAttributeStream(const xairo_uchar*,const xairo_uchar*, int, _attstream*);
	CStream* MakeAttributeStream(const xairo_uchar*,const xairo_uchar*, _attstream*);
	CStream* MakeRegExpStream(const xairo_uchar*,bool); 
	CStream* MakeFormStream(const xairo_uchar*,int); 
	CStream* MakeAllPosStream(int,const xairo_uchar**,const xairo_uchar**);
	CStream* MakeAllPos1Stream(const xairo_uchar*,const xairo_uchar*);
	CStream* MakePosStream(const xairo_uchar*,int,const xairo_uchar**,const xairo_uchar**);
	CStream* MakeLemmaStream(const xairo_uchar*,const xairo_uchar*);
	CStream* MakeLemmasStream(int,UnicodeString*,const xairo_uchar*);
	CStream* MakeAttributeRegexpStream(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*);
	CStream* MakeAttributeVarStream(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*);
	CStream* MakeAttributeValStream(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*);
	CStream* MakeAttrStream(const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,const xairo_uchar*,bool);
	CStream* MakeKeyStream(CDBKey*);
	CStream* MakeMemStream(int,xara_loc*);
	CStream* MakeStream(int,int*);
};

XAIRO_NAMESPACE_END
