#include "SServer.h"
#include "WordList.h"
#include "Parameters.h"
#include "DataReader.h"

XAIRO_NAMESPACE_BEGIN

CWordList::CWordList(CXairoServer* s):m_server(s)
{
	m_hi=NULL;
	m_wl=NULL;
	m_cont=NULL;
	m_tax=NULL;
}

CWordList::~CWordList()
{
	for (int i=0;i<m_wlsize;i++) {
		delete [] m_wl[i].word;
	}
	delete [] m_hi;
	delete [] m_wl;
	delete [] m_cont;
	delete [] m_tax;
}

const xairo_uchar* CWordList::GetWordW(int n)
{
	return m_wl[n].word;
}

const char* CWordList::GetWord(int n)
{
	const xairo_uchar* u=GetWordW(n);
	char* c=new char [wcslen(u)+1];
	wcstombs(c,u,wcslen(u)+1);
	return c;
}

bool CWordList::Load()
{
	FILE* f=_wfopen(m_server->GetParam()->exppath(PARAM_INDEXPATH,L("xdbwl")),L("rb"));
	if (f==NULL) {
		m_server->BootLog(L("The file xdbwl is missing\n"));
		return false;
	}
	CDataReader* d=new CDataReader(0,f);
	xara_wl_header h; 
	int x=d->ReadWLHeader(1,&h);
	if (x==0 || h.sig != DBSIG) {
		fclose(f);
		m_server->BootLog(L("Bad index signature %08X\n"),h.sig);
		return false;
	}
	if (h.version != DBVER) {
		fclose(f);
		m_server->BootLog(L("Bad index version %d\n"),h.version);
		return false;
	}
	m_hashsize=h.hashsize;
	m_wlsize=h.size;
	m_places=h.places;
	m_shift=h.shift;
	m_ncont=h.cont;
	m_ntax=h.ntax;
	m_nls=h.nls;
//	m_nlemma=-1; // for now
	int* t=new int[h.size];
	m_cont = new xara_wl_header_addkey [m_ncont];
	m_tax = new xara_wl_header_tax[m_ntax];
	d->ReadWLHeaderAddkey(m_ncont,m_cont);
	xara_wl_header_tax* tax=new xara_wl_header_tax [m_ntax];
	d->ReadWLHeaderTax(m_ntax,m_tax);
	m_wl=new xara_wl [h.size];
	m_hi=new int[h.hashsize];
	d->ReadWLHash(h.hashsize,m_hi);
	int i;
	for (i=0;i<h.size;i++) {
		xara_wl_file fl;
		d->ReadWLFile(1,&fl);
		t[i]=fl.len;
		m_wl[i].next=fl.next;
		m_wl[i].dicptr=fl.dicptr;
	}
	for (i=0;i<h.size;i++) {
		xairo_uchar* x=new xairo_uchar [t[i]+1];
		d->ReadWChars(t[i],x);
		x[t[i]]=0;
		m_wl[i].word=x;
	}
	delete [] t;
	fclose(f);
 	return true;
}

int CWordList::LookupW(const xairo_uchar* w)
{
	int h=hash(w);
	int k=m_hi[h];
	while (k != -1) {
		xairo_uchar *ww=m_wl[k].word;
		if (wcscmp(w,ww)==0) return k;
		k=m_wl[k].next;
	}
	return -1;
}

int CWordList::hash(const xairo_uchar* w)
{
	unsigned int k=0;
	for (int i=0;i<m_places;i++) {
		if (w[i]==0) break; // new
		k+=(w[i] << m_shift*i);
	}
	return k % m_hashsize;
}

int CWordList::addkey_index(const xairo_uchar* u)
{
	for (int i=0;i<m_ncont;i++) {
		if (m_cont[i].name==LookupW(u)) return i;
	}
	return -1;
}

int CWordList::lemma_index(const xairo_uchar* u)
{
	if (u[0]==0) return -1;
	for (int i=0;i<m_nls;i++) {
		if (wcscmp(m_ls[i],u)==0) return m_ncont+i;
	}
	return -1;
}

XAIRO_NAMESPACE_END
