#include "SServer.h"
#include "XPathTokeniser.h"
#include "StreamFactory.h"

XAIRO_NAMESPACE_BEGIN

bool UCLetter(xairo_uchar);
bool UCDigit(xairo_uchar);
bool UCCombiner(xairo_uchar);
bool UCExtender(xairo_uchar);
bool UCSpace(xairo_uchar);



CXPathTokeniser::CXPathTokeniser(CXairoServer* s):m_server(s)
{
}

CXPathTokeniser::~CXPathTokeniser(void)
{
}


#define NXSYM 11

const xairo_uchar* xtsym[NXSYM]={
	L("("),
	L(")"),
	L("["),
	L("]"),
	L(".."),
	L("."), 
	L("@"), 
	L(","),
	L("::"),
	L(":"),
	L("$")};

#define NXOP 11

const xairo_uchar* xtop[NXOP]={
	L("//"),
	L("/"),
	L("|"),
	L("+"),
	L("-"),
	L("="),
	L("!="),
	L("<="),
	L("<"),
	L(">="),
	L(">")
};

#define MAXPREC 100

struct _op {
	const xairo_uchar* name;
	int prec;
} ops[ETOP_MAX]= {
	{L("//"),80},
	{L("/"),80},
	{L("|"),70},
	{L("+"),50},
	{L("-"),50},
	{L("="),30},
	{L("!="),30},
	{L("<="),40},
	{L("<"),40},
	{L(">="),40},
	{L(">"),40},
	{L("*"),60},
	{L("or"),10},
	{L("and"),20},
	{L("mod"),60},
	{L("div"),60}
};

#define NFS


const xairo_uchar* fns[FN_MAX]=
	{L("last"),L("position"),L("count"),L("cql")};

void CXPathTokeniser::Adjust(int n,_xptoken* toks)
{
	bool bOp=false;
	int k;
	for (k=0;k<n;k++) {
		_tkop t=toks[k].ty;
		if (bOp) {
			if (t==XT_ALL||t==XT_NCNAME) toks[k].ty=XT_OP;
			bOp=false;
		} else {
			bOp=true;
			if (t==XT_AT||t==XT_AXIS||t==XT_ORB||t==XT_OSB||t==XT_OP) bOp=false;
			if (t==XT_NCNAME && k+1<n && toks[k+1].ty==XT_AXIS)
				toks[k].ty=XT_AXISNAME;
			if (t==XT_NCNAME && k+1<n && toks[k+1].ty==XT_ORB)
				toks[k].ty=XT_FUNCTION;

		}
	}
}

_xptoken CXPathTokeniser::NextToken(const xairo_uchar* s,int& k)
{
	while (UCSpace(s[k])) k++;
	if (s[k]==0) return _xptoken(XT_END,L(""));
	if (s[k]=='\"') {
		const xairo_uchar* ss=wcschr(s+k+1,'\"');
		int kk=ss?(ss-s):wcslen(s);
		xairo_uchar* x=new xairo_uchar [kk-k];
		wcsncpy(x,s+k+1,kk-k-1);
		x[kk-k-1]=0;
		k=kk+1;
		return _xptoken(XT_LITERAL,x);
	}
	if (s[k]=='\'') {
		const xairo_uchar* ss=wcschr(s+k+1,'\'');
		int kk=ss?(ss-s):wcslen(s);
		xairo_uchar* x=new xairo_uchar [kk-k];
		wcsncpy(x,s+k+1,kk-k-1);
		x[kk-k-1]=0;
		k=kk+1;
		return _xptoken(XT_LITERAL,x);
	}
	if (UCLetter(s[k])||s[k]=='_') {
		int i;
		for (i=k;i<(int) wcslen(s);i++) {
			if (!UCLetter(s[i]) && !UCDigit(s[i]) &&s [i] != '.' && s[i] != '-'  
				&& s[i] != '_' && !UCCombiner(s[i]) && !UCExtender(s[i]))
				break;
		}
		int kk=i;
		xairo_uchar* x=new xairo_uchar [kk-k+1];
		wcsncpy(x,s+k,kk-k);
		x[kk-k]=0;
		k=kk;
		return _xptoken(XT_NCNAME,x);
	}
	if (UCDigit(s[k])) {
		int i; 
		for (i=k;i<(int) wcslen(s);i++) {
			if (!UCDigit(s[i])) break;
		}
		int kk=i;
		xairo_uchar* x=new xairo_uchar [kk-k+1];
		wcsncpy(x,s+k,kk-k);
		x[kk-k]=0;
		k=kk;
		return _xptoken(XT_NUMBER,x);
	}
	if (s[k]=='*') {
		k++;
		return _xptoken(XT_ALL,L("*"));
	}
	for (int i=0;i<NXSYM;i++) {
		int t=wcslen(xtsym[i]);
		if (wcsncmp(s+k,xtsym[i],t)==0) {
			k+=t;
			return _xptoken((_tkop) i,xtsym[i]);
		}
	}
	for (int i=0;i<NXOP;i++) {
		int t=wcslen(xtop[i]);
		if (wcsncmp(s+k,xtop[i],t)==0) {
			k+=t;
			return _xptoken(XT_OP,xtop[i]);
		}
	}
	return _xptoken(XT_UNKNOWN,L(""));
}


_xpe* CXPathTokeniser::Parse(int i0,int i1)
{
	if (i0 > i1) return NULL;
	int levr=0; int levs=0;
	int jp=-1; int bp=MAXPREC; int ji=-1;
	for (int i=i0;i<=i1;i++) {
		if (m_toks[i].ty==XT_ORB) levr++;
		if (m_toks[i].ty==XT_CRB) levr--;
		if (levr < 0) return NULL;
		if (m_toks[i].ty==XT_OSB) levs++;
		if (m_toks[i].ty==XT_CSB) levs--;
		if (levs < 0) return NULL;
		if (levr || levs) continue;
		if (m_toks[i].ty != XT_OP) continue;
		int j;
		for (j=0;j<ETOP_MAX;j++) 
			if (wcscmp(ops[j].name,m_toks[i].val)==0) break;
		if (j==ETOP_MAX) return NULL;
		if (ops[j].prec <= bp) { // right assoc
			bp=ops[j].prec;
			jp=j;
			ji=i;
		}
	}
	if (levr || levs) return NULL;
	if (jp != -1) {
		_xpe* x=new _xpe;
		_xpe* x1=Parse(i0,ji-1);
//		if (x1==NULL) return NULL;
		_xpe* x2=Parse(ji+1,i1);
//		if (x2==NULL) return NULL;
		x->op=(etop) jp;
		x->op1=x1;
		x->op2=x2;
		x->etype=ET_OP;
		return x;
	}
// so we have a primary expression
	levs=0; int ns=0;
	for (int i=i0;i<=i1;i++) {
		if (m_toks[i].ty==XT_OSB && levs==0) ns++;
		if (m_toks[i].ty==XT_OSB) levs++;
		if (m_toks[i].ty==XT_CSB) levs--;
	}
	_xpe* x=new _xpe;
	x->ntests=ns;
	x->tests=new _xpe* [ns]; int j0, j1; int t=0;
	j1=i1;
	for (int i=i0;i<=i1;i++) {
		if (m_toks[i].ty==XT_OSB && levs==0) {
			j0=i; if (j1 >= i1) j1=i-1;
		}
		if (m_toks[i].ty==XT_OSB) levs++;
		if (m_toks[i].ty==XT_CSB) levs--;
		if (m_toks[i].ty==XT_CSB && levs==0) {
			x->tests[t++]=Parse(j0+1,i-1);
		}
	}
	switch (m_toks[i0].ty) {
		case XT_ORB:
			if (m_toks[j1].ty != XT_CRB) return NULL;
			x->etype=ET_BRACKETS;
			x->op1=Parse(i0+1,j1-1);
			break;
		case XT_LITERAL:
			x->etype=ET_LITERAL;
			x->s=m_toks[i0].val;
			break;
		case XT_NUMBER:
			x->etype=ET_NUMBER;
			x->s=m_toks[i0].val;
			break;
		case XT_AT:
//			x->etype=ET_NCATEST;
//			x->s=m_toks[i0+1].val;
			if (i0+3<=j1 && m_toks[i0+2].ty==XT_COLON){
				x->etype=ET_NCATEST;
				x->s=m_toks[i0+3].val;
				x->pre=m_toks[i0+1].val;
			} else {
				x->etype=ET_NCATEST;
				x->s=m_toks[i0+1].val;
				x->pre=L("");
			}
			break;
		case XT_NCNAME:
			if (i0+2<=j1 && m_toks[i0+1].ty==XT_COLON){
				x->etype=ET_NCTEST;
				x->s=m_toks[i0+2].val;
				x->pre=m_toks[i0].val;
			} else {
				x->etype=ET_NCTEST;
				x->s=m_toks[i0].val;
				x->pre=L("");
			}
			break;
		case XT_ALL:
			x->etype=ET_ALLTEST;
			break;
		case XT_FUNCTION:
			x->etype=ET_FUNCTION;
			x->s=m_toks[i0].val;
			int i;
			for (i=0;i<FN_MAX;i++) {
				if (wcscmp(fns[i],x->s)==0) break;
			}
			x->fn=(_fn) i;
			levr=0; int na=1;
			if (i0+2==j1) na = 0;
			else for (int i=i0+2;i<=j1-1;i++) {
				if (m_toks[i].ty==XT_ORB) levr++;
				if (m_toks[i].ty==XT_CRB) levr--;
				if (levr==0 && m_toks[i].ty==XT_COMMA) na++;
			}
			x->nargs=na;
			x->args=new _xpe* [na]; int a=0;
			int ip=i0+2;
			if (na) {
			for (int i=i0+2;i<=j1-1;i++) {
				if (m_toks[i].ty==XT_ORB) levr++;
				if (m_toks[i].ty==XT_CRB) levr--;
				if (levr==0 && m_toks[i].ty==XT_COMMA) {
					x->args[a++]=Parse(ip,i-1);
					ip=i+1;
				}
			}
			x->args[a]=Parse(ip,j1-1);
			}
			if (!PreProc(x)) return NULL;
			break;
			
	}
	return x;
}

bool CXPathTokeniser::PreProc(_xpe* x)
{
	if (x->fn==FN_CQL ) {
		x->usr=NULL;
		_xpe* a=x->args[0];
		if (a->etype==ET_LITERAL) {
			const xairo_uchar* q=a->s;
			const xairo_uchar* err;
			CStream* s=m_server->GetSF()->BuildXMLQuery(q,err);
			if (s==NULL) return false;
			x->usr=s;
		}
	}
	return true;
}


_xpe* CXPathTokeniser::Parse(const xairo_uchar* s)
{
	int k=0;
	int nt=0;
	while (k<(int) wcslen(s)) {
		if (NextToken(s,k).ty==XT_END) break;
		nt++;
	}
	m_toks=new _xptoken [nt];
	k=0;
	for (int i=0;i<nt;i++) {
		m_toks[i]=NextToken(s,k);
	}
	Adjust(nt,m_toks);
	return Parse(0,nt-1);
}







/*

	switch (s[0]) {
		case 0:
			return NULL;
			case 

tkbase
{L"(", TKX_ORB},
{L")", TKX_CRB},
{L"[", TKX_OSB},
{L"]", TKX_CSB},
{L"..",TKX_PARENT},
{L".", TKX_CTXT},
{L"@".TKX_AT},
{L",",TKX_COMMA},
{L"::",TKX_AXISSEP}

{L"or", TKX_OP,OP_OR}
{L"and", TKX_OP,OP_AND}
{L"mod", TKX_OP,OP_MOD}
{L"div", TKX_OP,OP_DIV}


{L"comment",TKX_NT,NT_COMMENT}
{L"text",TKX_NT,NT_TEXT}
{L"processing-instruction",TKX_NT,NT_PI}


letter -> ncname
_ -> ncname
digit ->number
" literal
' literal
* special
$ variable
: alone can be a nametest??
anything else - symbol or operator

Level 1 - find ncnmaes, literals &c
Level 2 - adjust



bool TkBase
{
	for (int i=0;i<NTKBASE;i++) {
		if (wcsncmp(t+k,ntb[i],wcslen(ntb[i]))==0) {
			tk.ty=ntb[i].ty;
			return true;
		}
	}
	return false;
}

bool TkOp
{





GetNextToken(const xairo_uchar* s,int& k)
{

*/

bool UCLetter(xairo_uchar w)
{
	if ('a' <= w && w <= 'z') return true;
	if ('A' <= w && w <= 'Z') return true;
	return false;
}

bool UCDigit(xairo_uchar w)
{
	if ('0' <= w && w <= '9') return true;
	return false;
}

bool UCExtender(xairo_uchar w)
{
	return false;
}

bool UCCombiner(xairo_uchar w)
{
	return false;
}

bool UCSpace(xairo_uchar c)
{
	return c==' '|| c=='\t' || c=='\n' || c=='\r';
}

XAIRO_NAMESPACE_END
