#include "SServer.h"
#include "XairoCorpusText.h"
#include "Stream.h"
#include "WordList.h"
#include "StreamFactory.h"
#include "XairoXpathValue.h"
#include "EntityManager.h"
#include "Thread.h"
#include "SB.h"
#include "XairoCorpusTextFilter.h"
#include "XairoTextEntity.h"

XAIRO_NAMESPACE_BEGIN

CXairoCorpusText::CXairoCorpusText(int i,CXairoServer* s):m_i(i),m_server(s)
{
	InitRange();
}

void CXairoCorpusText::InitRange()
{ 
	CAtomStream* s1=(CAtomStream*) m_server->GetSF()->MakeAtomStream(L("document"),XAIRA_URI,it_start);
	m_range.start=*s1->SetCursor(m_i);
	m_server->GetSF()->FreeStream(s1);
	CAtomStream* s2=(CAtomStream*) m_server->GetSF()->MakeAtomStream(L("document"),XAIRA_URI,it_end);
	m_range.end=*s2->SetCursor(m_i);
	m_server->GetSF()->FreeStream(s2);
	m_range.parent=0;
	m_range.pos=1;
	m_range.size=1;

}

CXairoCorpusText::~CXairoCorpusText()
{
}

_xrange* CXairoCorpusText::GetRootNode()
{
	_xrange* r=new _xrange;
	r->nt=NT_TEXT;
	CAtomStream* s1=(CAtomStream*) m_server->GetSF()->MakeAtomStream(L("document"),XAIRA_URI,it_start);
	r->start=*s1->SetCursor(m_i);
	m_server->GetSF()->FreeStream(s1);
	CAtomStream* s2=(CAtomStream*) m_server->GetSF()->MakeAtomStream(L("document"),XAIRA_URI,it_end);
	r->end=*s2->SetCursor(m_i);
	m_server->GetSF()->FreeStream(s2);
	r->parent=0;
	r->pos=1;
	r->size=1;
	r->def_uri=L("");
	return r;
}


const xairo_uchar* CXairoCorpusText::EvalPrefix(const xairo_uchar* p,_xrange* r)
{
	const xairo_uchar* uri=L("");
	if (wcscmp(p,L("xml"))==0) return L("http://www.w3.org/XML/1998/namespace");
	CAtomStream* s3=(CAtomStream*) m_server->GetSF()->MakeAtomAStream(p,L(""),it_spm);
	CAtomStream* s4=(CAtomStream*) m_server->GetSF()->MakeAtomAStream(p,L(""),it_epm);
	if (s3==NULL||s4==NULL) return uri;
	// the problem here is that the prefix map may come a bit after the document start
// edge forward to next start tag
	int k=r->start.seq; 
	while (true) {
		k++;
		int sbi=m_server->GetThread()->GetThread(k);
		xara_streambase* sb=m_server->GetSB()->GetSB(sbi);
		if (sb->w.t==it_start) 
			break;
	}
	if (s3->Advance(k,false)) {
		int kk=s3->m_seq;
		if (s4->Advance(kk,true)) {
			if (s4->m_seq >= k) {
				uri=m_server->GetWL()->GetWordW(s3->m_nw);
			}
		}
	}
	m_server->GetSF()->FreeStream(s3);
	m_server->GetSF()->FreeStream(s4);
	return uri;
}

const xairo_uchar* CXairoCorpusText::getTitle()
{
	CAtomStream* ds=m_server->GetDocumentStream();
	xara_loc* l=ds->SetCursor(m_i);
	return m_server->GetWL()->GetWordW(l->nw);
}

//CXairoIXPathValue* CXairoCorpusText::evalXPathUp(const xairo_uchar* path, int seq)
//{
//
//}

CXairoIXpathValue* CXairoCorpusText::evalXpath(const xairo_uchar* path)
{
	CXPathTokeniser xt(m_server);
	_xpe* e=xt.Parse(path);
	if (e==NULL) return NULL;
	_xrange* c=GetRootNode();
	return evalXpathinContext(e,c);
}

CXairoXpathValue* CXairoCorpusText::evalXpathinContext(_xpe* e,_xrange* v)
{
	switch (e->etype) {
			CXairoXpathValue* v1;
		case ET_OP:
			switch (e->op) {
		case ETOP_CHILD:
			if (e->op1) v1=evalXpathinContext(e->op1,v);
			else v1=new CXairoXpathValue(GetRootNode(),this);
			if (e->op2->etype==ET_NCTEST) 
				return ApplyNCTest(v1,e->op2,true);
			if (e->op2->etype==ET_NCATEST) 
				return ApplyNCATest(v1,e->op2);
			return NULL;
		case ETOP_DESC:
			if (e->op1) v1=evalXpathinContext(e->op1,v);
			else v1=new CXairoXpathValue(GetRootNode(),this);
			if (e->op2->etype==ET_NCTEST) 
				return ApplyNCTest(v1,e->op2,false);
			if (e->op2->etype==ET_NCATEST) 
				return ApplyNCATest(v1,e->op2);
			return NULL;
		case ETOP_EQUALS:
			return new CXairoXpathValue(Compare(e->op1,e->op2,v),this);
		case ETOP_NEQ:
			return new CXairoXpathValue(Compare(e->op1,e->op2,v),this);
		case ETOP_AND:
			return new CXairoXpathValue(evalXpathinContext(e->op1,v)->getBool() &&
				evalXpathinContext(e->op1,v)->getBool(),this);
		case ETOP_OR:
			return new CXairoXpathValue(evalXpathinContext(e->op1,v)->getBool() ||
				evalXpathinContext(e->op1,v)->getBool(),this);

		default:
			break;
			}
		case ET_ALLTEST:
			return new CXairoXpathValue(v,this);
		case ET_NCTEST:
			return ApplyNCTest(new CXairoXpathValue(v,this),e,true);
		case ET_NCATEST:
			return ApplyNCATest(new CXairoXpathValue(v,this),e);
		case ET_LITERAL:
			return new CXairoXpathValue(e->s,XPT_STRING,this);
		case ET_NUMBER:
			return new CXairoXpathValue(e->s,XPT_NUMBER,this);
		case ET_FUNCTION:
			return EvalFn(e,v);  // now we should filter in case the result was a nodeset eg id
		case ET_BRACKETS:
			return evalXpathinContext(e->op1,v); // ditto
		default:
			break;
	}
	return NULL;
}

CXairoXpathValue* CXairoCorpusText::ApplyNCTest(CXairoXpathValue *v,_xpe* test,bool bC)
{
	CXairoXpathValue* vv=new CXairoXpathValue(this);
	for (int i=0;i<v->Count();i++) {
		_xrange* r=v->Node(i);
		CXairoXpathValue* rr=ApplyNCTest(r,test,bC);
		if (rr) vv->Add(rr);
	}
	return vv;
}

CXairoXpathValue* CXairoCorpusText::ApplyNCATest(CXairoXpathValue* v,_xpe* t)
{
	CXairoXpathValue* vv=new CXairoXpathValue(this);
	for (int i=0;i<v->Count();i++) {
		_xrange* r=v->Node(i);
		CXairoXpathValue* rr=ApplyNCATest(r,t);
		if (rr) vv->Add(rr);
	}
	return vv;
}

CXairoXpathValue* CXairoCorpusText::ApplyNCATest(_xrange* r,_xpe* t)
{	
	const xairo_uchar* uri=EvalPrefix(t->pre,r);
	const xairo_uchar* w=EvalAttribute(t->s,uri,r->elt,r->start);
	if (w==NULL) return NULL;
	_xrange rr;
	rr.nt=NT_ATT;
	rr.elt=w;
	return new CXairoXpathValue(&rr,this);
}

// I suppose we need one of these for each axis

CXairoXpathValue* CXairoCorpusText::ApplyNCTest(_xrange* r,_xpe* t,bool bC)
{
	CXairoXpathValue* v=new CXairoXpathValue(this);
	const xairo_uchar* uri=EvalPrefix(t->pre,r);
	CAtomStream* s1=(CAtomStream*) m_server->GetSF()->MakeAtomStream(t->s,uri,it_start);
	CAtomStream* s2=(CAtomStream*) m_server->GetSF()->MakeAtomStream(t->s,uri,it_end);
	if (s1==NULL || s2==NULL) return v;
	int z=r->start.seq;
	bool bOK=false;
	while (	s1->Advance(z,true)) {
		xara_loc lz=*s1->SetCursor();
		xara_loc lze;
		int zz=lz.seq;
		if (lz.seq>r->end.seq) break;
		while (s2->Advance(zz,true)) {
			lze=*s2->SetCursor();
			if (lze.parent==lz.parent) break;
			zz=lze.seq+1;
		}
		_xrange nr;
		nr.nt=NT_TEXT;
		nr.start=lz;
		nr.end=lze;
		nr.parent=lz.seq;
		nr.elt=t->s;
		nr.def_uri=r->def_uri;
		if (!bC || lz.parent == r->parent) {
			v->Add(&nr);
		}
		z=lz.seq+1;
	}
	return FilterNodes(v,t);
}

// old version follows
CXairoXpathValue* CXairoCorpusText::evalXpathinContext(const xairo_uchar* path,_xrange r)
{
	int k=0; const xairo_uchar* elt;
	while (k<(int) wcslen(path)) {
		_xpc* pc=parseXpath(path,k);
		if (pc==NULL) return NULL;
		if (pc->bAtt) {
			if (!pc->bParent) return NULL;
			const xairo_uchar* w=EvalAttribute(pc->att,pc->ns,r.elt,r.start);
			return new CXairoXpathValue(w,XPT_STRING,this);
		}
		elt=pc->elt;
		if (wcscmp(elt,L("."))==0) continue;
		CAtomStream* s1=(CAtomStream*) m_server->GetSF()->MakeAtomStream(pc->elt,pc->ns,it_start);
		CAtomStream* s2=(CAtomStream*) m_server->GetSF()->MakeAtomStream(pc->elt,pc->ns,it_end);
		int z=r.start.seq;
		bool bOK=false;
		while (	s1->Advance(z,true)) {
			xara_loc lz=*s1->SetCursor();
			xara_loc lze;
			int zz=lz.seq;
			if (lz.seq>r.end.seq) break;
			while (s2->Advance(zz,true)) {
				lze=*s2->SetCursor();
				if (lze.parent==lz.parent) break;
				zz=lze.seq+1;
			}
			_xrange nr;
			nr.nt=NT_TEXT;
			nr.start=lz;
			nr.end=lze;
			nr.parent=lz.seq;
			nr.elt=pc->elt;
			nr.def_uri=r.def_uri;
			if (!pc->bParent || lz.parent==r.parent) {
//				if (TestNode(nr,pc))
				{
					r=nr;
					bOK=true;
					break;
				}
			}
			z=lz.seq+1;
		}
		if (!bOK) return false;
	}
	int it=r.start.fi;
	if (it != r.end.fi) return NULL;
	int p0=r.start.off;
	int p1=r.end.off;
	int k1=0; int k2=0;
	const xairo_uchar* s=m_server->GetEntityManager()->TextExtract(it,p0,p1-p0,&k1,&k2);
	return new CXairoXpathValue(s,XPT_STRING,this);
}

const xairo_uchar* CXairoCorpusText::NodeText(_xrange* r)
{
	switch (r->nt) {
		case NT_TEXT: 
			{
				int it=r->start.fi;
				if (it != r->end.fi) return NULL;
				int p0=r->start.off;
				int p1=r->end.off;
				int k1=0; int k2=0;
				const xairo_uchar* s=m_server->GetEntityManager()->TextExtract(it,p0,p1-p0,&k1,&k2);
				if (s==NULL) return L("");
				return s;
			}
		case NT_ATT:
			return r->elt;
	}
	return NULL;
}

CXairoXpathValue* CXairoCorpusText::FilterNodes(CXairoXpathValue* v,_xpe* p)
{
	for (int i=0;i<p->ntests;i++) {
		_xpe* test=p->tests[i];
		CXairoXpathValue* vv=FilterNodesOnce(test,v);
		v=vv; // debugging
	}
	return v;
}

CXairoXpathValue* CXairoCorpusText::FilterNodesOnce(_xpe* test,CXairoXpathValue* v)
{
	CXairoXpathValue* vv=new CXairoXpathValue(this);
//	PreProc(test);
	for (int i=0;i<v->Count();i++) {
		_xrange* r=v->Node(i);
		r->pos=i+1;
		r->size=v->Count();
		CXairoXpathValue* res=evalXpathinContext(test,r);
		if (res->IsNumber()) {
			if (res->getNumber()==i+1) 
				vv->Add(r);
		}
		else if (res->getBool()) vv->Add(r);
	}
	return vv;
}



bool CXairoCorpusText::pathchar(xairo_uchar c)
{
	if ('a' <= c && c <= 'z') return true;
	if ('A' <= c && c <= 'Z') return true;
	if ('0' <= c && c <= '9') return true;
	return false;
}


_xpc* CXairoCorpusText::parseXpath(const xairo_uchar* path,int& k)
{
	bool bParent=true; bool bAtt=false; bool bAbs=false;
	if (path[k] == '/') {
		bAbs=true;
		k++;
	}
	if (path[k] == '/') {
		bParent=false;
		k++;
	}
	if (path[k]=='@') {
		bAtt=true;
		k++;
	}
	int ii;
	for (ii=k;ii<(int) wcslen(path);ii++)
		if (!pathchar(path[ii])) break;
	int el=ii-k;
	xairo_uchar* e=new xairo_uchar [el+1];
	wcsncpy(e,path+k,el);
	e[el]=0;
	k+=el;
	_xpc* x=new _xpc;
	if (bAtt) x->att=e; else x->elt=e;
	x->ns=L("");
	x->bParent=bParent;
	x->bAtt=bAtt;
	x->bAbs=true;
	x->ntests=0;
	if (path[k]=='[') {
		k++;
		const xairo_uchar* kk=wcschr(path+k,']');
		if (kk==NULL) {
			k=wcslen(path);
			return x;
		}
		xairo_uchar* arg=new xairo_uchar [kk-(path+k)+1];
		wcsncpy(arg,path+k,kk-(path+k));
		arg[kk-(path+k)]=0;
		k=kk+1-path;
		x->ntests=1;
		x->tests=new _xtest [1];
		_xtest* t=&x->tests[0];
		if (arg[0] >= '0' && arg[0] <= '9') {
			t->xtty=XTTY_POS;
			t->pos=_wtoi(arg);
		}
		else if (wcschr(arg,'=')) {
			int kk1=wcschr(arg,'=')-arg;
			xairo_uchar* op1=new xairo_uchar[kk1+1];
			wcsncpy(op1,arg,kk1);
			op1[kk1]=0;
			xairo_uchar* op2=new xairo_uchar[wcslen(arg)-kk1];
			wcsncpy(op2,arg+kk1+1,wcslen(arg)-kk1-1);
			op2[wcslen(arg)-kk1-1]=0;
			t->xtty=XTTY_COMP;
			if (op1[0]=='\"') {
				t->op1.ty=XOP_STR;
				xairo_uchar* str=new xairo_uchar [wcslen(op1)-1];
				wcsncpy(str,op1+1,wcslen(op1)-2);
				str[wcslen(op1)-2]=0;
				t->op1.string=str;
			} else {
				t->op1.ty=XOP_PATH;
				t->op1.path=op1;
			}
			if (op2[0]=='\"') {
				t->op2.ty=XOP_STR;
				xairo_uchar* str=new xairo_uchar [wcslen(op2)-1];
				wcsncpy(str,op2+1,wcslen(op2)-2);
				str[wcslen(op2)-2]=0;
				t->op2.string=str;
			} else {
				t->op2.ty=XOP_PATH;
				t->op2.path=op2;
			}
		} else 
		{
			t->xtty=XTTY_PATH;
			t->xpath=arg;
		}
	}
	return x;
}


	

const xairo_uchar* CXairoCorpusText::EvalAttribute(const xairo_uchar* name,const xairo_uchar* uri,const xairo_uchar* elt,xara_loc l)
{
	CAtomStream *s=NULL;
	if (uri[0]==0)
		s=(CAtomStream*) m_server->GetSF()->MakeAtomBStream(name,elt,L(""),it_local_attribute_value);
	else 
		s=(CAtomStream*) m_server->GetSF()->MakeAtomAStream(name,uri,it_global_attribute_value);
	if (s==NULL) return NULL;
	if (!s->Advance(l.seq,true)) return NULL;
	xara_loc la;
	la =*s->SetCursor();
	if (la.fi != l.fi || la.off != l.off) return NULL;
	int nw=la.nw;
	m_server->GetSF()->FreeStream(s);
	return m_server->GetWL()->GetWordW(nw);
}

bool CXairoCorpusText::Compare(_xpe* a,_xpe* b,_xrange* v)
{
	CXairoXpathValue* v1=evalXpathinContext(a,v);
	CXairoXpathValue* v2=evalXpathinContext(b,v);
	return v1->Compare(v2);
}


CXairoXpathValue* CXairoCorpusText::EvalFn(_xpe* e,_xrange* v)
{
	int n=0;
	switch (e->fn) {
		CXairoXpathValue* a;
		case FN_LAST:
			return new CXairoXpathValue(v->size,this);
		case FN_POSITION:
			return new CXairoXpathValue(v->pos,this);
		case FN_COUNT:
			a=evalXpathinContext(e->args[0],v);
			n=a->Count();
			return new CXairoXpathValue(n,this);
		case FN_CQL:
			CStream* s=(CStream*) e->usr;
			if (s==NULL) return new CXairoXpathValue(false,this);
			if (!s->Advance(v->start.seq,true)) return new CXairoXpathValue(false,this);
			if (s->m_seq+s->m_nseq >= v->end.seq) return new CXairoXpathValue(false,this);
			return new CXairoXpathValue(true,this);
	}
	return NULL;
}


int CXairoCorpusText::classify(const CXairoIGrammar::Partition* p)
{
	CXairoGrammar::Partition* pp=(CXairoGrammar::Partition*) p;
	int t=pp->Classify(m_i);
	if (t < 0) return -1; else return t;
}

int CXairoCorpusText::getWordCount()
{
	CAtomStream* s=(CAtomStream*) m_server->GetSF()->MakeAtomStream(L("document"),XAIRA_URI,it_start);
	CAtomStream* s1=(CAtomStream*) m_server->GetSF()->MakeAtomStream(L("document"),XAIRA_URI,it_end);
	xara_loc* l=s->SetCursor(m_i);
	xara_loc* l1=s1->SetCursor(m_i);
	int n=l1->nw-l->nw;
	m_server->GetSF()->FreeStream(s);
	m_server->GetSF()->FreeStream(s1);
	return n;
}

bool CXairoCorpusText::isSelected(CXairoICorpusTextFilter* f)
{
	int* t;
	int n=((CXairoCorpusTextFilter*) f)->getTexts(t);
	bool b=m_i<n && (t[m_i] != 0);
	delete [] t;
	return b;
}

CXairoITextEntity* CXairoCorpusText::getTextEntity()
{
	CAtomStream* ds=m_server->GetDocumentStream();
	xara_loc* l=ds->SetCursor(m_i);
	int it=m_server->GetEntityManager()->GetOuterEntity(l->fi);
	return new CXairoTextEntity(it,m_server);
}

#define FCT L("ct(%d)")

const xairo_uchar* CXairoCorpusText::getId()
{
	return CXairoServer::MakeId(FCT,m_i);
}

/*
CXairoCorpusText::CXairoCorpusText(const xairo_uchar* id,CXairoServer* s):m_server(s)
{
	m_i=CXairoServer::ScanId(FCT,id);
	InitRange();
}
*/

XAIRO_NAMESPACE_END
