#include "SServer.h"
#include "XairoHit.h"
#include "EntityManager.h"
#include "XairoTextEntity.h"
#include "XairoCorpusText.h"
#include "Stream.h"
#include "WordList.h"
#include "EntityManager.h"
#include "XairoSolveError.h"

XAIRO_NAMESPACE_BEGIN

CXairoHit::CXairoHit(xara_loc l,int d,CXairoServer* s)
:m_loc(l),m_doc(d),m_server(s),m_bMissing(false)
{
}

CXairoHit::~CXairoHit(void)
{
}
// should retrieve doc object not doc title
CXairoITextEntity* CXairoHit::getTextEntity()
{
//	const xairo_uchar* p=m_server->GetEntityManager()->GetTextPath(m_fi);
	return new CXairoTextEntity(m_loc.fi,m_server);
}

CXairoICorpusText* CXairoHit::getCorpusText()
{
	return new CXairoCorpusText(m_doc,m_server);
}

// Here and everywhere else it would be better if accelerators
// were based on text entities

const xairo_uchar _s[]={0};

const xairo_uchar* CXairoHit::getText(int scope,int max)
{
	CStream* ss=m_server->GetScopeStream();
	CStream* ssd=m_server->GetDocumentStream();
	CStream* sse=m_server->GetDocumentEndStream();
	int p0=m_loc.seq; bool b=true;
	for (int i=0;i<scope+1;i++) {
		b=ss->Advance(p0,false,m_doc);
		if (b && ss->m_fi != m_loc.fi) b=false;
		if (!b) break;
		p0=ss->m_seq-1;
	}
	if (!b) {
		ssd->Advance(m_loc.seq,false,-1);
		p0=ssd->m_seq;
	}
	int pos0=(b?ss:ssd)->m_off;
	int fi0=(b?ss:ssd)->m_fi;
	int p1=m_loc.seq+m_loc.nw; // I doubt this - no it's right, see solveInDivision
	b=true;
	for (int i=0;i<scope+1;i++) {
		b=ss->Advance(p1,true,m_doc);
		if (b && ss->m_fi != m_loc.fi) b=false;
		if (!b) break;
		p1=ss->m_seq+1;
	}
	if (!b) {
		sse->Advance(m_loc.seq+m_loc.nw,true,-1);
		p1=sse->m_seq;
	}
	int pos1=(b?ss:sse)->m_off;
	int fi1=(b?ss:sse)->m_fi;
	if (fi0 != m_loc.fi || fi1 != m_loc.fi)
		throw new CXairoError(CXairoIError::ERR_NOTEXT,L("Cannot deduce text"));
	int ipos=m_loc.off-pos0;
	int ipos1=ipos+m_loc.len;
	const xairo_uchar* loc=_s;
	CStream* sl=m_server->GetLocaleStream();
	if (sl && sl->Advance(p0,false)) {
		int x=sl->m_nw;
		if (sl->Advance(sl->m_seq,true) && sl->m_seq >= p0) {
			loc=m_server->GetWL()->GetWordW(x);
		}
	}
	const xairo_uchar* w=m_server->GetEntityManager()->TextExtract(m_loc.fi,pos0,pos1-pos0,&ipos,&ipos1);
	const xairo_uchar* ew=Ellipsis(w,ipos,ipos1,max,loc);
	delete [] w;
	return ew;
}

#define HT_START L("<hns:xaira-hit xmlns:hns=\"http://xaira.org/ns/hit/1.0\">")
#define HT_START1 L("<hns:xaira-hit xmlns:hns=\"http://xaira.org/ns/hit/1.0\" xml:lang=\"")
#define HT_START2 L("\">")
#define HT_END L("</hns:xaira-hit>")
#define HT_HITSTART L("<hns:hit-start/>")
#define HT_HITEND L("<hns:hit-end/>")
#define HT_OMIT1 L("<hns:omitted-characters n=\"")
#define HT_OMIT2 L("\"/>")

enum hnt {HNT_TAG,HNT_TEXT,HNT_COMMENT,HNT_CDATA,HNT_PI,HNT_OMIT,HNT_CHARENT};

struct _charent {
	const xairo_uchar* name;
	xairo_uchar val;
	bool bK;
	_charent():name(NULL),val(0),bK(false) {}
};

struct _node {
	hnt ty;
	xairo_uchar* el;
	int start, length;
	bool bEnd;
	bool bEmpty;
	bool bSupplied, bEndSupplied;
	int d;
	int nom;
	_charent* e;
	const xairo_uchar* text;
	_node* next;
	_node():bEnd(false),bEmpty(false),bSupplied(false),bEndSupplied(false),d(0),e(NULL),next(NULL),text(NULL),el(NULL) {}
	~_node() {if (text) delete [] text; if (el) delete [] el;}
};


// ellipsis formation. 

const xairo_uchar* fstring(const xairo_uchar* fmt,...)
{
	va_list arg;
	va_start(arg,fmt);
	int n=_vscwprintf(fmt,arg);
	xairo_uchar* x=new xairo_uchar [n+1];
	va_end(arg);
	va_start(arg,fmt);
	_vsnwprintf(x,n+1,fmt,arg);
	va_end(arg);
	return x;
}

const xairo_uchar* CXairoHit::Ellipsis(const xairo_uchar* w,int i0,int i1,int max,const xairo_uchar* loc)
{
	int k=0; int k1=wcslen(w);
	_node* n;
	_node* fn=NULL;
	_node** next=&fn;
	int x=wcslen(w)+wcslen(HT_HITSTART)+wcslen(HT_HITEND);
	xairo_uchar* ww=new xairo_uchar[x+1];
	wcsncpy(ww,w,i0);
	wcscpy(ww+i0,HT_HITSTART);
	wcsncat(ww,w+i0,i1-i0);
	wcscpy(ww+wcslen(HT_HITSTART)+i1,HT_HITEND);
	wcscat(ww,w+i1);
	i1+=wcslen(HT_HITSTART)+wcslen(HT_HITEND);
	while (n=GetNode(ww,k)) {
		*next=n;
		next=&n->next;
	}
// now remove distant nodes
	n=fn;
	if (i0 > max && max >= 0) {
		while (n) {
			if (n->start+n->length >= i0-max) break;
			n=n->next;
		}
		if (n->ty==HNT_TEXT) {
			int n0=n->start;
			n->start=i0-max;
			n->length-=(i0-max-n0);
		}
		_node* ne=new _node;
		ne->ty=HNT_OMIT;
		ne->next=n;
		ne->nom=n->start;
		fn=ne;
	}
	if (i1+max<(int) wcslen(ww) && max >= 0)  {
		_node* n0=NULL;
		while (n) {
			if (n->start > i1+max) break;
			n0=n;
			n=n->next;
		}
		if (n0->ty==HNT_TEXT) {
			n0->length=i1+max-n0->start;
		}
		_node* ne=new _node;
		ne->ty=HNT_OMIT;
		ne->nom=wcslen(ww)-(n0->start+n0->length);
		n0->next=ne;
		next=&ne->next;
	}
// maybe truncate tail end tags
	n=fn;
	_node* lastn=n;
	int d=0;
	while (FALSE) { // was n
		n->d=d;
		if (n->ty==HNT_TAG) {
			if (n->bEnd) {
				d--;
				if (d < 0 && n->start>i1) {
					lastn->next=NULL;
					break;
				}
			}
			else if (n->bEmpty) {
			}
			else {
				d++;
			}

		}
		else
		{
		}
		lastn=n;
		n=n->next;
	}
// supply missing nodes
	n=fn;
	d=0;
	while (n) {
		n->d=d;
		if (n->ty==HNT_TAG) {
			if (n->bEnd) {
				d--;
				if (d < 0) {
					_node* n0=new _node;
					n0->ty=HNT_TAG;
					n0->el=_wcsdup(n->el);
					n0->bSupplied=true;
					n0->next=fn;
					n0->d=n->d;
					for (_node* m=fn;m != n;m=m->next)
						m->d++;
					fn=n0;
					d=0;
				}
			}
			else if (n->bEmpty) {
			}
			else {
				d++;
			}

		}
		else
		{
		}
		n=n->next;
	}
	for (int i=d-1;i>=0;i--) {
		_node* n=fn; _node* x=NULL;
		while (n)  {
			if (n->d==i) x=n;
			n=n->next;
		}
		_node* n1=new _node;
		n1->ty=HNT_TAG;
		n1->el=_wcsdup(x->el);
		n1->bSupplied=true;
		n1->bEnd=true;
		x->bEndSupplied=true;
		n1->d=i;
		*next=n1;
		next=&n1->next;
	}

	int t=wcslen(HT_START)+wcslen(HT_END);
	if (loc[0] && loc[0] != '#') t=wcslen(HT_START1)+wcslen(loc)+wcslen(HT_START2)+wcslen(HT_END);
	n=fn;
	while (n) {
		switch (n->ty) {
			case HNT_TAG:
				if (n->bSupplied) 
					if (n->bEnd) 
						n->text=fstring(L("</%s>"),n->el);
					else
						n->text=fstring(L("<%s hns:added=\"start\">"),n->el);
				else if (n->bEndSupplied) {
					const xairo_uchar* s=AttEnts(ww+n->start,n->length);
					int t=n->bEmpty?2:1;
					xairo_uchar* text=new xairo_uchar [wcslen(s)+wcslen(L(" hns:added=\"end\""))+1];
					wcsncpy(text,s,wcslen(s)-t);
					wcscpy(text+wcslen(s)-t,L(" hns:added=\"end\""));
					wcscat(text,s+wcslen(s)-t);
					n->text=text;
				}
				else 
					n->text=AttEnts(ww+n->start,n->length);
				break;
			case HNT_OMIT:
				n->text=fstring(L("<hns:omitted-characters n=\"%d\"/>"),n->nom);
				break;
			case HNT_CHARENT: {
				const xairo_uchar* x=CanonicalEntity(n->e,false);
				if (n->e->name) 
					n->text=fstring(L("<hns:entity original-name=\"%s\">%s</hns:entity>"),n->e->name,x);
				else 
					n->text=x;
				break; }
			default:
				xairo_uchar* text=new xairo_uchar[n->length+1];
				wcsncpy(text,ww+n->start,n->length);
				text[n->length]=0;
				n->text=text;
		}
		t+=wcslen(n->text);
		n=n->next;
	}
	delete [] ww;
	xairo_uchar* www=new xairo_uchar[t+1];
	if (loc[0] && loc[0] != '#') {
		wcscpy(www,HT_START1);
		wcscat(www,loc);
		wcscat(www,HT_START2);
	} else {
		wcscpy(www,HT_START);
	}
	n=fn;
	while (n) {
		wcscat(www,n->text);
		_node* nn=n->next;
		delete n;
		n=nn;
	}
	wcscat(www,HT_END);
	return www;
}

#define UNKENT '?'

const xairo_uchar* CXairoHit::CanonicalEntity(_charent *e,int context)
{
	bool bFull=false; 
	switch (context) {
		case 0: // content
			bFull=e->val=='&' || e->val== '<';
			break;
		case 1:
			bFull=e->val=='&' || e->val == '\'';
			break;
		case 2:
			bFull=e->val=='&' || e->val == '\"';
			break;
	}
	if (bFull) {
		switch (e->val) {
			case '&':
				return _wcsdup(L("&amp;"));
			case '<':
				return _wcsdup(L("&lt;"));
			case '\"':
				return _wcsdup(L("&quot;"));
			case '\'':
				return _wcsdup(L("&apos;"));
		}
	}
	xairo_uchar* x=new xairo_uchar[2];
	x[0]=e->val?e->val:UNKENT;
	x[1]=0;
	return x;
}

const xairo_uchar* CXairoHit::AttEnts(const xairo_uchar* w,int n)
{
	int k=0;
	bool bQ=false; bool bQQ=false; int t=0;
	while (k<n) {
		if (w[k]=='\"' && !bQ) {
			bQQ=!bQQ;
		}
		else if (w[k]=='\'' && !bQQ) {
			bQ=!bQ;
		} 
		else if (w[k]=='&' && (bQ || bQQ)) {
			_charent* e=GetEntity(w,k);
			const xairo_uchar* x=CanonicalEntity(e,bQ?1:2);
			t+=wcslen(x);
			continue;
		}
		k++; t++;
	}
	xairo_uchar* a=new xairo_uchar [t+1]; k=0; t=0; bQ=bQQ=false;
	while (k<n) {
		if (w[k]=='\"' && !bQ) {
			bQQ=!bQQ;
		}
		else if (w[k]=='\'' && !bQQ) {
			bQ=!bQ;
		} 
		else if (w[k]=='&' && (bQ || bQQ)) {
			_charent* e=GetEntity(w,k);
			const xairo_uchar* x=CanonicalEntity(e,bQ?1:2);
			wcscpy(a+t,x);
			t+=wcslen(x);
			continue;
		}
		a[t]=w[k];
		k++; t++;
	}
	a[t]=0;
	return a;
}

const xairo_uchar _s_commentstart[]={'<','!','-','-',0};
const xairo_uchar _s_commentend[]={'-','-','>',0};
const xairo_uchar _s_cdatastart[]={'<','!','[','C','D','A','T','A','[',0};
const xairo_uchar _s_cdataend[]={']',']','>',0};
const xairo_uchar _s_pistart[]={'<','?',0};
const xairo_uchar _s_piend[]={'?','>',0};



_node* CXairoHit::GetNode(const xairo_uchar* in,int& k)
{
	if (in[k]==0) return NULL;
	if (wcsncmp(in+k,_s_commentstart,4)==0) {
		const xairo_uchar* w=wcsstr(in+k,_s_commentend);
		int kk=w?w-in+3:wcslen(in);
		_node* n=new _node;
		n->ty=HNT_COMMENT;
		n->start=k;
		n->length=kk-k;
		k=kk;
		return n;
	}
	if (wcsncmp(in+k,_s_cdatastart,9)==0) {
		const xairo_uchar* w=wcsstr(in+k,_s_cdataend);
		int kk=w?w-in+3:wcslen(in);
		_node* n=new _node;
		n->ty=HNT_CDATA;
		n->start=k;
		n->length=kk-k;
		k=kk;
		return n;
	}
	if (wcsncmp(in+k,_s_pistart,2)==0) {
		const xairo_uchar* w=wcsstr(in+k,_s_piend);
		int kk=w?w-in+2:wcslen(in);
		_node* n=new _node;
		n->ty=HNT_PI;
		n->start=k;
		n->length=kk-k;
		k=kk;
		return n;
	}

	if (in[k]=='<') {
		return GetElementNode(in,k);
	}
	if (in[k]=='&') {
		int k0=k;
		_charent* e=GetEntity(in,k);
		_node* n=new _node;
		n->ty=HNT_CHARENT;
		n->e=e;
		n->start=k0;
		n->length=k-k0;
		return n;
	}
	const xairo_uchar* w=wcschr(in+k,'<');
	const xairo_uchar* w1=wcschr(in+k,'&');
	if (w1 != NULL && w1 < w) w=w1;
	int kk=w?w-in:wcslen(in);
	_node* n=new _node;
	n->ty=HNT_TEXT;
	n->start=k;
	n->length=kk-k;
	k=kk;
	return n;
}

bool namechar(int c)
{
	if (c=='>'||c=='/') return false;
	if (c < 128 && isspace((char) c)) return false;
	return true;
}

_node* CXairoHit::GetElementNode(const xairo_uchar* w,int& k)
{
	_node* n=new _node;
	n->ty=HNT_TAG;
	n->start=k;
	k++;
	if (w[k]=='/') {
		n->bEnd=true;
		k++;
	}
	int kk=k;
	while (w[kk] && namechar(w[kk])) kk++;
	n->el=new xairo_uchar [kk-k+1];
	wcsncpy(n->el,w+k,kk-k);
	n->el[kk-k]=0;
	bool bQ=false; bool bQQ=false;
	while (w[kk]) {
		if (w[kk]=='>' && !bQ && !bQQ) break;
		else if (w[kk]=='\"' && !bQQ) {
			bQ=!bQ;
		}
		else if (w[kk]=='\'' && !bQ) {
			bQQ=!bQQ;
		} 
//		else if (w[kk]=='&' && (bQ || bQQ)) {
//			_charent* e=GetEntity(w,kk);
//			continue;
//		}
		kk++;
	}
	if (w[kk-1]=='/') n->bEmpty=true;
	if (w[kk]) kk++;
	n->length=kk-n->start;
	k=kk;
	return n;
}

_charent* CXairoHit::GetEntity(const xairo_uchar* w,int& k)
{
	_charent* e=new _charent;
	const xairo_uchar* ww=wcschr(w+k,';');
	int kk=ww?ww-w:wcslen(w);
	xairo_uchar* name=new xairo_uchar [kk-k];
	wcsncpy(name,w+k+1,kk-k-1);
	name[kk-k-1]=0;
	e->name=name;
	if (e->name[0]=='#') {
		int r=10; int p=1; 
		if (name[1]=='x') {
			r=16;
			p=2;
		}
		int t=0;
		while (xairo_uchar x=name[p++]) {
			int d;
			if ('0' <= x && x <= '9') d=x-'0';
			if ('A' <= x && x <= 'Z') d=x-'A'+10;
			if ('a' <= x && x <= 'z') d=x-'a'+10;
			t=t*r+d;
		}
		e->val=t;
		delete [] e->name;
		e->name=NULL;
		e->bK=true;
	}
	else { 
		e->val=m_server->GetGrammarImpl()->EvalCharEntity(name);
		if (e->val > 0) e->bK=true;
	}
	k=ww?kk+1:kk;
	return e;
}

const xairo_uchar* CXairoHit::getUnit()
{
	CStream* us=m_server->GetUnitStream();
	us->Advance(m_loc.seq,false,m_doc);
	return m_server->GetWL()->GetWordW(us->m_nw);
}



XAIRO_NAMESPACE_END
