#include "SServer.h"
#include "XairoSolution.h"
#include "XairoHit.h"
#include "KDB.h"
#include "StreamFactory.h"
#include "Stream.h"
#include "WordList.h"
#include "EntityManager.h"
#include "Locs.h"
#include "RegularExpression.h"
#include "Thread.h"
#include "SB.h"
#include "Dictionary.h"
#include "XairoCollocationTable.h"
#include "XairoWord.h"
#include "XairoSolveError.h"
#include "XairoCorpusTextFilter.h"
#include "XairoAddkeyFilter.h"

XAIRO_NAMESPACE_BEGIN




CXairoSolution::CXairoSolution(CDBKey* k,const CXairoGrammar::Partition* p,CXairoServer* s):m_key(k),m_part(p),m_server(s)
{
	m_bOwns=true;
}

CXairoSolution::~CXairoSolution()
{
	if (m_bOwns) release();
}

int CXairoSolution::getCount()
{
	return m_server->GetDB()->countrecords(m_key);
}


// would it not be kinder to move text extraction to CXairoHit?
// This would mean stats calls were not burdened with it
CXairoHit* CXairoSolution::getHit(int n)
{
	xara_loc l;
	if (!m_server->GetDB()->recorded(m_key,n,&l)) return NULL;
	CStream* ds=m_server->GetDocumentStream();
	ds->Advance(l.seq,false);
	int doc=((CAtomStream*) ds)->GetCursor();
	CXairoHit* h=new CXairoHit(l,doc,m_server);
	return h;
}





CXairoISolution* CXairoSolution::thinInit(int i)
{
	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	if (i > getCount()) i=getCount();
	for (int j=0;j<i;j++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,j,&l);
		m_server->GetDB()->record(k,&l,sizeof(l),1);
	}
	m_server->GetDB()->closekey(k);
	return new CXairoSolution(k,m_part,m_server);
}

#define MAXRANDTHIN 1000

CXairoISolution* CXairoSolution::thinRand(int i,int seed)
{
	if (i > MAXRANDTHIN) 
		throw new CXairoError(CXairoIError::ERR_LIMIT,L("Too many random solutions"));
	srand(seed);
	int n=getCount();
	if (i > getCount()) i=getCount();
	if (i==getCount()) return thinInit(i);
	bool * x=new bool [n];
	for (int j=0;j<n;j++)
		x[j]=false;
  	if (i > getCount()) i=getCount();
	for (int j=0;j<i;j++) {
		int k;
		do {
			k=(int)((rand()*(__int64) n)/RAND_MAX);
		} while (k<n && x[k]);
		x[k]=true;
	}
  	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	for (int j=0;j<n;j++) {
		xara_loc l;
		if (x[j]) {
			m_server->GetDB()->recorded(m_key,j,&l);
			m_server->GetDB()->record(k,&l,sizeof(l),1);
		}
	}
	m_server->GetDB()->closekey(k);
	delete [] x;
	return new CXairoSolution(k,m_part,m_server);
}

int CXairoSolution::getCorpusTextCount()
{
	CAtomStream* ds=m_server->GetDocumentStream();
	int n=getCount(); int dseq=-1; int dc=0; int nt=0;
	m_server->GetDB()->fop(m_key);
	for (int i=0;i<n;i++) {
		xara_loc l;
		m_server->GetDB()->frecorded(m_key,i,&l);
		if (l.seq < dseq) continue;
		nt++;
		do {
			xara_loc* dl=ds->SetCursor(dc++);
			if (dl) dseq=dl->seq; else dseq=m_server->GetLocs()->GetMaxLoc(); // needs widening
		} while (dseq < l.seq);
	}
	m_server->GetDB()->fcl(m_key);
	return nt;
}

CXairoISolution* CXairoSolution::thinOnePerText()
{
 	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	CAtomStream* ds=m_server->GetDocumentStream();
	int n=getCount(); int dseq=-1; int dc=0; int nt=0;
	for (int i=0;i<n;i++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,i,&l);
		if (l.seq < dseq) continue;
		m_server->GetDB()->record(k,&l,sizeof (l),1);
		do {
			xara_loc* dl=ds->SetCursor(dc++);
			if (dl) dseq=dl->seq; else dseq=m_server->GetLocs()->GetMaxLoc();
		} while (dseq < l.seq);
	}
	m_server->GetDB()->closekey(k);
	return new CXairoSolution(k,m_part,m_server);
}

CXairoISolution* CXairoSolution::copy()
{
	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	for (int j=0;j<getCount();j++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,j,&l);
		m_server->GetDB()->record(k,&l,sizeof(l),1);
	}
	m_server->GetDB()->closekey(k);
	return new CXairoSolution(k,m_part,m_server);
}

CXairoISolution* CXairoSolution::thinSelection(int i,int* x)
{
	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	if (i > getCount()) i=getCount();
	for (int j=0;j<i;j++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,x[j],&l);
		m_server->GetDB()->record(k,&l,sizeof(l),1);
	}
	m_server->GetDB()->closekey(k);
	return new CXairoSolution(k,m_part,m_server);
}


CXairoISolution* CXairoSolution::thinCorpusTextFilter(const CXairoICorpusTextFilter* f)
{
	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	int* sc;
	int nsc=((CXairoCorpusTextFilter*) f)->getTexts(sc);
	CAtomStream* ds=m_server->GetDocumentStream();
	int n=getCount(); int dseq=-1; int dc=0; int nt=0;
	bool b=false; int d=0;
	for (int i=0;i<n;i++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,i,&l);
		if (l.seq < dseq) {
			if (b) m_server->GetDB()->record(k,&l,sizeof(l),1);
			continue;
		}
		do {
			b=(0 < dc && dc-1<nsc)?(sc[dc-1]!= 0):false;
			xara_loc* dl=ds->SetCursor(dc++);
			if (dl) dseq=dl->seq; else dseq=m_server->GetLocs()->GetMaxLoc();
		} while (dseq < l.seq);
		if (b) {m_server->GetDB()->record(k,&l,sizeof (l),1); d++;}
	}
	m_server->GetDB()->closekey(k);
	delete [] sc;
	if (d==0) return NULL;
	else return new CXairoSolution(k,m_part,m_server);
}

CXairoISolution* CXairoSolution::thinAddkeyFilter(const CXairoIAddkeyFilter* f)
{
	CDBKey* k=m_server->GetDB()->newkey(sizeof(xara_loc));
	CXairoAddkeyFilter* ff=(CXairoAddkeyFilter*) f;
	const xairo_uchar* ak=ff->getKey();
	int nk=m_server->GetWL()->addkey_index(ak);
	int n=getCount(); int d=0;
	for (int i=0;i<n;i++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,i,&l);
		int k1=l.seq;
		int k2=l.seq+l.nw; bool bOK=false;
		for (int j=k1;j<k2+100;j++) {
			int sbi=m_server->GetThread()->GetThread(j);
			xara_streambase* sb=m_server->GetSB()->GetSB(sbi);
			if (sb->w.t != it_content) continue;
			int a=sb->w.con[nk];
			int jj=0;
			for (jj=0;jj<ff->getCount();jj++) {
				if (a==ff->getValN(jj)) {
					break;
				}
			}
			bOK=jj<ff->getCount();
			break;
		}
		if (bOK) {
			m_server->GetDB()->record(k,&l,sizeof (l),1);
			d++;
		}
	}
	m_server->GetDB()->closekey(k);
	if (d==0) return NULL;
	else return new CXairoSolution(k,m_part,m_server);
}

CXairoICollocationTable* CXairoSolution::makeCollocationTable(const xairo_uchar* pattern,const xairo_uchar* key,CXairoCollocationTable::cmeth meth,CXairoCollocationTable::_cwin c,int maxn,double mins,double maxs,int lowf)
{
	CRegularExpression re(pattern,false);
	if (!re.Parse()) return NULL;
	int nhits=getCount();
	int scs=m_part?m_part->getWordCount():m_server->getWordCount();
	int d=0;
	int ik=m_server->GetWL()->lemma_index(key);
	if (ik < 0) ik=m_server->GetWL()->addkey_index(key);
	if (meth >= CXairoICollocationTable::CM_MAX || meth < 0) return NULL;
	CStream* s1=NULL, *s2=NULL;
	initCT();
	if (c.bX) {
		s1=m_server->GetSF()->MakeAtomStream(c.s,c.ns,it_start);
		if (s1==NULL) return NULL;
		s2=m_server->GetSF()->MakeAtomStream(c.s,c.ns,it_end);
		if (s2==NULL) return NULL;
	}
	for (int i=0;i<nhits;i++) {
		xara_loc loc;
		int wl, wr;
		m_server->GetDB()->recorded(m_key,i,&loc);
		if (c.bX) { 
// note that mw is not stored in the solution for same daft reason
// so we have to go and get it
			if (!s1->Advance(loc.seq-1,false)) continue;
			wl=loc.sb-s1->m_nw;
			if (wl < 0) wl=0;
			if (!s2->Advance(s1->m_seq+1,true)) continue;
			if (s2->m_seq < loc.seq) continue;
			wr=s2->m_nw-loc.sb;
			if (wr < 0) wr=0;
		}
		else {
			wl=c.wl;
			wr=c.wr;
		}
		for (int p=loc.seq-1;wl;p--)
		{
			int sbi=m_server->GetThread()->GetThread(p);
			xara_streambase* sb=m_server->GetSB()->GetSB(sbi);
			if (sb->w.t==it_enddoc) break;
			if (sb->w.t == it_content) {
				int p=m_server->GetDict(ik)->GetHeadwordIndex(&sb->w);
				int wn=m_server->GetDict(ik)->LookupEntry(p)->word[0];
				const xairo_uchar* wd=m_server->GetWL()->GetWordW(wn);
				if (re.Match(wd)) addCT(p,ik);
				d++;
				wl--;
			}
		}
		for (int p=loc.seq+loc.nw;wr;p++)
		{
			int sbi=m_server->GetThread()->GetThread(p);
			xara_streambase* sb=m_server->GetSB()->GetSB(sbi);
			if (sb->w.t==it_enddoc) break;
			if (sb->w.t == it_content) {
				int p=m_server->GetDict(ik)->GetHeadwordIndex(&sb->w);
				int wn=m_server->GetDict(ik)->LookupEntry(p)->word[0];
				const xairo_uchar* wd=m_server->GetWL()->GetWordW(wn);
				if (re.Match(wd)) addCT(p,ik);
				d++;
				wr--;
			}
		}
	}
	for (int i=0;i<m_ncolltab;i++) {
		m_colltab[i].score=cscore(meth,m_colltab[i].x,m_colltab[i].freq,d,nhits,scs);
	}
	qsort(m_colltab,m_ncolltab,sizeof (struct _colt),qc);
	CDBKey* ky=m_server->GetDB()->newkey(sizeof(struct _colt));
	int p=0;
	for (int i=0;i<m_ncolltab;i++) {
		_colt* c=&m_colltab[i];
		if (c->freq <= lowf) continue;
		if (c->score < mins || c->score >maxs) continue;
		if (p>=maxn) break;
		m_server->GetDB()->record(ky,&m_colltab[i],sizeof(struct _colt),1);
		p++;
	}
	m_server->GetDB()->closekey(ky);
	delete [] m_colltab;
	CXairoCollocationTable* x=new CXairoCollocationTable(ky,m_server);
	x->m_addkey=key;
	x->m_d=ik;
	if (c.bX) {
		c.wl=c.wr=0;
	}
	else {
		c.s=c.ns=L("");
	}
	x->m_cwin=c;
	return x;
}

void CXairoSolution::addCT(int p,int ik)
{
	int i;
	for (i=0;i<m_ncolltab;i++) {
		if (p==m_colltab[i].p) {
			m_colltab[i].x++;
			return;
		}
	}
	i=m_ncolltab;
	if (i >= m_ctsize) {
		_colt* t=new _colt[m_ctsize+CTALLOC];
		if (t==NULL) 
			throw new CXairoError(CXairoIError::ERR_MEMORY,L("Out of memory allocating collocation table"));
		for (int j=0;j<i;j++) {
			t[j]=m_colltab[j];
		}
		delete [] m_colltab;
		m_colltab=t;
		m_ctsize+=CTALLOC;
	}
	_colt* c=&m_colltab[i];
//	c->w=w;
	c->x=1;
	c->p=p;
	if (p != -1) 
	{
		xara_lemma* l=m_server->GetDict(ik)->LookupEntry(p);
		c->freq=l->freq;
	} else 
		c->freq=0; 
	m_ncolltab++;
}


void CXairoSolution::initCT()
{
	m_colltab=new _colt[CTALLOC];
	if (m_colltab==NULL) 
		throw new CXairoError(CXairoIError::ERR_MEMORY,L("Out of memory allocating collocation table"));
	m_ncolltab=0;
	m_ctsize=CTALLOC;
}

int CXairoSolution::qc(const void* v1,const void* v2)
{
	struct _colt* c1=(struct _colt*) v1;
	struct _colt* c2=(struct _colt*) v2;
	double c=c2->score-c1->score;
	if (c>0) return 1;
	else if (c<0) return -1;
	else return 0;
}



double CXairoSolution::cscore(CXairoCollocationTable::cmeth m,int k,int f,int d,int nh,int scs)
{
	double s; double p=((double) f)/scs;
	switch (m) {
		case CXairoCollocationTable::CM_Z:
			s=k-p*d;
			s=s/sqrt(d*p*(1-p));
			break;
		case CXairoCollocationTable::CM_MI:
			s=k*((double)scs)/f*nh;
			s=log(s)/log((double) 2);
			break;
	}
	return s;
}

CXairoISolution* CXairoSolution::thinCollocation(const xairo_uchar* key,CXairoCollocationTable::_cwin w,int n,const xairo_uchar** sel)
{
	int nhits=getCount();
	int ik=m_server->GetWL()->lemma_index(key);
	if (ik < 0) ik=m_server->GetWL()->addkey_index(key);
	CDBKey* nk=m_server->GetDB()->newkey(sizeof(xara_loc));
	CStream* s1=NULL, *s2=NULL; int t=0;
	if (w.bX) {
		s1=m_server->GetSF()->MakeAtomStream(w.s,w.ns,it_start);
		if (s1==NULL) return NULL;
		s2=m_server->GetSF()->MakeAtomStream(w.s,w.ns,it_end);
		if (s2==NULL) return NULL;
	}
	for (int i=0;i<nhits;i++) {
		bool b=false;
		xara_loc loc;
		int wl, wr;
		m_server->GetDB()->recorded(m_key,i,&loc);
		if (w.bX) { 
			if (!s1->Advance(loc.seq-1,false)) continue;
			wl=loc.sb-s1->m_nw;
			if (wl < 0) wl=0;
			if (!s2->Advance(s1->m_seq+1,true)) continue;
			if (s2->m_seq < loc.seq) continue;
			wr=s2->m_nw-loc.sb;
			if (wr < 0) wr=0;
		}
		else {
			wl=w.wl;
			wr=w.wr;
		}
		for (int p=loc.seq-1;wl;p--)
		{
			int sbi=m_server->GetThread()->GetThread(p);
			xara_streambase* sb=m_server->GetSB()->GetSB(sbi);
			if (sb->w.t==it_enddoc) break;
			if (sb->w.t == it_content) {
//				int w=(ik < 0)?sb->w.val:sb->w.con[ik];
//				const xairo_uchar* wd=m_server->GetWL()->GetWordW(w);
				int p=m_server->GetDict(ik)->GetHeadwordIndex(&sb->w);
				int wn=m_server->GetDict(ik)->LookupEntry(p)->word[0];
				const xairo_uchar* wd=m_server->GetWL()->GetWordW(wn);
				for (int j=0;j<n;j++) {
					if (wcscmp(wd,sel[j])==0) 
						b=true;
				}
				wl--;
			}
		}
		for (int p=loc.seq+loc.nw;wr;p++)
		{
			int sbi=m_server->GetThread()->GetThread(p);
			xara_streambase* sb=m_server->GetSB()->GetSB(sbi);
			if (sb->w.t==it_enddoc) break;
			if (sb->w.t == it_content) {
//				int w=(ik < 0)?sb->w.val:sb->w.con[ik];
//				const xairo_uchar* wd=m_server->GetWL()->GetWordW(w);
				int p=m_server->GetDict(ik)->GetHeadwordIndex(&sb->w);
				int wn=m_server->GetDict(ik)->LookupEntry(p)->word[0];
				const xairo_uchar* wd=m_server->GetWL()->GetWordW(wn);
				for (int j=0;j<n;j++) {
					if (wcscmp(wd,sel[j])==0) 
						b=true;
				}
				wr--;
			}
		}
		if (b) {m_server->GetDB()->record(nk,&loc,sizeof(loc),1); t++;}
	}
	m_server->GetDB()->closekey(nk);
	if (t==0) 
		return NULL;
	else 
		return new CXairoSolution(nk,m_part,m_server);
}



void CXairoSolution::getAnalysis(int np,int *p,int max,int* ana,int* tana)
{
	for (int i=0;i<max;i++)
		ana[i]=tana[i]=0;
	CAtomStream* ds=m_server->GetDocumentStream();
	int n=getCount(); int dseq=-1; int dc=0; int nt=0;
	int iSel=-1;
	for (int i=0;i<n;i++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,i,&l);
		if (l.seq > dseq) {
			bool b=ds->Advance(l.seq+1,true);
			int d=ds->Count()-1;
			if (b) d=ds->GetCursor()-1;
			iSel=(d >= 0 && d < np)?p[d]:0;
			if (iSel >= max) iSel=-1; else tana[iSel]++;
			if (b) dseq=ds->m_seq; else dseq=0x7fffffff;
		}
		if (iSel >=0) ana[iSel]++;
	}
}

void CXairoSolution::getAnalysisScope(int np,int *p,int max,int* ana,int* tana,const xairo_uchar* e)
{
	for (int i=0;i<max;i++)
		ana[i]=tana[i]=0;
	CAtomStream* ds=m_server->GetDocumentStream();
	CStream* s1=m_server->GetSF()->MakeAtomStream(e,L(""),it_start);
	CStream* s2=m_server->GetSF()->MakeAtomStream(e,L(""),it_end);
	int n=getCount(); int dseq=-1; int dc=0; int nt=0;
	int iSel=-1;
	for (int i=0;i<n;i++) {
		xara_loc l;
		m_server->GetDB()->recorded(m_key,i,&l);
		if (l.seq > dseq) {
			ds->Advance(l.seq+1,true);
			int d=ds->GetCursor()-1;
			iSel=(d >= 0 && d < np)?p[d]:0;
			if (iSel >= max) iSel=-1; else tana[iSel]++;
			dseq=ds->m_seq;
		}
		if (iSel >=0) ana[iSel]+=MeasureScope(&l,s1,s2);
	}
	if (s1) delete s1;
	if (s2) delete s2;
}

int CXairoSolution::MeasureScope(xara_loc *l,CStream* s1,CStream* s2)
{
	if (s1==NULL || s2==NULL) return 0;
	if (!s1->Advance(l->seq,false)) return 0;
	if (!s2->Advance(l->seq,true)) return 0;
	return s2->m_nw-s1->m_nw;
}

#define FSOL L("sol(%d)")

const xairo_uchar* CXairoSolution::getId()
{
	m_bOwns=false;
	return CXairoServer::MakeId(FSOL,m_key->getId());
}

CXairoSolution::CXairoSolution(const xairo_uchar* id,CXairoServer* s):m_server(s)
{
	int k;
	CXairoServer::ScanId(FSOL,id,&k);
	m_key=m_server->GetDB()->getKey(k,sizeof(xara_loc));
	m_part=NULL;
	m_bOwns=false;
}

void CXairoSolution::release()
{
	m_server->GetDB()->deletekey(m_key);
}

XAIRO_NAMESPACE_END
