#include "SServer.h"
#include "XairoXpathValue.h"
#include "XairoCorpusText.h"

XAIRO_NAMESPACE_BEGIN

CXairoXpathValue::CXairoXpathValue(int n,CXairoCorpusText* c):m_ct(c)
{
	xairo_uchar* w=new xairo_uchar [10];
	swprintf(w,L("%d"),n);
	m_string=w;
	m_type=XPT_NUMBER;
}

CXairoXpathValue::CXairoXpathValue(const xairo_uchar* s,_xptype t,CXairoCorpusText* c):m_ct(c),m_string(s)
{
	m_type=t;
}

CXairoXpathValue::CXairoXpathValue(bool b,CXairoCorpusText* c):m_bool(b),m_ct(c)
{
	m_type=XPT_BOOL;
}

CXairoXpathValue::~CXairoXpathValue(void)
{
}

CXairoXpathValue::CXairoXpathValue(_xrange* x,CXairoCorpusText* c):m_ct(c)
{
	m_type=XPT_NODESET;
	m_n=m_nalloc=1;
	m_nodes=new _xrange [1];
	m_nodes[0]=*x;
}

CXairoXpathValue::CXairoXpathValue(CXairoCorpusText* c):m_ct(c)
{
	m_type=XPT_NODESET;
	m_n=m_nalloc=0;
	m_nodes=NULL;
}

int CXairoXpathValue::Count()
{
	return m_n;
}

void CXairoXpathValue::Add(CXairoXpathValue* x)
{
	for (int i=0;i<x->Count();i++) 
		Add(x->Node(i));
}

/*
void CXairoXpathValue::Add(_xrange* x)
{
	_xrange* xn=new _xrange [m_n+1]; int d=0;
	bool b=false;
	for (int i=0;i<m_n;i++) {
		_xrange* t=&m_nodes[i];
		if (t->start.seq > x->start.seq && !b) 
		{
			xn[d]=*x;
			b=true; d++;
		}
		xn[d]=m_nodes[i]; d++;
	}
	if (!b) xn[d]=*x;
	if (m_nodes) delete [] m_nodes;
	m_nodes=xn;
	m_n++;
}
*/

#define ALLOC 1000

void CXairoXpathValue::Add(_xrange* x)
{
//	_xrange* xn=new _xrange [m_n+1]; int d=0;
//	bool b=false;
	if (m_n >= m_nalloc) {
		int n=m_nalloc+ALLOC;
		_xrange* xn=new _xrange[n];
		for (int i=0;i<m_n;i++) {
			xn[i]=m_nodes[i];
		}
		if (m_nodes) delete [] m_nodes;
		m_nodes=xn;
		m_nalloc=n;
	}
	int i=0;
	for (i=0;i<m_n;i++) {
		_xrange* t=&m_nodes[i];
		if (t->start.seq > x->start.seq) break;
	}
	for (int j=m_n-1;j >= i;j--) {

		{
			m_nodes[j+1]=m_nodes[j];
		}
	}
	m_nodes[i]=*x;
	m_n++;
}

_xrange* CXairoXpathValue::Node(int i)
{
	return &m_nodes[i];
}
	

const xairo_uchar* CXairoXpathValue::getText()
{
	switch (m_type) {
		case XPT_NODESET:
			if (Count()==0) return L("");
			else return m_ct->NodeText(&m_nodes[0]);
		case XPT_BOOL:
			return m_bool?L("true"):L("false");
		case XPT_STRING:
			return m_string;
		case XPT_NUMBER:
			return m_string;
	}
	return L("");
}


bool CXairoXpathValue::getBool()
{
	switch (m_type) {
		case XPT_NODESET:
			return Count() > 0;
		case XPT_BOOL:
			return m_bool;
		case XPT_STRING:
			return wcslen(m_string) > 0;
		case XPT_NUMBER:
			return _wtoi(m_string) != 0;
	}
	return false;
}


int CXairoXpathValue::getNumber()
{
	switch (m_type) {
		case XPT_NODESET:
			return _wtoi(getText());
		case XPT_BOOL:
			return m_bool?1:0;
		case XPT_STRING:
			return _wtoi(m_string);
		case XPT_NUMBER:
			return _wtoi(m_string);
	}
	return 0;
}

bool CXairoXpathValue::Compare(CXairoXpathValue* v)
{
	if (m_type==XPT_NODESET && v->m_type==XPT_NODESET) {
		for (int i=0;i<Count();i++) {
			for (int j=0;j<v->Count();j++) {
				const xairo_uchar* s1=m_ct->NodeText(&m_nodes[i]);
				const xairo_uchar* s2=m_ct->NodeText(&v->m_nodes[j]);
				if (wcscmp(s1,s2) ==0) return true;
			}
		}
		return false;
	} else if (m_type==XPT_NODESET) {
		switch (v->m_type) {
			case XPT_STRING:
		for (int i=0;i<Count();i++) {
			const xairo_uchar* s1=m_ct->NodeText(&m_nodes[i]);
			if (wcscmp(s1,v->m_string) ==0) return true;
		}
		return false;
			case XPT_NUMBER:
		for (int i=0;i<Count();i++) {
			const xairo_uchar* s1=m_ct->NodeText(&m_nodes[i]);
			if (_wtoi(s1)==_wtoi(v->m_string)) return true;
		}
		return false;
			case XPT_BOOL:
				return v->m_bool==(Count()>0);
		}
	} else if (v->m_type==XPT_NODESET)
		return v->Compare(this);
	if (m_type==XPT_BOOL) 
		return m_bool==v->getBool();
	else if (v->m_type==XPT_BOOL) 
		return getBool()==v->m_bool;
	else if (m_type==XPT_NUMBER) 
		return _wtoi(m_string)==v->getNumber();
	else if (v->m_type==XPT_NUMBER) 
		return getNumber()==_wtoi(v->m_string);
	else
		return wcscmp(m_string,v->m_string)==0;
}

XAIRO_NAMESPACE_END
