#define _CRT_NON_CONFORMING_SWPRINTFS
#define _CRT_SECURE_NO_DEPRECATE

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "RPCUtils.h"

namespace PROXY_UTILS {

DOMElement* RPCParam(DOMElement* e,XMLSize_t n)
{
	DOMNodeList* nl=e->getElementsByTagName(L("param"));
	XMLSize_t k=nl->getLength();
	if (n >= k) return NULL;
	DOMNode* nd=nl->item(n);
	return (DOMElement*) nd;
}

const xairo_uchar* RPCName(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("name"));
	XMLSize_t k=nl->getLength();
	if (k==0) return NULL;
	DOMNode* nd=nl->item(0);
	return nd->getTextContent();
}	


DOMElement* RPCValue(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("value"));
	XMLSize_t k=nl->getLength();
	if (k==0) return NULL;
	DOMNode* nd=nl->item(0);
	return (DOMElement*) nd;
}	

DOMElement* RPCStruct(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("struct"));
	DOMNode* nd=nl->item(0);
	return (DOMElement*) nd;
}	


const xairo_uchar* RPCString(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("string"));
	XMLSize_t k=nl->getLength();
	if (k==0) return L("");
	DOMNode* nd=nl->item(0);
	const xairo_uchar* w=L("");
	try {
		w=nd->getTextContent();
	}
	catch (DOMException &d) {
		xairo_uchar err[100];
		_snwprintf(err,100,L("DOM error %d"),d.code);
		RPCError(err);
	}
	return w;
}
	
int RPCInt(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("i4"));
	XMLSize_t k=nl->getLength();
	if (k==0) return 0;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* x=nd->getTextContent();
	return _wtoi(x);
}

double RPCDouble(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("double"));
	XMLSize_t k=nl->getLength();
	if (k==0) return 0;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* x=nd->getTextContent();
	return _wtof(x);
}
	
bool RPCBool(DOMElement* e)
{
	DOMNodeList* nl=e->getElementsByTagName(L("boolean")); // really?
	XMLSize_t k=nl->getLength();
	if (k==0) return 0;
	DOMNode* nd=nl->item(0);
	const xairo_uchar* x=nd->getTextContent();
	return _wtoi(x) != 0;
}
	

int RPCIntParam(DOMElement* d,int i)
{
	return RPCInt(RPCValue(RPCParam(d,i)));
}

double RPCDoubleParam(DOMElement* d,int i)
{
	return RPCDouble(RPCValue(RPCParam(d,i)));
}

bool RPCBoolParam(DOMElement* d,int i)
{
	return RPCBool(RPCValue(RPCParam(d,i)));
}

const xairo_uchar* RPCStringParam(DOMElement* d,int i)
{
	return RPCString(RPCValue(RPCParam(d,i)));
}

int RPCVectorParam(DOMElement* d,int i,int *& v)
{
	return RPCVector(RPCValue(RPCParam(d,i)),v);
}

int RPCVectorsParam(DOMElement* d,int i,const xairo_uchar**& v)
{
	return RPCVectors(RPCValue(RPCParam(d,i)),v);
}

int RPCVector(DOMElement* e,int*& v)
{
	DOMNodeList* nl=e->getElementsByTagName(L("data"));
	if (nl->getLength()==0) return 0;
	DOMElement* ex=(DOMElement*) nl->item(0);
	DOMNodeList* nll=ex->getChildNodes();
	int k=0;
	for (int i=0;i<(int) nll->getLength();i++) {
		DOMNode* n=nll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ne=(DOMElement*) n;
		if (wcscmp(ne->getTagName(),L("value")) != 0) continue;
		k++;
	}
	v=new int [k];
	k=0;
	for (int i=0;i<(int) nll->getLength();i++) {
		DOMNode* n=nll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ne=(DOMElement*) n;
		if (wcscmp(ne->getTagName(),L("value")) != 0) continue;
		v[k++]=RPCInt(ne);
	}
	return k;
}

int RPCVectors(DOMElement* e,const xairo_uchar**& v)
{
	DOMNodeList* nl=e->getElementsByTagName(L("data"));
	if (nl->getLength()==0) return 0;
	DOMElement* ex=(DOMElement*) nl->item(0);
	DOMNodeList* nll=ex->getChildNodes();
	int k=0;
	for (int i=0;i<(int) nll->getLength();i++) {
		DOMNode* n=nll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ne=(DOMElement*) n;
		if (wcscmp(ne->getTagName(),L("value")) != 0) continue;
		k++;
	}
	v=new const xairo_uchar* [k];
	k=0;
	for (int i=0;i<(int) nll->getLength();i++) {
		DOMNode* n=nll->item(i);
		if (n->getNodeType() != DOMNode::ELEMENT_NODE) continue;
		DOMElement* ne=(DOMElement*) n;
		if (wcscmp(ne->getTagName(),L("value")) != 0) continue;
		v[k++]=RPCString(ne);
	}
	return k;
}


#define FSTRINGRET L("<methodResponse><params><param><value><string>%s</string></value></param></params></methodResponse>")

const xairo_uchar* RPCStringRet(const xairo_uchar* s)
{
	int x=_scwprintf(FSTRINGRET,RPCEscape(s));
	xairo_uchar* ss=new xairo_uchar[x+1];
	swprintf(ss,FSTRINGRET,RPCEscape(s));
	return ss;
}

#define FI4RET L("<methodResponse><params><param><value><i4>%d</i4></value></param></params></methodResponse>")

const xairo_uchar* RPCIntRet(int n)
{
	int x=_scwprintf(FI4RET,n);
	xairo_uchar* ss=new xairo_uchar[x+1];
	swprintf(ss,FI4RET,n);
	return ss;
}

#define FBOOLEANRET L("<methodResponse><params><param><value><boolean>%d</boolean></value></param></params></methodResponse>")

const xairo_uchar* RPCBoolRet(bool b)
{
	int x=_scwprintf(FBOOLEANRET,b?L("1"):L("0"));
	xairo_uchar* ss=new xairo_uchar[x+1];
	swprintf(ss,FBOOLEANRET,b?1:0);
	return ss;
}

#define P1 L("<params>")
#define P2 L("</params>")
#define A1 L("<param><value><array><data>")
#define A2 L("</data></array></value></param>")

const xairo_uchar* RPCParams(int n,const xmlrpc_arg* x)
{
	int tx=(int) (wcslen(P1)+wcslen(P2));
	for (int i=0;i<n;i++) {
		switch (x[i].t) {
			int t;
			case XMLRPC_STRING:
				t=_scwprintf(L("<param><value><string>%s</string></value></param>"),RPCEscape(x[i].s));
				tx+=t;
				break;
			case XMLRPC_I4:
				t=_scwprintf(L("<param><value><i4>%d</i4></value></param>"),x[i].i);
				tx+=t;
				break;
			case XMLRPC_BOOL:
				t=_scwprintf(L("<param><value><boolean>%d</boolean></value></param>"),x[i].i);
				tx+=t;
				break;
			case XMLRPC_DOUBLE:
				t=_scwprintf(L("<param><value><double>%f</double></value></param>"),x[i].d);
				tx+=t;
				break;
			case XMLRPC_VECTOR:
				tx+=(int) (wcslen(A1)+wcslen(A2));
				for (int j=0;j<x[i].nvec;j++) {
					tx+=_scwprintf(L("<value><i4>%d</i4></value>"),x[i].vec[j]);
				}
				break;
			case XMLRPC_VECTORS:
				tx+=(int) (wcslen(A1)+wcslen(A2));
				for (int j=0;j<x[i].nvec;j++) {
					tx+=_scwprintf(L("<value><string>%s</string></value>"),RPCEscape(x[i].vecs[j]));
				}
				break;
		}
	}
	xairo_uchar* w=new xairo_uchar [tx+1];
	wcscpy(w,P1); xairo_uchar* ww=w+wcslen(P1);
	for (int i=0;i<n;i++) {
		switch (x[i].t) {
			case XMLRPC_STRING:
				ww+=swprintf(ww,L("<param><value><string>%s</string></value></param>"),RPCEscape(x[i].s));
				break;
			case XMLRPC_I4:
				ww+=swprintf(ww,L("<param><value><i4>%d</i4></value></param>"),x[i].i);
				break;
			case XMLRPC_BOOL:
				ww+=swprintf(ww,L("<param><value><boolean>%d</boolean></value></param>"),x[i].i);
				break;
			case XMLRPC_DOUBLE:
				ww+=swprintf(ww,L("<param><value><double>%f</double></value></param>"),x[i].d);
				break;
			case XMLRPC_VECTOR:
				wcscpy(ww,A1);
				ww+=wcslen(A1);
				for (int j=0;j<x[i].nvec;j++) {
					ww+=swprintf(ww,L("<value><i4>%d</i4></value>"),x[i].vec[j]);
				}
				wcscpy(ww,A2);
				ww+=wcslen(A2);
				break;
			case XMLRPC_VECTORS:
				wcscpy(ww,A1);
				ww+=wcslen(A1);
				for (int j=0;j<x[i].nvec;j++) {
					ww+=swprintf(ww,L("<value><string>%s</string></value>"),RPCEscape(x[i].vecs[j]));
				}
				wcscpy(ww,A2);
				ww+=wcslen(A2);
				break;
		}
	}
	wcscpy(ww,P2);
	return w;
}


#define MR1 L("<methodResponse>")
#define MR2 L("</methodResponse>")
#define MN1 L("<methodName>")
#define MN2 L("</methodName>")
#define MC1 L("<methodCall>")
#define MC2 L("</methodCall>")

const xairo_uchar* RPCRet(int n,xmlrpc_arg * x)
{
	const xairo_uchar* p=RPCParams(n,x);
	int tx=(int) (wcslen(p)+wcslen(MR1)+wcslen(MR2));
	xairo_uchar* w=new xairo_uchar[tx+1];
	wcscpy(w,MR1);
	wcscat(w,p);
	wcscat(w,MR2);
	return w;
}

const xairo_uchar* RPCCall(const xairo_uchar* m,int n,const xmlrpc_arg* x)
{
	const xairo_uchar* p=RPCParams(n,x);
	int tx=int(wcslen(m)+wcslen(p)+wcslen(MC1)+wcslen(MC2)+wcslen(MN1)+wcslen(MN2));
	xairo_uchar* w=new xairo_uchar[tx+1];
	wcscpy(w,MC1);
	wcscat(w,MN1);
	wcscat(w,m);
	wcscat(w,MN2);
	wcscat(w,p);
	wcscat(w,MC2);
	delete [] p;
	return w;
}

const xairo_uchar* RPCEscape(const xairo_uchar* s)
{
	int t=0;
	for (int i=0;i<(int) wcslen(s);i++) {
		if (s[i]=='<') t+=4;
		else if (s[i]=='&') t+=5;
		else t++;
	}
	xairo_uchar* w=new xairo_uchar [t+1]; w[0]=0;
	xairo_uchar* ww=w;
	for (int i=0;i<(int) wcslen(s);i++) {
		if (s[i]=='<') {
			wcscpy(ww,L("&lt;")); ww+=4;
		}
		else if (s[i]=='&') {
			wcscpy(ww,L("&amp;")); ww+=5;
		}
		else {
			*(ww++)=s[i]; 
		}
	}
	*ww=0;
	return w;
}

const xairo_uchar* RPCUnescape(const xairo_uchar* s)
{
	xairo_uchar* w=new xairo_uchar [wcslen(s)+1];
	xairo_uchar* ww=w;
	for (int i=0;i<(int) wcslen(s);i++) {
		if (s[i]=='&') {
			if (wcsncmp(s+i,L("&lt;"),4)==0) {
				*ww++='<'; i+=3;
			} else if (wcsncmp(s+i,L("&amp;"),5)==0) {
				*ww++='&'; i+=4;
			}
			*ww++=s[i];
		}
	}
	*ww=0;
	return w;
}

#define FF L("<methodResponse><fault><value><struct><member><name>faultCode</name><value><i4>%d</i4></value></member><member><name>faultString</name><value><string>%s</string></value></member></struct></value></fault></methodResponse>")

const xairo_uchar* RPCFault(int n,const xairo_uchar* msg)
{
	return RPCFormat(FF,n,msg);
}

const xairo_uchar* RPCFormat(const xairo_uchar* fmt,...)
{
	va_list arg;
	va_start(arg,fmt);
	int x=_vscwprintf(fmt,arg);
	xairo_uchar* s=new xairo_uchar [x+1];
	vswprintf(s,fmt,arg);
	va_end(arg);
	return s;
}

void (*f_err)(const xairo_uchar*)=NULL;

void RPCSetErrorHandler(void f(const xairo_uchar*))
{
}

void RPCError(const xairo_uchar* s)
{
	if (f_err) f_err(_wcsdup(s));
}

}